/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.TPeriodicMatrix;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.CloneException;

/**
 * TData Editor
 * 
 * @author Christoph Halbmayr
 */
public class Editor {

	/** Editor Action Listener */
	private ActionListener actionListner;

	/** MainTable - Data Overview */
	private final EditorListPanel editorListPanel;

	/** Contains MainTables */
	private final Vector<EditorPanelInterface> EditorPanels;

	/** Editor Key Listener */
	private KeyListener keyListener;

	/** Editor Mouse Listener */
	private MouseListener mouseListner;

	/** Navigation Label - on the bottom */
	private final JLabel navigationBar;

	/** Contains the text for the Navigation Label */
	private final Vector<String> navigationBarText;

	/** Editor JPanel */
	private final JPanel panel;

	/** Index of actual selectet EditorPanel */
	private int panelIndex = 0;

	/** Editor Table Mode lListener */
	private TableModelListener tableModelListner;

	/**
	 * Creats an empty Object
	 */
	public Editor() {
		this.panel = new JPanel();

		SwingUtilities.updateComponentTreeUI(this.panel);

		this.panel.setLayout(new BorderLayout());
		this.editorListPanel = new EditorListPanel();
		this.initActionListener();
		this.initMouseListener();
		this.initKeyListener();
		this.initTableModelListener();
		this.editorListPanel.addListener(this.actionListner, this.keyListener, this.mouseListner);
		this.editorListPanel.addTableModelListener(this.tableModelListner);
		this.editorListPanel.addNewVarRow();
		this.EditorPanels = new Vector<EditorPanelInterface>();
		this.EditorPanels.add(this.editorListPanel);

		this.panel.add(this.editorListPanel.getPanel(), BorderLayout.CENTER);

		this.panel.addKeyListener(this.keyListener);

		this.navigationBar = new JLabel();

		this.panel.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, true), "showKey");
		this.navigationBarText = new Vector<String>();
		this.navigationBarText.add("> ");
		String text = this.navigationBarText.toString();
		text = text.replaceAll("\\[", "");
		text = text.replaceAll("\\]", "");

	}

	/**
	 * Add a data to the editor
	 * 
	 * @param name
	 *            Name of data
	 * @param data
	 *            Data to add
	 */
	public void addData(String name, TData data) {

		try {
			this.editorListPanel.addData(name, data.deepClone());
		} catch (final CloneException e) {

			e.printStackTrace();
		}

		this.panel.repaint();
		this.panel.revalidate();
		this.updatePnav();

	}

	/**
	 * Returns the Editor JPanel
	 * 
	 * @return the panel
	 */
	public JPanel getPanel() {
		return this.panel;
	}

	/**
	 * Returns the TData to key
	 * 
	 * @param key
	 *            Key
	 * @return TData
	 */
	public TData getTData(final String key)

	{
		final TStruct struct = (TStruct) this.EditorPanels.get(0).getTData();

		return struct.get(key);

	}

	/**
	 * Returns the keys
	 * 
	 * @return Key List
	 */
	public String[] getTDataList() {

		final TStruct struct = (TStruct) this.EditorPanels.get(0).getTData();

		return struct.getKeys();

	}

	/**
	 * Generates a FileChooser with *.xml Filter
	 * 
	 * @return fc JFileChooser
	 */
	private JFileChooser createFileChooser() {

		final FileNameExtensionFilter filter = new FileNameExtensionFilter("xml files", "xml");
		final JFileChooser fc = new JFileChooser();
		fc.addChoosableFileFilter(filter);
		return fc;

	}

	/**
	 * Initializes the ActionListener
	 */
	private void initActionListener() {
		this.actionListner = new ActionListener() {

			/**
			 * Action Event, needed for the Navigation Panel
			 * 
			 * @param arg0
			 *            Action Event
			 */
			@Override
			public void actionPerformed(final ActionEvent arg0) {

				if (arg0.getActionCommand().equals("Open Selection")) {
					Editor.this.openSelection();
				}

				if (arg0.getActionCommand().equals("show Generator")) {
					((EditorChangePanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).showGenerator();
				}

				if (arg0.getSource().getClass().getName().equals("javax.swing.JTextField")) {
					((EditorChangePanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).changeOffset();
					Editor.this.updatePnav();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();

					Editor.this.TDataWriteBack();
				}

				if (arg0.getActionCommand().equals("Import Data")) {

					final JFileChooser fc = Editor.this.createFileChooser();
					final int rval = fc.showOpenDialog(Editor.this.panel);

					if (rval == JFileChooser.APPROVE_OPTION) {
						Editor.this.EditorPanels.get(Editor.this.panelIndex).importTData(fc.getSelectedFile());
					} else if (rval == JFileChooser.CANCEL_OPTION) {
					}

					Editor.this.TDataWriteBack();

				}
				if (arg0.getActionCommand().equals("Save Data")) {

					final JFileChooser fc = Editor.this.createFileChooser();

					final int rval = fc.showSaveDialog(Editor.this.panel);

					if (rval == JFileChooser.APPROVE_OPTION) {
						File f = fc.getSelectedFile();

						if (!f.toString().endsWith(".xml")) {
							f = new File(f.getPath() + ".xml");
						}

						Editor.this.EditorPanels.get(Editor.this.panelIndex).saveTData(f);

					} else if (rval == JFileChooser.CANCEL_OPTION) {
					}

				}

				if (arg0.getActionCommand().equals("Delete")) {

					Editor.this.EditorPanels.get(Editor.this.panelIndex).delete();

					if (Editor.this.panelIndex != 0) {
						Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
					} else {
						Editor.this.editorListPanel.reload();
					}
				}

				if (arg0.getActionCommand().equals("orderDelete")) {
					Editor.this.TDataWriteBack();
				}

				if (arg0.getActionCommand().equals("up")) {
					Editor.this.up();
				}

				if (arg0.getActionCommand().equals("Print")) {
					Editor.this.EditorPanels.get(Editor.this.panelIndex).printme();
				}

				if (arg0.getActionCommand().equals("SET")) {

					((EditorMatrixPanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).addNewTData();

					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
					Editor.this.updatePnav();
				}

				if (arg0.getActionCommand().equals("Row")) {

					Editor.this.EditorPanels.get(Editor.this.panelIndex).deleteRow();

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
				}

				if (arg0.getActionCommand().equals("addRow")) {

					(Editor.this.EditorPanels.get(Editor.this.panelIndex)).insertRow();

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
				}
				if (arg0.getActionCommand().equals("addCol")) {

					(Editor.this.EditorPanels.get(Editor.this.panelIndex)).insertColumn();

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
				}

				if (arg0.getActionCommand().equals("TPMPreview")) {
					((EditorMatrixPanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).showTPMPreview();
				}

				if (arg0.getActionCommand().equals("Column")) {

					Editor.this.EditorPanels.get(Editor.this.panelIndex).deleteColumn();

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
				}

				if (arg0.getActionCommand().equals("RESIZE")) {

					((EditorChangePanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).tnumResize();

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
				}
				if (arg0.getActionCommand().equals("writeback&reload")) {
					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
					Editor.this.EditorPanels.get(Editor.this.panelIndex - 1).reload();
					Editor.this.EditorPanels.get(Editor.this.panelIndex).reload();
					((EditorChangePanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).createHeaders();
				}
			}

		};
	}

	/**
	 * Initializes the Editor Key Listener
	 */
	private void initKeyListener() {
		this.keyListener = new KeyListener() {

			/**
			 * Key Press Key event
			 * 
			 * @param arg0
			 *            KeyEvent
			 */
			@Override
			public void keyPressed(final KeyEvent arg0) {
				final int keycode = arg0.getKeyCode();

				if (keycode == KeyEvent.VK_ENTER) {
					try {
						Editor.this.openSelection();

					} catch (final Exception e) {
					}
				} else if (keycode == KeyEvent.VK_ESCAPE) {
					Editor.this.EditorPanels.get(Editor.this.panelIndex).reload();
				} else if (keycode == KeyEvent.VK_BACK_SPACE) {
					Editor.this.up();
				}

			}

			/**
			 * Not in use
			 */
			@Override
			public void keyReleased(final KeyEvent arg0) {
			}

			/**
			 * Not in use
			 */
			@Override
			public void keyTyped(final KeyEvent arg0) {

			}

		};
	}

	/**
	 * Initializes the Editor Mouse Listener
	 */
	private void initMouseListener() {
		this.mouseListner = new MouseListener() {

			/**
			 * Mouse Clicked, needed to select a signal to edit
			 * 
			 * @param e
			 *            {@link MouseEvent}
			 */
			@Override
			public void mouseClicked(final MouseEvent e) {

				if ((e.getButton() == MouseEvent.BUTTON1) && (e.getClickCount() == 2)) {

					Editor.this.openSelection();

					Editor.this.updatePnav();

				}

			}

			/**
			 * Not in use
			 */
			@Override
			public void mouseEntered(final MouseEvent e) {

			}

			/**
			 * Not in use
			 */
			@Override
			public void mouseExited(final MouseEvent e) {

			}

			/**
			 * Not in use
			 */
			@Override
			public void mousePressed(final MouseEvent e) {

			}

			/**
			 * Not in use
			 */
			@Override
			public void mouseReleased(final MouseEvent e) {
			}

		};

	}

	/**
	 * Initializes the Table Model Listener
	 */
	private void initTableModelListener() {
		this.tableModelListner = new TableModelListener() {

			/**
			 * tableChanged is triggers by the TableModeListener
			 * 
			 * @param arg0
			 *            {@link TableModelEvent}
			 */
			@Override
			public void tableChanged(final TableModelEvent arg0) {

				if (arg0.getType() == TableModelEvent.INSERT) {
				} else if (arg0.getType() == TableModelEvent.UPDATE) {
					final int row = arg0.getFirstRow();
					final int column = arg0.getColumn();

					if (arg0.getSource() instanceof TableModelChangePanel) {
						((EditorChangePanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).changed(row, column);
					} else if (arg0.getSource() instanceof TableModelListPanel) {
						((EditorListPanel) Editor.this.EditorPanels.get(Editor.this.panelIndex)).changed(row, column);
					}

					Editor.this.updatePnav();
					Editor.this.TDataWriteBack();
				} else if (arg0.getType() == TableModelEvent.DELETE) {
				}

			}
		};
	}

	/**
	 * opens a selection
	 */
	private void openSelection() {

		final TData data = this.EditorPanels.get(this.panelIndex).getSelectedTData();
		if (data != null) {
			if (data instanceof TArray) {
				final TArray ta = (TArray) data;
				final EditorMatrixPanel mt = new EditorMatrixPanel(ta, this.tableModelListner);

				this.panel.removeAll();
				this.panel.add(mt.getPanel(), BorderLayout.CENTER);

				this.EditorPanels.add(mt);
				this.panelIndex++;
				this.panel.repaint();
				this.panel.revalidate();
				mt.addListener(this.actionListner, this.keyListener, this.mouseListner);

			} else if (data instanceof TPeriodicMatrix) {
				final TPeriodicMatrix tpm = (TPeriodicMatrix) data;
				final EditorMatrixPanel mt = new EditorMatrixPanel(tpm, this.tableModelListner);

				this.panel.removeAll();
				this.panel.add(mt.getPanel(), BorderLayout.CENTER);

				this.EditorPanels.add(mt);
				this.panelIndex++;
				this.panel.repaint();
				this.panel.revalidate();
				mt.addListener(this.actionListner, this.keyListener, this.mouseListner);

			} else if (data instanceof TStruct) {
				final TStruct ts = (TStruct) data;
				final EditorListPanel ms = new EditorListPanel();
				ms.addTStruct(ts);
				ms.addTableModelListener(this.tableModelListner);
				this.panel.removeAll();
				this.panel.add(ms.getPanel(), BorderLayout.CENTER);

				this.EditorPanels.add(ms);
				this.panelIndex++;
				this.panel.repaint();
				this.panel.revalidate();
				ms.addListener(this.actionListner, this.keyListener, this.mouseListner);

			} else if ((data instanceof TString) || (data instanceof TBoolean) || (data instanceof TNumeric)
					|| (data instanceof TPeriodic)) {
				final EditorChangePanel et = new EditorChangePanel(data);
				et.addTableModelListener(this.tableModelListner);
				et.addListener(this.actionListner);
				this.panel.removeAll();
				this.panel.add(et.getPanel(), BorderLayout.CENTER);

				this.EditorPanels.add(et);
				this.panelIndex++;
				this.panel.repaint();
				this.panel.revalidate();
			}
		}

	}

	/**
	 * Writes back the TData from the selectet Panel to the Ground Panel
	 * (panlindex =0)
	 */
	private void TDataWriteBack() {

		for (int i = this.panelIndex; i > 0; i--) {
			this.EditorPanels.get(i - 1).updateChangedSignal(this.EditorPanels.get(i).getTData());
			this.EditorPanels.get(i - 1).reload();
		}
	}

	/**
	 * Navigates one level up
	 */
	private void up() {
		this.panel.removeAll();
		if (this.panelIndex > 0) {
			this.EditorPanels.remove(this.panelIndex);

			this.panel.add(this.EditorPanels.get(this.panelIndex - 1).getPanel());
			this.panelIndex--;
		} else {
			this.panel.add(this.editorListPanel.getPanel());
		}
		this.updatePnav();
		this.panel.repaint();
		this.panel.revalidate();

	}

	/**
	 * updates the navigation Bar
	 */
	private void updatePnav() {
		this.navigationBarText.removeAllElements();
		TData data;

		for (int i = 0; i < this.panelIndex; i++) {
			data = this.EditorPanels.get(i).getTData();
			if (data instanceof TArray) {
				this.navigationBarText.add("{" + ((EditorMatrixPanel) this.EditorPanels.get(i)).getSelectedRow() + ","
						+ ((EditorMatrixPanel) this.EditorPanels.get(i)).getSelectedColumn() + "}");
			} else if (data instanceof TPeriodicMatrix) {
				this.navigationBarText.add("(" + ((EditorMatrixPanel) this.EditorPanels.get(i)).getSelectedRow() + ","
						+ ((EditorMatrixPanel) this.EditorPanels.get(i)).getSelectedColumn() + ")");
			} else if (data instanceof TStruct) {
				this.navigationBarText.add(((EditorListPanel) this.EditorPanels.get(i)).getSelectedStructName());
			}

		}
		String text = this.navigationBarText.toString();
		text = text.replaceAll("\\[", "");
		text = text.replaceAll("\\]", "");
		text = text.replaceAll("\\,", ".");
		this.navigationBar.setText(text);
		this.panel.remove(this.navigationBar);
		this.panel.add(this.navigationBar, BorderLayout.SOUTH);

	}
}
