/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.table.JTableHeader;
import javax.swing.text.JTextComponent;

import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicModel;

/**
 * This panel allows to analyze and manipulate TPeriodic. It based on
 * TPeriodicModel and TPeriodicTableModel for JTables.
 * 
 * @author Johannes Kaisinger
 */
public class TPeriodicPanel extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Local reference of the tNumericModel. It allows to manipulate the
	 * TPeriodic.
	 */
	private TPeriodicModel tPeriodicModel;

	/** Contains a table model for the JTable. */
	private TPeriodicTableModel tableModel;

	/** Editor as table. */
	private JTable table;

	/** Serialization version number */
	private TPeriodicSignalBuilder signalBuilder;

	/** Serialization version number */
	private RowHeaderTable rowHeaderTable;

	/**
	 * Creates an object.
	 * 
	 * @param tPeriodicModel
	 *            TPeriodicModel
	 */
	public TPeriodicPanel(TPeriodicModel tPeriodicModel) {

		this.tPeriodicModel = tPeriodicModel;
		this.setUpPanel();
	}

	/**
	 * Set up the whole Panel. Should be used when the tPeriodicModel has
	 * changed.
	 */
	private void setUpPanel() {
		this.setLayout(new BorderLayout());

		JToolBar toolbar = this.createToolBar();
		this.add(toolbar, BorderLayout.NORTH);

		this.tableModel = new TPeriodicTableModel(this.tPeriodicModel);
		this.table = new JTable(this.tableModel) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			// Implement table cell tool tips.
			@Override
			public String getToolTipText(MouseEvent e) {
				String tip = null;
				java.awt.Point p = e.getPoint();
				int rowIndex = this.rowAtPoint(p);
				int colIndex = this.columnAtPoint(p);

				try {
					tip = this.getValueAt(rowIndex, colIndex).toString();
				} catch (RuntimeException e1) {
					// catch null pointer exception if mouse is over an empty
					// line
				}

				return tip;
			}
		};

		this.table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

		this.table.setDefaultEditor(Double.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(final JTable table, final Object value,
					final boolean isSelected, final int row, final int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();
				return c;
			}
		});

		this.createMouseListener();

		JTableHeader tableHeader = this.table.getTableHeader();
		tableHeader.setReorderingAllowed(false);

		JScrollPane scroll = new JScrollPane(this.table);

		this.rowHeaderTable = new RowHeaderTable(this.table);
		scroll.setRowHeaderView(this.rowHeaderTable);
		scroll.setCorner(ScrollPaneConstants.UPPER_LEFT_CORNER, this.rowHeaderTable.getTableHeader());

		final JLabel dcLabel = new JLabel("DC-Offset: ", SwingConstants.RIGHT);

		final JTextField dcField = new JTextField();

		dcField.setText("" + this.tPeriodicModel.getDcOffset());
		dcField.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e) {
				try {
					TPeriodicPanel.this.tPeriodicModel.setDcOffset(Double.parseDouble(dcField.getText()));
				} catch (NumberFormatException ex) {
					dcField.setText("" + TPeriodicPanel.this.tPeriodicModel.getDcOffset());
				}
			}

			@Override
			public void focusGained(FocusEvent e) {
				dcField.selectAll();
			}
		});

		dcField.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				dcLabel.requestFocus();
			}
		});

		JPanel dcPanel = new JPanel();
		dcPanel.setLayout(new GridLayout(1, 2));
		dcPanel.setBorder(BorderFactory.createLineBorder(Color.gray));

		dcPanel.add(dcLabel);
		dcPanel.add(dcField);

		this.createPopupMenu();

		this.add(scroll, BorderLayout.CENTER);
		this.add(dcPanel, BorderLayout.SOUTH);
		this.setVisible(true);
	}

	/**
	 * Adds a mouse listener to the table.
	 */
	private void createMouseListener() {
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				if (e.getClickCount() == 2) {
					Component editor = TPeriodicPanel.this.table.getEditorComponent();
					editor.requestFocusInWindow();
					((JTextComponent) editor).selectAll();
				}
			}
		});
	}

	/**
	 * Add a sub-signal.
	 */
	private void addSignal() {
		this.tPeriodicModel.addSignal();
		this.tableModel.fireTableDataChanged();

		this.table.addRowSelectionInterval(this.table.getRowCount() - 1, this.table.getRowCount() - 1);
		this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
	}

	/**
	 * Removes the selected sub.signal.
	 */
	private void removeSignal() {
		int row = this.table.getSelectedRow();

		if (row > -1) {

			int order = (Integer) this.table.getValueAt(row, 0);

			this.tPeriodicModel.removeSignal(order);
			this.tableModel.fireTableDataChanged();

			if (this.table.getRowCount() > row) {
				this.table.addRowSelectionInterval(row, row);
				this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
			} else if (this.table.getRowCount() > 0) {
				this.table.addRowSelectionInterval(this.table.getRowCount() - 1, this.table.getRowCount() - 1);
				this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
			}
		}

	}

	/**
	 * Create the popupMenu and mousListener for the TPeriodicPanel.
	 * 
	 * @return popupMenu
	 */
	private JPopupMenu createPopupMenu() {

		final JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem addSignal = new JMenuItem("Add signal");
		JMenuItem removeSignal = new JMenuItem("Remove signal");

		popupMenu.add(addSignal);
		popupMenu.add(removeSignal);

		addSignal.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.addSignal();
			}
		});

		removeSignal.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.removeSignal();
			}
		});

		this.table.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					JTable source = (JTable) e.getSource();
					int row = source.rowAtPoint(e.getPoint());
					int column = source.columnAtPoint(e.getPoint());

					if (!source.isRowSelected(row)) {
						source.changeSelection(row, column, false, false);
					}

					popupMenu.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		});

		return popupMenu;
	}

	/**
	 * Create the toolBar for the TPeriodicPanel.
	 * 
	 * @return toolBar
	 */
	private JToolBar createToolBar() {

		JToolBar tb = new JToolBar();
		tb.setPreferredSize(new Dimension(500, 50));

		Vector<JButton> buttons = new Vector<JButton>();

		final JButton addRow = new JButton("Add", IconEditor.ICON_ADD);
		addRow.setToolTipText("Insert a row ");
		addRow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.addSignal();
			}
		});
		buttons.add(addRow);

		final JButton removeRow = new JButton("Remove", IconEditor.ICON_REMOVE);
		removeRow.setToolTipText("Delete a row ");
		removeRow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.removeSignal();
			}
		});
		buttons.add(removeRow);

		final JButton buildSignal = new JButton("Signal", IconEditor.ICON_GENERATOR);
		buildSignal.setToolTipText("Build a new signal");
		buildSignal.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.signalBuilder = new TPeriodicSignalBuilder();
				TPeriodicPanel.this.signalBuilder.addWindowListener(TPeriodicPanel.this.closeMonitor);
			}
		});
		buttons.add(buildSignal);

		final JButton showTime = new JButton("Time", IconEditor.ICON_XYPLOT);
		showTime.setToolTipText("Show time plot");
		showTime.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.tPeriodicModel.showTime();
			}
		});
		buttons.add(showTime);

		final JButton showSpectrum = new JButton("Spectrum", IconEditor.ICON_SPECTRUM);
		showSpectrum.setToolTipText("Show spectrum plot");
		showSpectrum.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicPanel.this.tPeriodicModel.showSpectrum();
			}
		});
		buttons.add(showSpectrum);

		for (final JButton i : buttons) {

			Dimension d = new Dimension(55, 50);
			i.setVerticalTextPosition(SwingConstants.BOTTOM);
			i.setHorizontalTextPosition(SwingConstants.CENTER);
			i.setMaximumSize(d);
			i.setMinimumSize(d);
			i.setPreferredSize(d);
			tb.add(i);
		}
		return tb;
	}

	/**
	 * WindowListener for the SignalBuilder. Get the signal which has been
	 * design in the SignalBuilder.
	 */
	private WindowListener closeMonitor = new WindowAdapter() {
		@Override
		public void windowClosing(WindowEvent e) {

		}

		@Override
		public void windowClosed(WindowEvent e) {

			// TODO: This is a complete reset off the panel.
			// This is not the best way to do that.
			// Instead only the table should be updated.

			TPeriodicModel temp = TPeriodicPanel.this.signalBuilder.getSignal();

			TPeriodicPanel.this.tPeriodicModel.setData(temp.getData());
			TPeriodicPanel.this.removeAll();
			TPeriodicPanel.this.setUpPanel();

			TPeriodicPanel.this.repaint();
			TPeriodicPanel.this.revalidate();
		}
	};
}
