/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.text.NumberFormat;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.text.JTextComponent;

import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicSignalBuilderModel;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TPeriodicSignalBuilderModel.Signal;

/**
 * @author Johannes Kaisinger
 *
 */
public class TPeriodicSignalBuilder extends JFrame {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Local reference of the tPeriodicSignalBuilderModel. It allows to
	 * manipulate the TPeriodicSignal.
	 */
	private TPeriodicSignalBuilderModel tPeriodicSignalBuilderModel;

	/** Editor as table. */
	private JTable table;

	/** Contains a table model for the JTable. */
	private TPeriodicSignalBuilderTableModel tableModel;

	// private JTextField maxOrderValue;

	/**
	 * Creates a signalbuilder.
	 */
	public TPeriodicSignalBuilder() {
		this.setLayout(new BorderLayout());

		JToolBar toolbar = this.createToolBar();
		this.add(toolbar, BorderLayout.NORTH);

		this.tPeriodicSignalBuilderModel = new TPeriodicSignalBuilderModel();
		this.tableModel = new TPeriodicSignalBuilderTableModel(this.tPeriodicSignalBuilderModel);
		this.table = new JTable(this.tableModel) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			// Implement table cell tool tips.
			@Override
			public String getToolTipText(MouseEvent e) {
				String tip = null;
				java.awt.Point p = e.getPoint();
				int rowIndex = this.rowAtPoint(p);
				int colIndex = this.columnAtPoint(p);

				try {
					tip = this.getValueAt(rowIndex, colIndex).toString();
				} catch (RuntimeException e1) {
					// catch null pointer exception if mouse is over an empty
					// line
				}

				return tip;
			}
		};
		this.table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

		JTableHeader tableHeader = this.table.getTableHeader();
		tableHeader.setReorderingAllowed(false);

		this.table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		this.table.addRowSelectionInterval(0, 0);
		this.table.addColumnSelectionInterval(0, 0);

		this.setUpTableEditors();
		this.createMouseListener();
		this.setUpClosePanel();

		JScrollPane scroll = new JScrollPane(this.table);

		RowHeaderTable rowHeaderTable = new RowHeaderTable(this.table);
		scroll.setRowHeaderView(rowHeaderTable);
		scroll.setCorner(ScrollPaneConstants.UPPER_LEFT_CORNER, rowHeaderTable.getTableHeader());

		JPanel centerPanel = new JPanel();
		centerPanel.setLayout(new BorderLayout());

		JPanel orderPanel = new JPanel();
		orderPanel.setBorder(BorderFactory.createLineBorder(Color.gray));
		orderPanel.setLayout(new GridLayout(1, 2));

		final JLabel maxOrderLabel = new JLabel("Max order: ", SwingConstants.RIGHT);

		NumberFormat integerFieldFormatter = NumberFormat.getIntegerInstance();

		final JFormattedTextField maxOrderValue = new JFormattedTextField(integerFieldFormatter);
		maxOrderValue.setText("" + this.tPeriodicSignalBuilderModel.getMaxOrder());

		maxOrderValue.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e) {
				try {
					int val = Integer.parseInt(maxOrderValue.getText());
					TPeriodicSignalBuilder.this.tPeriodicSignalBuilderModel.setMaxOrder(val);
				} catch (NumberFormatException nfe) {
				}
			}

			@Override
			public void focusGained(FocusEvent e) {
				maxOrderValue.selectAll();
			}
		});

		maxOrderValue.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				maxOrderLabel.requestFocus();
			}
		});

		orderPanel.add(maxOrderLabel);
		orderPanel.add(maxOrderValue);

		centerPanel.add(orderPanel, BorderLayout.SOUTH);
		centerPanel.add(scroll, BorderLayout.CENTER);

		this.add(centerPanel, BorderLayout.CENTER);

		this.createPopupMenu();

		this.setSize(500, 500);
		this.setLocationRelativeTo(null);
		this.setTitle("TPeriodic Signal Builder");
		this.setVisible(true);
	}

	/**
	 * Get all defined signals as one TPeriodicModel.
	 * 
	 * @return Combine signal. (TPeriodic)
	 */
	public TPeriodicModel getSignal() {
		return this.tPeriodicSignalBuilderModel.getAllSignals();
	}

	/**
	 * Set up the table cell dditors for Integer and Double classes.
	 */
	private void setUpTableEditors() {

		TableColumn sigColumn = this.table.getColumnModel().getColumn(0);
		JComboBox<Signal> comboBox = new JComboBox<Signal>(Signal.values());
		sigColumn.setCellEditor(new DefaultCellEditor(comboBox));

		this.table.setDefaultEditor(Integer.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(final JTable table, final Object value,
					final boolean isSelected, final int row, final int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();
				return c;
			}
		});

		this.table.setDefaultEditor(Double.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(final JTable table, final Object value,
					final boolean isSelected, final int row, final int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();
				return c;
			}
		});
	}

	/**
	 * Adds a mouse listener to the table.
	 */
	private void createMouseListener() {
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				if (e.getClickCount() == 2) {
					Component editor = TPeriodicSignalBuilder.this.table.getEditorComponent();
					try {
						editor.requestFocusInWindow();
						((JTextComponent) editor).selectAll();
					} catch (Exception ex) {
					}

				}
			}
		});
	}

	/**
	 * Insert a new signal above the selected signal.
	 * 
	 * @param sig
	 *            New signal to be inserted.
	 */
	private void insertSignalAbove(Signal sig) {
		int row = this.table.getSelectedRow();

		if (row > -1) {
			this.tPeriodicSignalBuilderModel.insertSignal(row, sig);
			this.tableModel.fireTableDataChanged();

			this.table.addRowSelectionInterval(row, row);
		}
	}

	/**
	 * Insert a new signal below the selected signal.
	 * 
	 * @param sig
	 *            New signal to be inserted.
	 */
	private void insertSignalBelow(Signal sig) {
		int row = this.table.getSelectedRow();

		if (row > -1) {
			this.tPeriodicSignalBuilderModel.insertSignal(row + 1, sig);
			this.tableModel.fireTableDataChanged();

			this.table.addRowSelectionInterval(row + 1, row + 1);
		}
	}

	/**
	 * Remove the selected signal.
	 */
	private void removeSignal() {
		int row = this.table.getSelectedRow();
		int maxRow = this.table.getRowCount();

		if ((maxRow > 1) && (row > -1)) {
			this.tPeriodicSignalBuilderModel.removeSignal(row);
			this.tableModel.fireTableDataChanged();

			if (row < (maxRow - 1)) {
				this.table.addRowSelectionInterval(row, row);
			} else if (row == (maxRow - 1)) {
				this.table.addRowSelectionInterval(row - 1, row - 1);
			}
		}
	}

	/**
	 * Show the preview of the selected signal(Spectrum).
	 */
	private void showTime() {
		int row = this.table.getSelectedRow();

		if (row > -1) {
			this.tPeriodicSignalBuilderModel.showTime(row);
		}
	}

	/**
	 * Show the preview of the selected signal(Time).
	 */
	private void showSpectrum() {
		int row = this.table.getSelectedRow();

		if (row > -1) {
			this.tPeriodicSignalBuilderModel.showSpectrum(row);
		}
	}

	/**
	 * 
	 */
	private void showAll() {
		this.tPeriodicSignalBuilderModel.showTimeAll();
	}

	/**
	 * Create the popupMenu and mousListener for the TPeriodicSignalBuilder;
	 * 
	 * @return popupMenu
	 */
	private JPopupMenu createPopupMenu() {

		final JPopupMenu popupMenu = new JPopupMenu();

		JMenu insertSignalAbove = new JMenu("Insert signal above");
		JMenuItem insertRectAbove = new JMenuItem("Rectangle");
		JMenuItem insertSinAbove = new JMenuItem("Sinus");
		JMenuItem insertTriAbove = new JMenuItem("Triangle");

		JMenu insertSignalBelow = new JMenu("Insert signal below");
		JMenuItem insertRectBelow = new JMenuItem("Rectangle");
		JMenuItem insertSinBelow = new JMenuItem("Sinus");
		JMenuItem insertTriBelow = new JMenuItem("Triangle");

		JMenuItem removeSignal = new JMenuItem("Remove signal");

		popupMenu.add(insertSignalAbove);
		insertSignalAbove.add(insertRectAbove);
		insertSignalAbove.add(insertSinAbove);
		insertSignalAbove.add(insertTriAbove);

		popupMenu.add(insertSignalBelow);
		insertSignalBelow.add(insertRectBelow);
		insertSignalBelow.add(insertSinBelow);
		insertSignalBelow.add(insertTriBelow);

		popupMenu.add(removeSignal);

		insertRectAbove.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalAbove(Signal.RECT);
			}
		});

		insertSinAbove.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalAbove(Signal.SIN);
			}
		});

		insertTriAbove.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalAbove(Signal.TRI);
			}
		});

		insertRectBelow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalBelow(Signal.RECT);
			}
		});

		insertSinBelow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalBelow(Signal.SIN);
			}
		});

		insertTriBelow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalBelow(Signal.TRI);
			}
		});

		removeSignal.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.removeSignal();
			}
		});

		this.table.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					JTable source = (JTable) e.getSource();
					int row = source.rowAtPoint(e.getPoint());
					int column = source.columnAtPoint(e.getPoint());

					if (!source.isRowSelected(row)) {
						source.changeSelection(row, column, false, false);
					}
					popupMenu.show(e.getComponent(), e.getX(), e.getY());
				}
			}
		});

		return popupMenu;
	}

	/**
	 * Create the toolBar for the TPeriodicSignalBuilder.
	 * 
	 * @return toolBar
	 */
	private JToolBar createToolBar() {

		JToolBar tb = new JToolBar();
		tb.setPreferredSize(new Dimension(500, 50));

		Vector<JButton> buttons = new Vector<JButton>();

		final JButton insertSignalAbove = new JButton("Above", IconEditor.ICON_ADD);
		insertSignalAbove.setToolTipText("Insert a row ");
		insertSignalAbove.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalAbove(Signal.RECT);
			}
		});
		buttons.add(insertSignalAbove);

		final JButton insertSignalBelow = new JButton("Below", IconEditor.ICON_ADD);
		insertSignalBelow.setToolTipText("Insert a row ");
		insertSignalBelow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.insertSignalBelow(Signal.RECT);
			}
		});
		buttons.add(insertSignalBelow);

		final JButton removeRow = new JButton("Remove", IconEditor.ICON_REMOVE);
		removeRow.setToolTipText("Delete a row ");
		removeRow.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.removeSignal();
			}
		});
		buttons.add(removeRow);

		final JButton showTime = new JButton("Time", IconEditor.ICON_XYPLOT);
		showTime.setToolTipText("Show time plot");
		showTime.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.showTime();
			}
		});
		buttons.add(showTime);

		final JButton showSpectrum = new JButton("Spectrum", IconEditor.ICON_SPECTRUM);
		showSpectrum.setToolTipText("Show spectrum plot");
		showSpectrum.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.showSpectrum();
			}
		});
		buttons.add(showSpectrum);

		final JButton showAll = new JButton("All", IconEditor.ICON_XYPLOT);
		showAll.setToolTipText("Show all signals plot");
		showAll.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.showAll();
			}
		});
		buttons.add(showAll);

		for (final JButton i : buttons) {

			Dimension d = new Dimension(55, 50);
			i.setVerticalTextPosition(SwingConstants.BOTTOM);
			i.setHorizontalTextPosition(SwingConstants.CENTER);
			i.setMaximumSize(d);
			i.setMinimumSize(d);
			i.setPreferredSize(d);
			tb.add(i);
		}

		return tb;
	}

	/**
	 * Set up the close panel with the buttons (Apply, Cancel).
	 */
	private void setUpClosePanel() {
		JPanel closePanel = new JPanel();
		closePanel.setLayout(new FlowLayout());

		JButton apply = new JButton("Apply");
		JButton cancel = new JButton("Cancel");

		apply.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this.dispose();
			}
		});

		cancel.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TPeriodicSignalBuilder.this
						.dispatchEvent(new WindowEvent(TPeriodicSignalBuilder.this, WindowEvent.WINDOW_CLOSING));
			}

		});

		closePanel.add(apply);
		closePanel.add(cancel);

		this.add(closePanel, BorderLayout.SOUTH);
	}
}
