/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1022 $
 * $Date: 2020-06-03 14:13:51 +0200 (Wed, 03 Jun 2020) $
 */
package at.lcm.bu21.general.dtypes.unit;

import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;

import org.json.simple.JSONArray;
import org.json.simple.parser.JSONParser;

/**
 * Unit representation
 *
 * @author Werner Koppelsttter
 */
public class Unit {

	/** List of base-units */
	public static List<Unit> baseUnits;

	/** Unit-string */
	String unitString;

	/** Exponent of unit */
	int exponent = 1;

	/** Gain of unit */
	double gain = 1;

	/** Offset of unit */
	double offset = 0;

	/** Base-unit for unit-string. Is null if this is an base-unit */
	Unit baseUnit = null;

	/** List of SI-units for base-units. Is null if this is no base-unit */
	List<Unit> siUnits = null;

	static {
		FileReader reader;
		JSONParser parser;
		JSONArray ja, unit;
		Unit u;
		File unitFile, tmp;

		Unit.baseUnits = new ArrayList<Unit>();

		try {
			Unit.baseUnits.add(new Unit("?", 1, true));

			unitFile = new File(
					new File(new File(Unit.class.getProtectionDomain().getCodeSource().getLocation().toURI())
							.getParentFile().getParentFile(), "Unit"),
					"Unit.json");
			if (!unitFile.exists()) {
				tmp = new File(new File(new File(Unit.class.getProtectionDomain().getCodeSource().getLocation().toURI())
						.getParentFile().getParentFile().getParentFile().getParentFile(), "Unit"), "Unit.json");
				if (tmp.exists()) {
					unitFile = tmp;
				}
			}

			reader = new FileReader(unitFile);

			parser = new JSONParser();

			ja = (JSONArray) parser.parse(reader);
			for (Object obj : ja) {
				unit = (JSONArray) obj;
				if (unit.get(0) != null) {
					u = new Unit((String) unit.get(0), 1, true);
					if (!unit.get(0).equals(unit.get(1))) {
						try {
							u.siUnits = new Parser(unit.get(1).toString()).getSyntax().getUnits(true);
						} catch (Exception e) {
							System.err.println("Error by parsing base-units: \"" + unit.get(1).toString() + "\" ("
									+ e.getMessage() + ")");
							e.printStackTrace();
						} // SI units

						u.gain = (double) unit.get(3); // gain
						u.offset = (double) unit.get(4); // offset
					}
					Unit.baseUnits.add(u);
				}
			}
		} catch (Exception e) {
			System.err.println("Error by reading base-units: " + e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Create unit from string
	 * 
	 * @param unitString
	 *            Unit-string
	 * @param exponent
	 *            Exponent of unit
	 * @throws SyntaxException
	 */
	public Unit(String unitString, int exponent) throws SyntaxException {
		this(unitString, exponent, false);
	}

	/**
	 * Create unit from string
	 * 
	 * @param unitString
	 *            Unit-string
	 * @param exponent
	 *            Exponent of unit
	 * @param baseUnit
	 *            Is base-unit
	 * @throws SyntaxException
	 */
	Unit(String unitString, int exponent, boolean baseUnit) throws SyntaxException {
		int index;

		this.unitString = unitString;
		this.exponent = exponent;

		index = Unit.baseUnits.indexOf(this);
		if (index != -1) {
			this.baseUnit = Unit.baseUnits.get(index);
		} else if (!baseUnit) {
			for (Prefix p : Prefix.values()) {
				if ((unitString.length() > p.getSymbol().length()) && unitString.startsWith(p.getSymbol())) {
					index = Unit.baseUnits
							.indexOf(new Unit(this.unitString.substring(p.getSymbol().length()), 0, baseUnit));
					if (index != -1) {
						this.baseUnit = Unit.baseUnits.get(index);
						this.unitString = this.baseUnit.unitString;
						this.gain = p.getGain();

						break;
					}
				}
			}
		}

		if (!baseUnit && (this.unitString != "1") && (this.baseUnit == null)) {
			throw new SyntaxException("Undefined unit: " + this.unitString);
		}
	}

	/**
	 * Get unit-string of unit
	 * 
	 * @return unit-string
	 */
	public String getUnitString() {
		return this.unitString;
	}

	/**
	 * Get exponent of unit
	 * 
	 * @return exponent
	 */
	public int getExponent() {
		return this.exponent;
	}

	/**
	 * Get gain of unit
	 * 
	 * @return gain
	 */
	public double getGain() {
		return this.gain;
	}

	/**
	 * Get offset of unit
	 * 
	 * @return offset
	 */
	public double getOffset() {
		return this.offset;
	}

	@Override
	public String toString() {
		return this.unitString + "\t^ " + this.exponent + "\t* " + this.gain + "\t+ " + this.offset;
	}

	/**
	 * Equal of unit-string
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Unit other = (Unit) obj;
		if (this.unitString == null) {
			if (other.unitString != null) {
				return false;
			}
		} else if (!this.unitString.equals(other.unitString)) {
			return false;
		}
		return true;
	}

}
