/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.math.interpolator2d;

import javax.swing.JFrame;
import javax.swing.WindowConstants;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.plot.HOLD;
import at.lcm.bu21.general.plot.Plot3D;
import at.lcm.bu21.general.plot.Plot3Surface;
import at.lcm.bu21.general.plot.Style;

/**
 * 
 * 
 * @author Gnther Weidenholzer
 */

public abstract class Interpolator2d {

	protected double[] xnorm;
	protected double[] ynorm;
	protected double[] zdata;

	protected double xmax;
	protected double xmin;
	protected double ymax;
	protected double ymin;
	protected double zmax;
	protected double zmin;

	protected double xrange;
	protected double yrange;
	protected double zrange;

	double rel_tol = 5e-2;
	private boolean checkHull = true;

	/**
	 * min x value
	 * 
	 * @return min x value
	 */
	public double getXmin() {
		return this.xmin;
	}

	/**
	 * max x value
	 * 
	 * @return max x value
	 */
	public double getXmax() {
		return this.xmax;
	}

	/**
	 * min y value
	 * 
	 * @return min y value
	 */
	public double getYmin() {
		return this.ymin;
	}

	/**
	 * max y value
	 * 
	 * @return max y value
	 */
	public double getYmax() {
		return this.ymax;
	}

	/**
	 * 
	 * @param x
	 * @param y
	 * @param z
	 * @return vectors
	 * @throws Exception
	 */
	public static double[][] dataVectors(TNumeric x, TNumeric y, TNumeric z)
			throws Exception {

		// linearize data
		int m = x.getNumRows();
		int n = x.getNumColumns();
		double[] xdata = new double[m * n];
		double[] ydata = new double[m * n];
		double[] zdata = new double[m * n];

		int idx = 0;
		for (int row = 0; row < m; row++) {
			for (int col = 0; col < n; col++) {
				xdata[idx] = x.getReal(row, col);
				ydata[idx] = y.getReal(row, col);
				zdata[idx] = z.getReal(row, col);
				idx++;

			}
		}

		return new double[][] { xdata, ydata, zdata };

	}

	/**
	 * Normalize Data and remove points, if they are too close
	 * 
	 * @param xdata
	 * @param ydata
	 * @throws Exception
	 */
	protected Interpolator2d(double[] xdata, double[] ydata, double[] zdata)
			throws Exception {

		int npoints = xdata.length;
		boolean[] isNonUnique = new boolean[npoints];

		this.xmax = Double.NEGATIVE_INFINITY;
		this.xmin = Double.POSITIVE_INFINITY;
		this.ymax = Double.NEGATIVE_INFINITY;
		this.ymin = Double.POSITIVE_INFINITY;

		for (int k = 0; k < npoints; k++) {
			this.xmax = xdata[k] > this.xmax ? xdata[k] : this.xmax;
			this.xmin = xdata[k] < this.xmin ? xdata[k] : this.xmin;
			this.ymax = ydata[k] > this.ymax ? ydata[k] : this.ymax;
			this.ymin = ydata[k] < this.ymin ? ydata[k] : this.ymin;
			this.zmax = zdata[k] > this.zmax ? zdata[k] : this.zmax;
			this.zmin = zdata[k] < this.zmin ? zdata[k] : this.zmin;

			isNonUnique[k] = Double.isNaN(zdata[k]);
		}

		this.xrange = this.xmax - this.xmin;
		this.yrange = this.ymax - this.ymin;
		this.zrange = this.zmax - this.zmin;

		this.xnorm = new double[npoints];
		this.ynorm = new double[npoints];

		for (int s = 0; s < npoints; s++) {

			if (isNonUnique[s]) {
				continue;
			}

			double x1 = xdata[s] / this.xrange;
			double y1 = ydata[s] / this.yrange;

			for (int k = 0; k < s; k++) {

				if (isNonUnique[k]) {
					continue;
				}

				double x2 = xdata[k] / this.xrange;
				double y2 = ydata[k] / this.yrange;

				double dx = x1 - x2;
				double dy = y1 - y2;

				double r = Math.sqrt((dx * dx) + (dy * dy));

				if ((r < 1e-3)) {
					isNonUnique[s] = true;
				}

			}

			int cntUniquePoints = 0;
			for (int k = 0; k < npoints; k++) {
				if (!isNonUnique[k]) {
					cntUniquePoints++;
				}
			}

			this.xnorm = new double[cntUniquePoints];
			this.ynorm = new double[cntUniquePoints];
			this.zdata = new double[cntUniquePoints];

			int idxUnique = 0;
			for (int k = 0; k < npoints; k++) {
				if (!isNonUnique[k]) {

					this.xnorm[idxUnique] = xdata[k] / this.xrange;
					this.ynorm[idxUnique] = ydata[k] / this.yrange;
					this.zdata[idxUnique] = zdata[k];
					idxUnique++;
				}
			}
		}

	}

	/**
	 * @param x
	 * @param y
	 * @return interpolated value
	 */
	public double value(double x, double y) {

		double val = Double.NaN;

		try {
			if (!this.checkHull || this.contains(x, y)) {
				val = this.value_(x / this.xrange, y / this.yrange);
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;
	}

	abstract double value_(double xn, double yn);

	/**
	 * 
	 * @param x
	 * @param y
	 * @return first x-derivative
	 */
	public double dz_dx(double x, double y) {

		double val = Double.NaN;

		try {

			if (!this.checkHull || this.contains(x, y)) {
				val = this.dz_dx_(x / this.xrange, y / this.yrange)
						/ this.xrange;
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;

	}

	abstract double dz_dx_(double x, double y);

	/**
	 * 
	 * @param x
	 * @param y
	 * @return first y-derivative
	 */
	public double dz_dy(double x, double y) {

		double val = Double.NaN;

		try {

			if (!this.checkHull || this.contains(x, y)) {
				val = this.dz_dy_(x / this.xrange, y / this.yrange)
						/ this.yrange;
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;

	}

	abstract double dz_dy_(double x, double y);

	/**
	 * 
	 * @param x
	 * @param y
	 * @return second x-derivative
	 */
	public double dz_dx2(double x, double y) {

		double val = Double.NaN;

		try {

			if (!this.checkHull || this.contains(x, y)) {
				val = this.dz_dx2_(x / this.xrange, y / this.yrange)
						/ this.xrange / this.xrange;
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;

	}

	abstract double dz_dx2_(double x, double y);

	/**
	 * 
	 * @param x
	 * @param y
	 * @return second y-derivative
	 */
	public double dz_dy2(double x, double y) {

		double val = Double.NaN;

		try {

			if (!this.checkHull || this.contains(x, y)) {
				val = this.dz_dy2_(x / this.xrange, y / this.yrange)
						/ this.yrange / this.yrange;
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;

	}

	abstract double dz_dy2_(double x, double y);

	/**
	 * 
	 * @param x
	 * @param y
	 * @return mixed xy-derivative
	 */
	public double dz_dxy(double x, double y) {

		double val = Double.NaN;

		try {

			if (!this.checkHull || this.contains(x, y)) {
				val = this.dz_dxy_(x / this.xrange, y / this.yrange)
						/ this.xrange / this.yrange;
			}
		} catch (Exception e) {
			// e.printStackTrace();
		}

		return val;

	}

	abstract double dz_dxy_(double x, double y);

	/**
	 * @param x
	 * @param y
	 * @return true if inside hull, with tolerance!
	 */
	public boolean contains(double x, double y) {

		final int n = 5;

		for (int k = 0; k < n; k++) {
			double x_ = (x / this.xrange)
					+ (this.rel_tol * Math.cos((2.0 * Math.PI * k) / n));
			double y_ = (y / this.yrange)
					+ (this.rel_tol * Math.sin((2.0 * Math.PI * k) / n));
			if (this.contains_(x_, y_)) {
				return true;
			}
		}

		return false;

	}

	abstract boolean contains_(double x_, double y_);

	/**
	 * set true to enable extrapolation
	 * 
	 * @param allow
	 * @throws Exception
	 */
	public void allowExtrapolation(boolean allow) throws Exception {
		this.checkHull = !allow;
	}

	/**
	 * Preview of interpolated data
	 * 
	 * @return Panel with plot
	 * @throws Exception
	 */
	public Plot3D previewPlot() throws Exception {

		Plot3D plot;
		int n = 50;

		plot = new Plot3Surface();

		plot.addSignal("Points", new TNumeric(this.xnorm).scale(this.xrange),
				new TNumeric(this.ynorm).scale(this.yrange), new TNumeric(
						this.zdata));

		Style style = plot.getSignal(0).getStyle();
		style.setMarkerVisible(true);
		style.setLineVisible(false);
		style.setSurfaceVisible(false);

		double[] x = new double[n * n];
		double[] y = new double[n * n];
		double[] z = new double[n * n];

		int cnt = 0;
		for (int p = 0; p < n; p++) {
			for (int q = 0; q < n; q++) {

				double x1 = this.xmin
						+ ((p * (this.xmax - this.xmin)) / (n - 1));
				double y1 = this.ymin
						+ ((q * (this.ymax - this.ymin)) / (n - 1));
				double z1 = this.value(x1, y1);

				if (!Double.isNaN(z1)) {

					x[cnt] = x1;
					y[cnt] = y1;
					z[cnt] = z1;

					cnt++;

				}

			}
		}

		plot.setHold(HOLD.ON);
		plot.addSignal("Surf", new TNumeric(x), new TNumeric(y),
				new TNumeric(z));

		style = plot.getSignal(1).getStyle();
		style.setMarkerVisible(false);
		style.setLineVisible(true);
		style.setSurfaceVisible(true);

		plot.setXLabel("x");
		plot.setYLabel("y");
		plot.setZLabel("z");

		return plot;
	}

	/**
	 * 
	 * @throws Exception
	 */
	public void showPreview() throws Exception {

		JFrame f = new JFrame("2d interpolation preview");
		f.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		f.setContentPane(this.previewPlot().getPanel());
		f.setSize(500, 500);
		f.setVisible(true);
	}

}
