/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import org.jfree.chart.plot.XYPolarPlot;
import org.jfree.chart.renderer.xy.DefaultXYPolarItemRenderer;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Class for polar plotting
 * 
 * @author Werner Koppelsttter
 */

public class Plot2Polar extends Plot2D {

	/**
	 * Creats an empty Object
	 */
	public Plot2Polar() {
		super();
		this.plot.setRenderer(0, new DefaultXYPolarItemRenderer());
	}

	@Override
	protected void addSignal(Signal signal) {
		double[] phi, r;
		double x, y;

		phi = signal.getXRef();
		r = signal.getYRef();

		for (int k = 0; k < phi.length; k++) {
			x = r[k] * Math.cos(phi[k]);
			y = r[k] * Math.sin(phi[k]);
			phi[k] = x;
			r[k] = y;
		}

		// phi = signal.getXRef();
		// r = signal.getYRef();
		//
		// for (int k = 0; k < phi.length; k++) {
		// // convert from rad to grad
		// phi[k] = (((phi[k]) * -180.0) / Math.PI) + 90.0;
		//
		// // limit rang of radius 0...Inf
		// if (r[k] < 0.0) {
		// phi[k] += 180.0;
		// r[k] *= -1.0;
		// }
		// }

		super.addSignal(signal);
	}

	/**
	 * Adds data to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param data
	 *            Data to plot
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric data)
			throws PlotException, TDataException, CloneException {
		Signal sig;
		double[] r, phi;
		double re, im;

		sig = new Signal(name, new TNumeric(data.getRealMatrix()),
				new TNumeric(data.getImagMatrix()));

		phi = sig.getXRef();
		r = sig.getYRef();
		for (int k = 0; k < phi.length; k++) {
			re = phi[k];
			im = r[k];

			phi[k] = Math.atan2(im, re);
			r[k] = Math.hypot(re, im);
		}

		this.addSignal(sig);
	}

	/**
	 * Add data to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param phi
	 *            Angle
	 * @param r
	 *            Radius
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[] phi, double[] r)
			throws PlotException, TDataException, CloneException {
		this.addSignal(new Signal(name, phi, r));
	}

	/**
	 * Add data to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param phi
	 *            Angle
	 * @param r
	 *            Radius
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric phi, TNumeric r)
			throws PlotException, TDataException, CloneException {
		this.addSignal(new Signal(name, phi, r));
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param per
	 *            TPeriodic signal
	 * @throws PlotException
	 *             Plot exception
	 */
	public void addSignal(String name, TPeriodic per) throws PlotException {
		this.addSignal(name, per, 360);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param re
	 *            TPeriodic real-signal
	 * @param im
	 *            TPeriodic im-signal
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TPeriodic re, TPeriodic im)
			throws PlotException, TDataException, CloneException {
		this.addSignal(name, re, im, 360);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param re
	 *            TPeriodic real-signal
	 * @param im
	 *            TPeriodic im-signal
	 * @param nPoints
	 *            Number of point to plot
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TPeriodic re, TPeriodic im, int nPoints)
			throws PlotException, TDataException, CloneException {
		double[] r, phi;
		double re2, im2, phi2;

		r = new double[nPoints + 1];
		phi = new double[nPoints + 1];
		for (int k = 0; k <= nPoints; k++) {
			phi2 = ((double) k / nPoints) * 2 * Math.PI;
			re2 = re.valueAt(phi2);
			im2 = im.valueAt(phi2);
			r[k] = Math.hypot(re2, im2);
			phi[k] = Math.atan2(im2, re2);
		}

		this.addSignal(name, phi, r);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param per
	 *            TPeriodic signal
	 * @param nPoints
	 *            Number of point to plot
	 * @throws PlotException
	 *             Plot exception
	 */
	public void addSignal(String name, TPeriodic per, int nPoints)
			throws PlotException {
		this.addSignal(new Signal(name, per, nPoints));
	}

	/**
	 * Updates the boundaries
	 */
	@Override
	void updateBounds() {
		this.plot.getDomainAxis().setLowerBound(this.xBoundMin * 1.15);
		this.plot.getDomainAxis().setUpperBound(this.xBoundMax * 1.15);

		this.plot.getRangeAxis().setLowerBound(this.yBoundMin * 1.15);
		this.plot.getRangeAxis().setUpperBound(this.yBoundMax * 1.15);

		this.getPlotPanel().repaint();
	}

	/**
	 * This Methode zooms the whole plot to the right perspective
	 */
	@Override
	public void zoom() {
		double limit = 0.0;

		limit = ((XYPolarPlot) this.plot).getMaxUsedRadius();

		// limitsX = MathUtils.roundLimits(limitsX);
		// limitsY = MathUtils.roundLimits(limitsY);

		if ((this.xLimitSet == false)) {
			this.xLimitMin = -limit;
			this.xLimitMax = limit;
		}
		this.xBoundMin = this.xLimitMin;
		this.xBoundMax = this.xLimitMax;

		if ((this.yLimitSet == false)) {
			this.yLimitMin = -limit;
			this.yLimitMax = limit;
		}
		this.yBoundMin = this.yLimitMin;
		this.yBoundMax = this.yLimitMax;

		this.updateBounds();
	}
}
