/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JLabel;
import javax.swing.SwingConstants;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import org.jzy3d.chart.Chart;
import org.jzy3d.chart.controllers.keyboard.camera.NewtCameraKeyController;
import org.jzy3d.chart.controllers.mouse.camera.NewtCameraMouseController;
import org.jzy3d.chart.factories.NewtChartComponentFactory;
import org.jzy3d.colors.ColorMapper;
import org.jzy3d.colors.IColorMappable;
import org.jzy3d.colors.colormaps.IColorMap;
import org.jzy3d.maths.Coord3d;
import org.jzy3d.plot3d.builder.Builder;
import org.jzy3d.plot3d.primitives.Point;
import org.jzy3d.plot3d.primitives.Shape;
import org.jzy3d.plot3d.primitives.axes.layout.renderers.DefaultDecimalTickRenderer;
import org.jzy3d.plot3d.rendering.canvas.CanvasNewtAwt;
import org.jzy3d.plot3d.rendering.canvas.Quality;
import org.jzy3d.plot3d.rendering.legends.colorbars.AWTColorbarLegend;

import com.jogamp.newt.event.MouseAdapter;
import com.jogamp.newt.event.MouseEvent;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Class for plotting 3D signals
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public abstract class Plot3D extends Plot {

	/**
	 * View mode for plot
	 *
	 * @author Werner Koppelsttter
	 */
	public enum ViewPositionMode {
	/** Free position */
	FREE,
	/** Profile */
	PROFILE,
	/** Top position */
	TOP;
	}

	/**
	 * The ColorMap is a prograssive transition from blue over green to red. The
	 * mix between these 3 colors is expressed by the definition of 2 functions.
	 * 
	 * <code><br>
	 *           blue   green  red         <br>
	 *  1.0 |   /------X------X------\   | <br>
	 *      |  /:     /:\    /:\     :\  | <br>
	 *      | / :    / : \  / : \    : \ | <br>
	 *  0.5 |/  :   /  :  \/  :  \   :  \| <br>
	 *      |   :  /   :  /\  :   \  :   | <br>
	 *      |   : /    : /  \ :    \ :   | <br>
	 *      |   :/     :/    \:     \:   | <br>
	 *  0.0 |-------------|--------------| <br>
	 *      0   :      : 0.5  :      :   1 <br>
	 *         L1     L2     L3     L4     <br>
	 * <br></code>
	 * 
	 * The input is is from min- to max-value of colorable-parameter
	 * 
	 * @author Werner Koppelsttter
	 */
	private class ColorMap implements IColorMap {

		/** Is it is the standard color gradient */
		private boolean direction = true;

		/** Level where blue reach the maximum and green starts */
		private double level1 = 0.125f; // 1/8

		/**
		 * Level where blue ends the maximum, green reach the maximum and red
		 * starts
		 */
		private double level2 = 0.375f; // 3/8

		/**
		 * Level where blue ends to minimum, green ends the maximum and red
		 * reach the maximum starts
		 */
		private double level3 = 0.625f; // 5/8

		/** Level where green ends to minimum, red ends the maximum */
		private double level4 = 0.875f; // 7/8

		@Override
		public org.jzy3d.colors.Color getColor(IColorMappable colorable,
				double x, double y, double z) {
			return this.getColor(colorable, z);
		}

		@Override
		public org.jzy3d.colors.Color getColor(IColorMappable colorable,
				double val) {
			double r, g, b;
			org.jzy3d.colors.Color color = org.jzy3d.colors.Color.WHITE;

			if (!Double.isNaN(val)) {
				if ((val < colorable.getMin()) || (colorable.getMax() < val)) {
					color = org.jzy3d.colors.Color.MAGENTA;
				} else {
					val = (val - colorable.getMin())
							/ (colorable.getMax() - colorable.getMin());

					r = this.colorComponentAbsolute(val, this.level2,
							2.0 - this.level4, this.level3, this.level4);
					g = this.colorComponentAbsolute(val, this.level1,
							this.level4, this.level2, this.level3);
					b = this.colorComponentAbsolute(val, -this.level1,
							this.level3, this.level1, this.level2);

					if (this.direction) {
						color = new org.jzy3d.colors.Color((float) r, (float) g,
								(float) b);
					} else {
						color = new org.jzy3d.colors.Color((float) b, (float) g,
								(float) r);
					}
				}
			}
			return color;
		}

		@Override
		public boolean getDirection() {
			return this.direction;
		}

		@Override
		public void setDirection(boolean isStandard) {
			this.direction = isStandard;
		}

		// /**
		// * List color-levels
		// *
		// * @return List of color-levels
		// */
		// public float[] getColorLevels() {
		// return new float[] { this.level1, this.level2, this.level3,
		// this.level4 };
		// }

		/**
		 * Set color-levels
		 * 
		 * @param levels
		 *            List of color-levels
		 */
		public void setColorLevels(float[] levels) {
			if ((levels != null) && (levels.length == 4)) {
				this.setColorLevels(levels[0], levels[1], levels[2], levels[3]);
			}
		}

		/**
		 * Set color-levels
		 * 
		 * @param level1
		 *            First color-level
		 * @param level2
		 *            Second color-level
		 * @param level3
		 *            Third color-level
		 * @param level4
		 *            Fourth color-level
		 */
		public void setColorLevels(double level1, double level2, double level3,
				double level4) {
			this.level1 = level1;
			this.level2 = level2;
			this.level3 = level3;
			this.level4 = level4;
		}

		@Override
		public double colorComponentAbsolute(double value, double bLeft,
				double bRight, double tLeft, double tRight) {
			double col;

			if ((value <= bLeft) || (bRight <= value)) {
				col = 0.0f;
			} else if ((tLeft <= value) && (value <= tRight)) {
				col = 1.0f;
			} else if ((bLeft < value) && (value < tLeft)) {
				col = (value - bLeft) / (tLeft - bLeft);
			} else if ((tRight < value) && (value < bRight)) {
				col = (value - bRight) / (tRight - bRight);
			} else {
				col = 0.0f;
			}

			return col;
		}

		@Override
		public double colorComponentRelative(double value, double center,
				double topWidth, double bottomWidth) {
			return this.colorComponentAbsolute(value,
					center - (bottomWidth / 2.0),
					(center + (bottomWidth / 2.0)), center - (topWidth / 2.0),
					center + (topWidth / 2.0));
		}

	}

	/** 3D Chart */
	private Chart chart;

	/** List of displayed shapes */
	private LinkedList<Shape> shapes = new LinkedList<Shape>();

	/** Label for title */
	private JLabel titleLabel;

	/** Show color legend */
	private boolean showColorLegend = false;

	/**
	 * Get shape to values
	 * 
	 * @param style
	 *            Style of values
	 * @param values
	 *            Values to plot
	 * @return Shape for values
	 */
	private Shape getShape(Style style, double[][] values) {
		Shape shape;
		Point point;
		Color c;

		shape = new Shape();
		if (style.isMarkerVisible()) {
			for (int k = 0; k < values[0].length; k++) {
				point = new org.jzy3d.plot3d.primitives.Point(
						new Coord3d(values[0][k], values[1][k], values[2][k]));
				c = style.getColor();
				point.setColor(new org.jzy3d.colors.Color(c.getRed(),
						c.getGreen(), c.getBlue()));
				point.setWidth(style.getMarkerSize());
				shape.add(point);
			}
		}

		return shape;
	}

	/**
	 * Get shape for surface and wireframe
	 * 
	 * @param signal
	 *            Signal to plot
	 * @return Shape for plot
	 */
	private Shape getSurfaceShape(Signal signal) {
		List<Coord3d> coords;
		double[] x, y, z;
		ColorMapper color;
		Style style;
		Shape shape;
		Color c;
		ColorMap colorMap;

		shape = new Shape();
		style = signal.getStyle();
		if (style.isLineVisible() || style.isSurfaceVisible()) {
			x = signal.getX();
			y = signal.getY();
			z = signal.getZ();
			coords = new ArrayList<Coord3d>(x.length);
			for (int k = 0; k < x.length; k++) {
				coords.add(new Coord3d(x[k], y[k], z[k]));
			}
			shape = Builder.buildDelaunay(coords);

			colorMap = new ColorMap();
			colorMap.setDirection(this.getColorDirection());
			colorMap.setColorLevels(this.getColorLevels());

			// surface and Legend
			color = new ColorMapper(colorMap, shape.getBounds().getZmin(),
					shape.getBounds().getZmax());
			shape.setColorMapper(color);
			shape.setFaceDisplayed(style.isSurfaceVisible());
			if (style.isSurfaceVisible() && this.getColorLegend()) {
				shape.setLegend(new AWTColorbarLegend(shape,
						this.chart.getView().getAxe().getLayout(),
						org.jzy3d.colors.Color.BLACK, null));
				shape.setLegendDisplayed(true);
			}

			// wireframe
			shape.setWireframeWidth(style.getLineWidth());
			c = style.getColor();
			shape.setWireframeColor(new org.jzy3d.colors.Color(c.getRed(),
					c.getGreen(), c.getBlue()));
			shape.setWireframeDisplayed(style.isLineVisible());
		}

		return shape;
	}

	/**
	 * Creates an empty object
	 */
	public Plot3D() {
		super();

		// chart
		// this.chart = AWTChartComponentFactory.chart(Quality.Intermediate);
		// this.chart.addController(new AWTCameraKeyController());
		// this.chart.addController(new AWTCameraMouseController());
		this.chart = NewtChartComponentFactory.chart(Quality.Intermediate);
		this.chart.addController(new NewtCameraKeyController());
		this.chart.addController(new NewtCameraMouseController());

		// axes values
		this.chart.getAxeLayout()
				.setXTickRenderer(new DefaultDecimalTickRenderer(2));
		this.chart.getAxeLayout()
				.setYTickRenderer(new DefaultDecimalTickRenderer(2));
		this.chart.getAxeLayout()
				.setZTickRenderer(new DefaultDecimalTickRenderer(2));

		// title
		this.titleLabel = new JLabel();
		this.titleLabel.setHorizontalAlignment(SwingConstants.CENTER);
		this.titleLabel.setAutoscrolls(true);
		this.titleLabel.setFont(new Font("Helvetica", Font.PLAIN, 20));
		this.titleLabel.setBackground(Color.WHITE);
		this.titleLabel.setOpaque(true);

		// add Panel
		this.getPlotPanel().add((Component) this.chart.getCanvas(),
				BorderLayout.CENTER);
		this.setViewPoint((Math.PI * 5) / 4, Math.PI / 8, 0.0);
		this.redraw();

		// add context menu
		((CanvasNewtAwt) this.chart.getCanvas())
				.addMouseListener(new MouseAdapter() {
					@Override
					public void mouseClicked(MouseEvent e) {
						if (e.getButton() == com.jogamp.newt.event.MouseEvent.BUTTON3) {
							Plot3D.this.popup.getPopupMenu().show(
									(Component) Plot3D.this.chart.getCanvas(),
									e.getX(), e.getY());
						}
					}
				});
		// ((CanvasAWT) this.chart.getCanvas())
		// .addMouseListener(new java.awt.event.MouseAdapter() {
		// @Override
		// public void mouseClicked(MouseEvent e) {
		// if (e.getButton() == com.jogamp.newt.event.MouseEvent.BUTTON3) {
		// Plot3D.this.popup.getPopupMenu().show(
		// (Component) Plot3D.this.chart.getCanvas(),
		// e.getX(), e.getY());
		// }
		// }
		// });

		this.getPlotPanel().addAncestorListener(new AncestorListener() {
			@Override
			public void ancestorRemoved(AncestorEvent event) {
				((CanvasNewtAwt) Plot3D.this.chart.getCanvas()).getAnimator()
						.stop();
			}

			@Override
			public void ancestorMoved(AncestorEvent event) {
			}

			@Override
			public void ancestorAdded(AncestorEvent event) {
				((CanvasNewtAwt) Plot3D.this.chart.getCanvas()).getAnimator()
						.start();
			}
		});
	}

	/**
	 * o Adds a 3D Signal
	 * 
	 * @param name
	 *            Name of plot
	 * @param xSignal
	 *            X values
	 * @param ySignal
	 *            Y values
	 * @param zSignal
	 *            Z values
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[] xSignal, double[] ySignal,
			double[] zSignal) throws PlotException, CloneException {
		this.addSignal(new Signal(name, xSignal, ySignal, zSignal));
		this.redraw();
	}

	/**
	 * o Adds a 3D Signal
	 * 
	 * @param name
	 *            Name of plot
	 * @param zSignal
	 *            Z values
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[][] zSignal)
			throws PlotException, CloneException {
		double[] xSignal, ySignal;

		ySignal = new double[zSignal.length];
		for (int k = 0; k < ySignal.length; k++) {
			ySignal[k] = k + 1;
		}

		xSignal = new double[zSignal[0].length];
		for (int k = 0; k < xSignal.length; k++) {
			xSignal[k] = k + 1;
		}

		this.addSignal(new Signal(name, xSignal, ySignal, zSignal));
		this.redraw();
	}

	/**
	 * o Adds a 3D Signal
	 * 
	 * @param name
	 *            Name of plot
	 * @param xSignal
	 *            X values
	 * @param ySignal
	 *            Y values
	 * @param zSignal
	 *            Z values
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[] xSignal, double[] ySignal,
			double[][] zSignal) throws PlotException, CloneException {
		this.addSignal(new Signal(name, xSignal, ySignal, zSignal));
		this.redraw();
	}

	/**
	 * o Adds a 3D Signal
	 * 
	 * @param name
	 *            Name of plot
	 * @param zSignal
	 *            Z values
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric zSignal)
			throws PlotException, CloneException, TDataException {
		double[] xSignal, ySignal;

		ySignal = new double[zSignal.size()[0]];
		for (int k = 0; k < ySignal.length; k++) {
			ySignal[k] = k + 1;
		}

		xSignal = new double[zSignal.size()[1]];
		for (int k = 0; k < xSignal.length; k++) {
			xSignal[k] = k + 1;
		}

		this.addSignal(new Signal(name, new TNumeric(xSignal),
				new TNumeric(ySignal), zSignal));
	}

	/**
	 * o Adds a 3D Signal
	 * 
	 * @param name
	 *            Name of plot
	 * @param xSignal
	 *            X values
	 * @param ySignal
	 *            Y values
	 * @param zSignal
	 *            Z values
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric xSignal, TNumeric ySignal,
			TNumeric zSignal)
			throws PlotException, CloneException, TDataException {
		this.addSignal(new Signal(name, xSignal, ySignal, zSignal));
	}

	/**
	 * Get the view-Point
	 * 
	 * @return Coordinates of view-point
	 */
	public double[] getViewPoint() {
		double[] point;
		Coord3d coord;

		coord = this.chart.getViewPoint();
		point = new double[3];
		point[0] = coord.x;
		point[1] = coord.y;
		point[2] = coord.z;

		return point;
	}

	/**
	 * Set the view-point
	 * 
	 * @param x
	 *            X angle (rad)
	 * @param y
	 *            Y angle (rad)
	 * @param z
	 *            Z
	 */
	public void setViewPoint(double x, double y, double z) {
		this.chart.setViewPoint(new Coord3d(x, y, z));
	}

	/**
	 * Get the view position mode
	 * 
	 * @return ViewPositionMode
	 */
	public ViewPositionMode getViewPositionMode() {
		ViewPositionMode mode;
		switch (this.chart.getViewMode()) {
		case FREE:
			mode = ViewPositionMode.FREE;
			break;
		case PROFILE:
			mode = ViewPositionMode.PROFILE;
			break;
		case TOP:
			mode = ViewPositionMode.TOP;
			break;
		default:
			mode = null;
			break;
		}
		return mode;
	}

	/**
	 * Set the view position mode
	 * 
	 * @param vpm
	 *            ViewPositionMode
	 */
	public void setViewPositionMode(ViewPositionMode vpm) {
		switch (vpm) {
		case FREE:
			this.chart.setViewMode(
					org.jzy3d.plot3d.rendering.view.modes.ViewPositionMode.FREE);
			break;
		case PROFILE:
			this.chart.setViewMode(
					org.jzy3d.plot3d.rendering.view.modes.ViewPositionMode.PROFILE);
			break;
		case TOP:
			this.chart.setViewMode(
					org.jzy3d.plot3d.rendering.view.modes.ViewPositionMode.TOP);
			break;
		default:
			break;
		}
	}

	@Override
	void repaint() {
		if (this.chart != null) {
			this.chart.render();
		}
		this.getPlotPanel().repaint();
		this.getPlotPanel().revalidate();
	}

	@Override
	void redraw() {
		Signal sig;
		Shape shape;

		// title
		if (this.getTitle() == null) {
			this.getPlotPanel().remove(this.titleLabel);
		} else {
			this.titleLabel.setText(this.getTitle());
			this.getPlotPanel().add(this.titleLabel, BorderLayout.NORTH);
		}
		this.getPlotPanel().validate();

		for (Shape s : this.shapes) {
			this.chart.getScene().getGraph().remove(s);
		}
		this.shapes.clear();

		for (int k = 0; k < this.getNumSignals(); k++) {
			sig = this.getSignal(k);

			// surface
			shape = this.getSurfaceShape(sig);
			if (shape.size() > 0) {
				this.shapes.add(shape);
				this.chart.getScene().getGraph().add(shape);
			}

			// values
			shape = this.getShape(sig.getStyle(), sig.getValues());
			if (shape.size() > 0) {
				this.shapes.add(shape);
				this.chart.getScene().getGraph().add(shape);
			}

			// selected values
			shape = this.getShape(sig.getSelectStyle(),
					sig.getSelectedValues());
			if (shape.size() > 0) {
				this.shapes.add(shape);
				this.chart.getScene().getGraph().add(shape);
			}
		}

		// labels
		if (this.getXLabel() != null) {
			this.chart.getAxeLayout().setXAxeLabel(this.getXLabel());
		} else {
			this.chart.getAxeLayout().setXAxeLabel("x");
		}
		if (this.getYLabel() != null) {
			this.chart.getAxeLayout().setYAxeLabel(this.getYLabel());
		} else {
			this.chart.getAxeLayout().setYAxeLabel("y");
		}
		if (this.getZLabel() != null) {
			this.chart.getAxeLayout().setZAxeLabel(this.getZLabel());
		} else {
			this.chart.getAxeLayout().setZAxeLabel("z");
		}

		// view
		// this.chart.getView().getCamera().setStretchToFill(true);
		this.chart.getView().setMaximized(true);
		this.chart.render();

	}

	@Override
	public void reset() {
		this.redraw();
	}

	/**
	 * Get 3d-chart of jzy3d
	 * 
	 * @return 3D-chart
	 */
	public Chart getChart() {
		return this.chart;
	}

	/**
	 * Returns the color legend visibility
	 * 
	 * @return Is visible
	 */
	public boolean getColorLegend() {
		return this.showColorLegend;
	}

	/**
	 * Set color legend visibility
	 * 
	 * @param show
	 *            Visibility of legend
	 */
	public void setColorLegend(boolean show) {
		this.showColorLegend = show;
		this.redraw();
	}

	// @Override
	// protected BufferedImage getBufferedImage() {
	// BufferedImage im;
	// Component image;
	// JPanel plot;
	//
	// plot = (JPanel) this.chart.getCanvas();
	//
	// // chart image
	// im = this.chart.screenshot();
	//
	// // display screenshot
	// image = new JImagePanel(im);
	// image.setSize(plot.getSize());
	// this.getPlotPanel().remove(plot);
	// this.getPlotPanel().add(image, BorderLayout.CENTER);
	//
	// // get new image
	// im = super.getBufferedImage();
	//
	// // restore structure
	// this.getPlotPanel().remove(image);
	// this.getPlotPanel().add((Component) this.chart.getCanvas(),
	// BorderLayout.CENTER);
	// // this.getPlotPanel().validate();
	//
	// plot.paint(plot.getGraphics());
	//
	// return im;
	// }

	@Override
	protected boolean isVectorGraphicPossible() {
		return false;
	}

}
