/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot.interpolation;

/**
 * Bilinear interpolation of values of a grid. Requiredment is increasing order
 * of x- and y-vectors. Values outside the original-values will be NaN
 * 
 * @author Werner Koppelsttter
 */
public class BilinearInterpolator {

	/** List of x-values */
	double[] xValues;

	/** List of y-values */
	double[] yValues;

	/** Matrix of z-values. First dimenstion is y and the second is x */
	double[][] zValues;

	/**
	 * Create a bilinear interpolator.
	 * 
	 * @param xValues
	 *            List of x-values (increasing order)
	 * @param yValues
	 *            List of y-values (increasing order)
	 * @param zValues
	 *            Matrix of z-values. (First dimenstion is y and the second is
	 *            x)
	 * @throws Exception
	 */
	public BilinearInterpolator(double[] xValues, double[] yValues,
			double[][] zValues) throws Exception {

		if ((yValues.length != zValues.length)
				|| (xValues.length != zValues[0].length)) {
			throw new Exception("invalid size");
		}
		this.xValues = xValues;
		this.yValues = yValues;
		this.zValues = zValues;
	}

	/**
	 * Interplolate one point for x,y values. Values outside the original-values
	 * will be NaN
	 * 
	 * @param xValue
	 *            x-value
	 * @param yValue
	 *            y-value
	 * @return Interpolated z-value
	 */
	public double interpolate(double xValue, double yValue) {
		int x, y;
		double r1, r2, z = Double.NaN;

		if ((this.xValues[0] <= xValue)
				&& (xValue <= this.xValues[this.xValues.length - 1])
				&& (this.yValues[0] <= yValue)
				&& (yValue <= this.yValues[this.yValues.length - 1])) {

			x = 1;
			while ((x < (this.xValues.length - 1))
					&& (xValue > this.xValues[x])) {
				x++;
			}

			y = 1;
			while ((y < (this.yValues.length - 1))
					&& (yValue > this.yValues[y])) {
				y++;
			}

			r1 = (((this.xValues[x] - xValue) / (this.xValues[x] - this.xValues[x - 1])) * this.zValues[y - 1][x - 1]);
			r1 += (((xValue - this.xValues[x - 1]) / (this.xValues[x] - this.xValues[x - 1])) * this.zValues[y - 1][x]);
			r2 = (((this.xValues[x] - xValue) / (this.xValues[x] - this.xValues[x - 1])) * this.zValues[y][x - 1]);
			r2 += (((xValue - this.xValues[x - 1]) / (this.xValues[x] - this.xValues[x - 1])) * this.zValues[y][x]);

			z = (((this.yValues[y] - yValue) / (this.yValues[y] - this.yValues[y - 1])) * r1);
			z += (((yValue - this.yValues[y - 1]) / (this.yValues[y] - this.yValues[y - 1])) * r2);
		}

		return z;
	}

	/**
	 * Interplolate a matrix of x,y values. Values outside the original-values
	 * will be NaN
	 * 
	 * @param xValues
	 *            x-values (increasing order)
	 * @param yValues
	 *            y-values (increasing order)
	 * @return Matrix of interpolated z-value (First dimenstion is y and the
	 *         second is x)
	 */
	public double[][] interpolate(double[] xValues, double[] yValues) {
		double[][] z;

		z = new double[yValues.length][xValues.length];

		for (int y = 0; y < yValues.length; y++) {
			for (int x = 0; x < xValues.length; x++) {
				z[y][x] = this.interpolate(xValues[x], yValues[y]);
			}
		}

		return z;
	}
}
