/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1028 $
 * $Date: 2020-07-24 15:03:48 +0200 (Fri, 24 Jul 2020) $
 */
package at.lcm.bu21.general.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.nio.channels.FileChannel;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.LinkedList;

/**
 * Additional File utilities
 * 
 * @author Werner Koppelsttter
 */
public class FileUtils {

	/**
	 * Max count of transfered byte at one cycle
	 */
	// magic number for Windows
	private final static int MAX_COUNT = (64 * 1024 * 1024) - (23 * 1024);

	/**
	 * Copy file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @throws IOException File exception
	 */
	public static void copy(String source, String destination) throws IOException {
		FileUtils.copy(new File(source), new File(destination));
	}

	/**
	 * Copy file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @throws IOException File exception
	 */
	public static void copy(File source, File destination) throws IOException {
		FileUtils.copy(source, destination, false);
	}

	/**
	 * Copy file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @param force Force to override destination
	 * @throws IOException File exception
	 */
	public static void copy(File source, File destination, boolean force) throws IOException {
		String[] files;
		FileChannel inChannel, outChannel;
		long pos, size;

		if (!source.exists()) {
			throw new IOException("Source does not exist (" + source.getAbsolutePath() + ")");
		}
		if (!force && destination.exists()) {
			throw new IOException("Destination does already exist (" + destination.getAbsolutePath() + ")");
		}

		if (source.isDirectory()) { // directory
			if (source.getName().compareTo(".svn") != 0) {
				destination.mkdirs();
				files = source.list();
				for (String file : files) {
					FileUtils.copy(new File(source, file), new File(destination, file));
				}
			}
		} else { // file
			destination.getParentFile().mkdirs();
			inChannel = (new FileInputStream(source)).getChannel();
			outChannel = (new FileOutputStream(destination)).getChannel();
			size = inChannel.size();
			pos = 0;
			while (pos < size) {
				pos += inChannel.transferTo(pos, FileUtils.MAX_COUNT, outChannel);
			}
			inChannel.close();
			outChannel.close();
		}
	}

	/**
	 * Copy inputstream to outputstream
	 * 
	 * @param is Input stream
	 * @param os Output stream
	 * @throws IOException File exception
	 */
	public static void copy(InputStream is, OutputStream os) throws IOException {
		byte[] buf = new byte[1024];
		int len = 0;
		while ((len = is.read(buf)) != -1) {
			os.write(buf, 0, len);
		}
	}

	/**
	 * Move file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @throws IOException File exception
	 */
	public static void move(String source, String destination) throws IOException {
		FileUtils.move(new File(source), new File(destination));
	}

	/**
	 * Move file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @throws IOException File exception
	 */
	public static void move(File source, File destination) throws IOException {
		FileUtils.move(source, destination, false);
	}

	/**
	 * Move file object from given source to given destination with all subfiles
	 * 
	 * @param source Source file/directory
	 * @param destination Destination file/directory
	 * @param force Force to override destination
	 * @throws IOException File exception
	 */
	public static void move(File source, File destination, boolean force) throws IOException {
		FileUtils.copy(source, destination, force);
		FileUtils.delete(source);
	}

	/**
	 * Delete file object from given source with all subfiles
	 * 
	 * @param del file/directory to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean delete(String del) {
		return FileUtils.delete(new File(del));
	}

	/**
	 * Delete file object from given source with all subfiles
	 * 
	 * @param del file/directory to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean delete(File del) {
		boolean bool = true;

		if (del.exists()) {
			bool &= FileUtils.deleteContent(del);
			bool &= del.delete();
		}
		return bool;
	}

	/**
	 * Delete file object from given source with all subfiles
	 * 
	 * @param del file/directories to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean delete(String... del) {
		boolean success = true;
		for (String d : del) {
			success &= FileUtils.delete(d);
		}
		return success;
	}

	/**
	 * Delete file object from given source with all subfiles
	 * 
	 * @param del file/directories to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean delete(File... del) {
		boolean success = true;
		for (File d : del) {
			success &= FileUtils.delete(d);
		}
		return success;
	}

	/**
	 * Delete all subfiles and folders of given source
	 * 
	 * @param del directory to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean deleteContent(String del) {
		return FileUtils.deleteContent(new File(del));
	}

	/**
	 * Delete all subfiles and folders of given source
	 * 
	 * @param del directory to delete
	 * @return true if and only if the file or directory is successfully
	 * deleted; false otherwise
	 */
	public static boolean deleteContent(File del) {
		String[] files;
		boolean bool = true;

		if (del.exists()) {
			if (del.isDirectory()) {
				files = del.list();
				for (String file : files) {
					bool &= FileUtils.delete(new File(del, file));
				}
			}
		}
		return bool;
	}

	/**
	 * Delete all files (no folder) in given source-directory
	 * 
	 * @param del directory to delete
	 * @return true if and only if the file is successfully deleted; false
	 * otherwise
	 */
	public static boolean deleteContentFiles(String del) {
		return FileUtils.deleteContentFiles(new File(del));
	}

	/**
	 * Delete all files (no folder) in given source-directory
	 * 
	 * @param del directory to delete
	 * @return true if and only if the file is successfully deleted; false
	 * otherwise
	 */
	public static boolean deleteContentFiles(File del) {
		File[] files;
		boolean bool = true;

		if (del.exists()) {
			if (del.isDirectory()) {
				files = del.listFiles();
				for (File file : files) {
					if (!file.isDirectory()) {
						bool &= file.delete();
					}
				}
			}
		}
		return bool;
	}

	/**
	 * Convert the content of a file to a representative string
	 * 
	 * @param file Name of the file to check
	 * @return String with represent the content of the file
	 * @throws IOException File exception
	 * @throws NoSuchAlgorithmException Undefined algorithm
	 */
	public static String checkSum(String file) throws IOException, NoSuchAlgorithmException {
		return FileUtils.checkSum(new File(file));
	}

	/**
	 * Convert the content of a file to a representative string
	 * 
	 * @param file File to check
	 * @return String with represent the content of the file
	 * @throws IOException File exception
	 * @throws NoSuchAlgorithmException Undefined algorithm
	 */
	public static String checkSum(File file) throws IOException, NoSuchAlgorithmException {
		return FileUtils.checkSum(new FileInputStream(file));
	}

	/**
	 * Convert the content of a file to a representative string
	 * 
	 * @param stream Stream to check
	 * @return String with represent the content of the file
	 * @throws IOException File exception
	 * @throws NoSuchAlgorithmException Undefined algorithm
	 */
	public static String checkSum(InputStream stream) throws IOException, NoSuchAlgorithmException {
		MessageDigest md;
		DigestInputStream is;
		byte[] buff = new byte[1024];

		md = MessageDigest.getInstance("MD5");
		is = new DigestInputStream(stream, md);
		while (is.read(buff) > 0) {
			;
		}
		is.close();

		return (new BigInteger(1, md.digest())).toString(16);
	}

	/**
	 * Convert the string to a representative string
	 * 
	 * @param content String to check
	 * @return String with represent the content
	 * @throws NoSuchAlgorithmException Undefined algorithm
	 */
	public static String getCheckSum(String content) throws NoSuchAlgorithmException {
		MessageDigest md;

		md = MessageDigest.getInstance("MD5");
		md.update(content.getBytes());

		return (new BigInteger(1, md.digest())).toString(16);
	}

	/**
	 * Get file for the given path
	 * 
	 * @param path Path of file
	 * @return File
	 */
	public static File getFile(String... path) {
		File file = new File("");

		if (path.length > 0) {
			if (path[0].contains(":") && !path[0].contains(":" + File.separator)) {
				file = new File(path[0] + File.separator);
			} else {
				file = new File(path[0]);
			}
			for (int k = 1; k < path.length; k++) {
				file = new File(file, path[k]);
			}
		}

		return file;
	}

	/**
	 * Trim path of file. Remove "." and ".." strings
	 * 
	 * @param path Path to trim
	 * @return Trimmed path
	 */
	public static String trimPath(String path) {
		String[] strList;
		LinkedList<String> pathList = new LinkedList<String>();
		int index;

		// trim backslash to 2
		while (path.contains("\\\\\\")) {
			path = path.replaceAll("\\\\\\\\\\\\", "\\\\");
		}

		// trim to last
		if (path.contains("\\\\")) {
			path = path.substring(path.lastIndexOf("\\\\"));
		}

		if (path.contains(":")) {
			path = path.substring(path.lastIndexOf(File.separator, path.lastIndexOf(":")) + 1);
		}

//		if (path.contains("=")) {
//			path = path.substring(
//					path.lastIndexOf(File.separator, path.lastIndexOf("="))
//							+ 1);
//		}

		if (path.startsWith("\\\\")) {
			index = path.indexOf(File.separator, 2);
			if (index != -1) {
				pathList.add(path.substring(0, index));
				path = path.substring(index + 1);
			} else {
				pathList.add(path);
				path = "";
			}
		}

		if (path.contains(":")) {
			index = path.indexOf(":");
			index = path.indexOf(File.separator, index + 1);
			if (index != -1) {
				pathList.add(path.substring(0, index));
				path = path.substring(index + 1);
			} else {
				pathList.add(path);
				path = "";
			}
		}

		strList = path.split(File.separator.replace("\\", "\\\\"));
		for (String str : strList) {
			if (str.compareTo(".") == 0) {
			} else if ((str.compareTo("..") == 0) && (pathList.size() > 0) && !pathList.getLast().contains(":")
					&& !pathList.getLast().contains("\\\\")
//					&& !pathList.getLast().contains("=")
					&& (pathList.getLast().compareTo("..") != 0)) {
				pathList.removeLast();
			} else {
				pathList.add(str);
			}
		}

		return FileUtils.getFile(pathList.toArray(new String[0])).getPath();
	}

	/**
	 * Trim path of file. Remove "." and ".." strings
	 * 
	 * @param file file to trim
	 * @return Trimmed path
	 */
	public static File trimPath(File file) {
		return new File(FileUtils.trimPath(file.getPath()));
	}

	/**
	 * Get relative path from start-path to end-path
	 * 
	 * @param from Starting path
	 * @param to End path
	 * @return Relative Path
	 */
	public static String getRelativePath(String from, String to) {
		LinkedList<String> pathFrom, pathTo;
		boolean add = false;
		int index;

		pathFrom = new LinkedList<String>();
		if (from.startsWith("\\\\")) {
			index = from.indexOf(File.separator, 2);
			if (index != -1) {
				pathFrom.add(from.substring(0, index));
				from = from.substring(index + 1);
			} else {
				pathFrom.add(from);
				from = "";
			}
		}
		if (!from.isEmpty() || pathFrom.isEmpty()) {
			pathFrom.addAll(Arrays.asList(FileUtils.trimPath(from).split(File.separator.replace("\\", "\\\\"))));
		}

		pathTo = new LinkedList<String>();
		if (to.startsWith("\\\\")) {
			index = to.indexOf(File.separator, 2);
			if (index != -1) {
				pathTo.add(to.substring(0, index));
				to = to.substring(index + 1);
			} else {
				pathTo.add(to);
				to = "";
			}
		}
		if (!to.isEmpty() || pathTo.isEmpty()) {
			pathTo.addAll(Arrays.asList(FileUtils.trimPath(to).split(File.separator.replace("\\", "\\\\"))));
		}

		if (pathFrom.get(0).compareToIgnoreCase(pathTo.get(0)) == 0) {
			for (String str : pathFrom) {
				if (!add) {
					add = (pathTo.size() == 0) || (str.compareToIgnoreCase(pathTo.getFirst()) != 0);
				}

				if (add) {
					pathTo.addFirst("..");
				} else {
					pathTo.removeFirst();
				}
			}
		}

		return FileUtils.getFile(pathTo.toArray(new String[0])).getPath();
	}

	/**
	 * Get relative path from start-path to end-path
	 * 
	 * @param from Starting path
	 * @param to End path
	 * @return Relative Path
	 */
	public static File getRelativePath(File from, File to) {
		return new File(FileUtils.getRelativePath(from.getPath(), to.getPath()));
	}

	/**
	 * Split file-name in name and extension
	 * 
	 * @param name Name to split
	 * @return List of name an extension
	 */
	public static String[] splitName(String name) {
		String[] file;
		int index;

		file = new String[2];
		index = name.lastIndexOf('.');
		if (index == -1) {
			file[0] = name;
			file[1] = "";
		} else {
			file[0] = name.substring(0, index);
			file[1] = name.substring(index);
		}

		return file;
	}

	/**
	 * List files and folders of directory recursive
	 * 
	 * @param path Path of directory
	 * @return List of files and folders
	 */
	public static String[] listFiles(String path) {
		File[] list;
		String[] files;

		list = FileUtils.listFiles(new File(path));
		files = new String[list.length];
		for (int k = 0; k < list.length; k++) {
			files[k] = list[k].getPath();
		}

		return files;
	}

	/**
	 * List files and folders of directory recursive
	 * 
	 * @param folder Directory to list
	 * @return List of files and folders
	 */
	public static File[] listFiles(File folder) {
		LinkedList<File> list = new LinkedList<File>();

		FileUtils.collectFiles(folder, list);

		return list.toArray(new File[0]);
	}

	/**
	 * Collect files of directory
	 * 
	 * @param folder Folder to list
	 * @param list List of collected files
	 */
	private static void collectFiles(File folder, LinkedList<File> list) {
		for (File f : folder.listFiles()) {
			list.add(f);
			if (f.isDirectory()) {
				FileUtils.collectFiles(f, list);
			}
		}
	}
}
