#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 944 $
$LastChangedDate:: 2019-02-20 16:44:30 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import numpy as np
import lcm.utils.pytperiodic as pm
from abc import ABCMeta, abstractmethod
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec


class AbstractAnalyticFunction(metaclass=ABCMeta):

    def __init__(self, n_x, n_y, n_phi):
        self.n_fncs = 4
        self.n_x = n_x
        self.n_y = n_y
        self.n_phi = n_phi
        self.phi = np.linspace(0, 2 * np.pi, self.n_phi, endpoint=False)

    def getXc(self):
        xc_samples = np.empty((2, self.n_x * self.n_y, self.n_phi), dtype=np.float)
        for col, y0 in enumerate(np.linspace(-2, 2, self.n_y)):
            for row, x0 in enumerate(np.linspace(-1, 1, self.n_x)):
                idx = col * self.n_x + row
                xc_samples[0, idx] = x0 * np.ones_like(self.phi)
                xc_samples[1, idx] = y0 * np.ones_like(self.phi)

        return xc_samples

    def getZc(self):
        xc_samples = self.getXc()

        '''zc_samples'''
        n_samples = xc_samples.shape[1]
        zc_samples = np.empty((self.n_fncs, n_samples, self.n_phi), dtype=np.float)
        '''dz/dx[0]'''
        for (idx_p, p) in enumerate(self.phi):
            for idx_center in range(0, n_samples):
                zc_samples[:, idx_center, idx_p] = self.getZiDiscrete(*xc_samples[:, idx_center, idx_p], p)

        return zc_samples

    def getZi(self, x, y, phi):
        """
        Definition der Testfunktion [f1, f2, f3] = fun1(x, y, phi)
        """

        zi_disc = self.getZiDiscrete(x, y, phi)

        samples = np.empty((self.n_fncs, 1, phi.size), dtype=np.float)
        for row in range(self.n_fncs):
            samples[row, 0] = zi_disc[row, :]
        return samples

    def getXcPer(self):
        return pm.PyTPeriodicMatrix(self.getXc(), mode="samples")

    def getZcPer(self):
        return pm.PyTPeriodicMatrix(self.getZc(), mode="samples")

    def getXiPer(self, xi):
        return pm.PyTPeriodicMatrix(xi[:, np.newaxis, :], mode="samples")

    def getZiPer(self, x, y, phi):
        samples = self.getZi(x, y, phi)
        return pm.PyTPeriodicMatrix(samples, mode='samples')

    def get_dzi_dxPer(self, x, y, phi):
        samples = self.get_dzi_dx(x, y, phi)
        return pm.PyTPeriodicMatrix(samples, mode='samples')

    def get_dzi_dphiPer(self, x, y, phi, omega):
        samples = self.get_dzi_dphi(x, y, phi, omega)
        return pm.PyTPeriodicMatrix(samples, mode='samples')

    @abstractmethod
    def getZiDiscrete(self, x, y, phi):
        raise NotImplementedError

    @abstractmethod
    def get_dzi_dx(self, x, y, phi):
        raise NotImplementedError

    @abstractmethod
    def get_dzi_dphi(self, x, y, phi, omega):
        raise NotImplementedError

    def plotComparison(self, x_interp, x_analytic, atol, rtol, x_periodic=None):

        nrows = x_interp.shape[0]
        ncols = x_interp.shape[1]
        fig = plt.figure(figsize=(6 * 2, 4 * 2))
        gs = GridSpec(nrows, ncols * 2, figure=fig)

        for row in range(nrows):
            for col in range(ncols):
                a = x_interp[row, col, :]
                b = x_analytic[row, col, :]

                error = np.abs(a - b)
                threshold = atol + rtol * np.abs(b)

                ax1 = fig.add_subplot(gs[row, col])
                ax1.plot(self.phi, a, 'r-.', label='z_rbf[{},{}]'.format(row, col))
                ax1.plot(self.phi, b, 'g--', label='z_analytic[{},{}]'.format(row, col))
                if x_periodic is not None:
                    ax1.plot(self.phi, x_periodic[row, col, :], 'b:', label='periodic[{},{}]'.format(row, col))
                ax1.grid(True)
                ax1.legend(loc=1)

                ax2 = fig.add_subplot(gs[row, col + ncols])
                ax2.plot(self.phi, error, 'r-.', label='error[{},{}]'.format(row, col))
                ax2.plot(self.phi, threshold, 'g--', label='threshold[{},{}]'.format(row, col))
                ax2.grid(True)
                ax2.legend(loc=1)

                plt.tight_layout()

        plt.show()

    def calcError(self, a, b):
        return np.abs(a - b)


class AnalyticFunction1(AbstractAnalyticFunction):
    def getZiDiscrete(self, x, y, phi):
        """
        Definition der Testfunktion [f1, f2, f3] = fun1(x, y, phi)
        """

        f1 = 2. * x - 3. * y
        f2 = 2. * x - 3. * y + 0.5 * np.sin(2. * phi)
        f3 = 2. * x - 3. * y - x * y ** 2
        f4 = 2. * x - 3. * y - x * y ** 2 + np.sin(4. * phi)

        return np.array([f1, f2, f3, f4])[:self.n_fncs]

    def get_dzi_dx(self, x, y, phi):
        samples = np.empty((self.n_fncs, 2, phi.size), dtype=np.float)
        '''dz/dx[0]'''
        samples[0, 0] = 2. * np.ones_like(x)
        samples[1, 0] = 2. * np.ones_like(x)
        samples[2, 0] = 2. * np.ones_like(x) - y ** 2
        samples[3, 0] = 2. * np.ones_like(x) - y ** 2
        '''dz/dx[1]'''
        samples[0, 1] = -3. * np.ones_like(y)
        samples[1, 1] = -3. * np.ones_like(y)
        samples[2, 1] = -3. * np.ones_like(y) - 2 * y * x
        samples[3, 1] = -3. * np.ones_like(y) - 2 * y * x
        return samples

    def get_dzi_dphi(self, x, y, phi, omega):
        samples = np.empty((self.n_fncs, 1, phi.size), dtype=np.float)
        samples[0, 0] = np.zeros_like(phi)
        samples[1, 0] = np.cos(2. * phi)
        samples[2, 0] = np.zeros_like(phi)
        samples[3, 0] = 4 * np.cos(4. * phi)

        return samples * omega


class AnalyticFunction2(AbstractAnalyticFunction):
    def getZiDiscrete(self, x, y, phi):
        """
        Definition der Testfunktion [f1, f2, f3] = fun1(x, y, phi)
        """

        f1 = 2. * x - 3. * y
        f2 = 2. * x - 3. * y + 0.5
        f3 = 2. * x - 3. * y
        f4 = 2. * x - 3. * y

        return np.array([f1, f2, f3, f4])[:self.n_fncs]

    def get_dzi_dx(self, x, y, phi):
        samples = np.empty((self.n_fncs, 2, phi.size), dtype=np.float)
        '''dz/dx[0]'''
        samples[0, 0] = 2. * np.ones_like(x)
        samples[1, 0] = 2. * np.ones_like(x)
        samples[2, 0] = 2. * np.ones_like(x)
        samples[3, 0] = 2. * np.ones_like(x)
        '''dz/dx[1]'''
        samples[0, 1] = -3. * np.ones_like(y)
        samples[1, 1] = -3. * np.ones_like(y)
        samples[2, 1] = -3. * np.ones_like(y)
        samples[3, 1] = -3. * np.ones_like(y)
        return samples

    def get_dzi_dphi(self, x, y, phi, omega):
        samples = np.empty((self.n_fncs, 1, phi.size), dtype=np.float)
        samples[0, 0] = np.zeros_like(phi)
        samples[1, 0] = np.zeros_like(phi)
        samples[2, 0] = np.zeros_like(phi)
        samples[3, 0] = np.zeros_like(phi)

        return samples * omega
