#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 1037 $
$LastChangedDate:: 2020-11-13 10:49:19 +0100#$

This file is part of SyMSpace.

Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import warnings
import numpy as np
from lcm.tdata_interface.pyjavadata_interface import PyJavaDataInterface
from lcm.utils.pytperiodic import PyTPeriodicMatrix


class PyTPeriodicInterface(PyJavaDataInterface):
    """TPeriodic conversion

    Java            | Python
    -------------------------
    TPeriodic       | PyTPeriodicMatrix
    TPeriodicMatrix | PyTPeriodicMatrix

    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "Periodic_mat"
        equivalent to given Java object of type "TPeriodic" or "TPeriodicMatrix"
        """

        if jobj.getClassName() == "TPeriodic":

            max_order = jobj.getMaxOrder()
            nbins = 1 + 2 * max_order
            rfft = np.zeros((1, 1, nbins))

            ''' dc part'''
            rfft[0, 0, 0] = jobj.mean() * nbins

            for order in range(1, max_order + 1):
                ''' cosine coefficient'''
                rfft[0, 0, 2 * order - 1] = jobj.getCoeffA(order) * nbins / 2.

                ''' sine coefficient'''
                rfft[0, 0, 2 * order] = -jobj.getCoeffB(order) * nbins / 2.

            pyobj = PyTPeriodicMatrix(rfft, mode="rfft", nopack=True)

        elif jobj.getClassName() == "TPeriodicMatrix":

            sz = jobj.size()
            rows = sz[0]
            cols = sz[1]
            pyobj = PyTPeriodicMatrix(rows, cols, mode="zero", nopack=True)

            for r in range(0, rows):
                for c in range(0, cols):
                    pyobj[r, c] = PyTPeriodicInterface.convert_J2Py(jobj.get(r, c))

        else:
            raise ValueError("unsupported argument type")

        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "TPeriodic" or "TPeriodicMatrix"
        equivalent to given python object of type "Periodic_mat"
        """
        # noinspection PyUnresolvedReferences
        from at.lcm.bu21.general.dtypes import TPeriodic
        # noinspection PyUnresolvedReferences
        from at.lcm.bu21.general.dtypes import TPeriodicMatrix
        # noinspection PyUnresolvedReferences
        from org.apache.commons.math3.linear import ArrayRealVector

        rows = pyobj.nrows()
        cols = pyobj.ncols()

        if rows > 1 or cols > 1:

            ''' it is a periodic matrix '''
            jobj = TPeriodicMatrix(rows, cols)

            for r in range(0, rows):
                for c in range(0, cols):
                    jobj.set(r, c, PyTPeriodicInterface.convert_Py2J(pyobj[r, c]))

        else:
            ''' it is a single periodic signal '''
            (dc, cos_coeffs, sin_coeffs) = pyobj.get_coeffs()

            dc = dc[0, 0]

            n = cos_coeffs.shape[2]

            cos_vec = ArrayRealVector(n)
            for k in range(0, n):
                cos_vec.setEntry(k, cos_coeffs[0, 0, k])

            n = sin_coeffs.shape[2]
            sin_vec = ArrayRealVector(n)
            for k in range(0, n):
                sin_vec.setEntry(k, sin_coeffs[0, 0, k])

            jobj = TPeriodic()

            ''' Workaround, it was not possible to pass double-arrays to setCoeffs  '''
            jobj.setCoeffs2(dc, cos_vec, sin_vec)

        return jobj


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.general.dtypes import TPeriodic
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface

    magnitude = 2.0
    py_cos = PyTPeriodicMatrix(1., magnitude, 0.)  # cosine
    py_sin = PyTPeriodicMatrix(1, magnitude, -np.pi / 2)  # sine

    j_cos = TPeriodic(int(1), float(magnitude), float(0.))  # cosine
    j_sin = TPeriodic(int(1), float(magnitude), float(-np.pi / 2))  # sine

    tmp_jobj = PyJDataInterface.convert_Py2J(py_cos)
    tmp_pyobj = PyJDataInterface.convert_J2Py(tmp_jobj)

    print('java:' + str(tmp_jobj))
    print('python:' + str(tmp_pyobj))
    print('python type:' + str(type(tmp_pyobj)))

    max_order = j_cos.getMaxOrder()
    a_coeffs = np.zeros(max_order)
    b_coeffs = np.zeros(max_order)

    for order in range(max_order):
        a_coeffs[order] = j_cos.getCoeffA(order + 1)
        b_coeffs[order] = j_cos.getCoeffB(order + 1)

    dc, cos_coeffs, sin_coeffs = py_cos.get_coeffs(range(max_order + 1))


