/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.gui.views;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JComponent;
import javax.swing.JSplitPane;

/**
 * Split-panel for views
 *
 * @author Werner Koppelsttter
 */
public class SplitView extends View {

	/** Left view */
	protected View left = null;
	/** Right view */
	protected View right = null;

	/** Divider weight */
	protected double divider = 0.5;

	/** Split horizontal or vertiacal */
	private boolean horizontal = true;

	/** Split-pane */
	JSplitPane split = null;

	private boolean visible = false;

	/**
	 * Create split-view
	 * 
	 * @param manager
	 *            View-manager
	 */
	public SplitView(ViewManager manager) {
		super(manager);
		this.split = new JSplitPane();
		this.setHorizontal(this.horizontal);
		this.split.setResizeWeight(this.divider);
		this.split.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY, new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				if (SplitView.this.visible) {
					if (SplitView.this.horizontal) {
						SplitView.this.divider = ((double) SplitView.this.split.getDividerLocation())
								/ (SplitView.this.split.getWidth() - SplitView.this.split.getDividerSize());
					} else {
						SplitView.this.divider = ((double) SplitView.this.split.getDividerLocation())
								/ (SplitView.this.split.getHeight() - SplitView.this.split.getDividerSize());
					}

				} else if ((SplitView.this.split.getWidth() != 0) && (SplitView.this.split.getHeight() != 0)) {
					SplitView.this.visible = true;
				}
			}
		});
	}

	@Override
	public JComponent getComponent() {
		return this.split;
	}

	/**
	 * Set horizontal or vertical split
	 * 
	 * @param hor
	 *            Horizontal split
	 */
	public void setHorizontal(boolean hor) {
		this.horizontal = hor;
		if (this.horizontal) {
			this.split.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
		} else {
			this.split.setOrientation(JSplitPane.VERTICAL_SPLIT);
		}
		this.split.revalidate();
		this.split.repaint();
	}

	/**
	 * Set divider weight
	 * 
	 * @param div
	 *            Weight
	 */
	public void setDivider(double div) {
		this.divider = div;
		// this.split.setResizeWeight(this.divider);
		// this.split.setDividerLocation(proportionalLocation);ResizeWeight(this.divider);
	}

	/**
	 * Set left view
	 * 
	 * @param view
	 *            View
	 */
	public void setLeftView(View view) {
		this.left = view;
		if (this.left != null) {
			this.split.setLeftComponent(this.left.getComponent());
		} else {
			this.split.setLeftComponent(null);
		}
		this.split.revalidate();
		this.split.repaint();
	}

	/**
	 * Set right view
	 * 
	 * @param view
	 *            View
	 */
	public void setRightView(View view) {
		this.right = view;
		if (this.right != null) {
			this.split.setRightComponent(this.right.getComponent());
		} else {
			this.split.setRightComponent(null);
		}
		this.split.revalidate();
		this.split.repaint();
	}

	@Override
	public void refresh() {
		this.split.setDividerLocation(this.divider);

		if (this.left != null) {
			this.left.refresh();
		}
		if (this.right != null) {
			this.right.refresh();
		}

	}

	@Override
	public View getView(ViewItem item) {
		View view = null;

		if (this.left != null) {
			view = this.left.getView(item);
		}
		if ((view == null) && (this.right != null)) {
			view = this.right.getView(item);
		}

		return view;
	}

	@Override
	public void setEnabled(boolean aFlag) {
		if (this.left != null) {
			this.left.setEnabled(aFlag);
		}
		if (this.right != null) {
			this.right.setEnabled(aFlag);
		}
	}
}
