/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9814 $
 * $Date: 2020-04-30 16:26:03 +0200 (Thu, 30 Apr 2020) $
 */
package at.lcm.symspace.optimizer.views;

import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import at.lcm.bu21.optim.gui.mapping.PanelMapping;
import at.lcm.bu21.optim.gui.mapping.PanelSensitivity;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import at.lcm.symspace.gui.views.DisablePanel;
import at.lcm.symspace.gui.views.SplitView;
import at.lcm.symspace.gui.views.TabView;
import at.lcm.symspace.gui.views.ViewManager;
import at.lcm.symspace.optimizer.OptimizerController;
import at.lcm.symspace.optimizer.views.compare.PanelCompare;
import at.lcm.symspace.optimizer.views.pool.PoolViewPanel;
import magopt.gui.console.Console;
import magopt.gui.console.PythonConsole;
import magopt.gui.event.EventView;
import magopt.gui.view.ParamListener;

/**
 * Manager for all views of optimizer
 *
 * @author Werner Koppelsttter
 */
public class OptimizerViewManager extends ViewManager<OptimizerViewItem> implements ParamListener {

	/** Pool view identifier */
	public OptimizerViewItem POOL = new OptimizerViewItem(OptimizerViewType.POOL, "Pool");
	/** Logging view identifier */
	public OptimizerViewItem EVENT = new OptimizerViewItem(OptimizerViewType.EVENT, "Logging");
	/** Console view identifier */
	public OptimizerViewItem CONSOLE = new OptimizerViewItem(OptimizerViewType.CONSOLE, "Console");
	/** Python console view identifier */
	public OptimizerViewItem PYTHON = new OptimizerViewItem(OptimizerViewType.PYTHON, "Python Console");
	/** Polt views identifier */
	public LinkedList<OptimizerViewItem> PLOTS = new LinkedList<OptimizerViewItem>();
	/** Compare view identifier */
	public OptimizerViewItem COMPARE = new OptimizerViewItem(OptimizerViewType.COMPARE, "Compare");
	/** Mapping view identifier */
	public OptimizerViewItem MAPPING = new OptimizerViewItem(OptimizerViewType.MAPPING, "Mapping");
	/** Sensitivity view identifier */
	public OptimizerViewItem SENSITIVITY = new OptimizerViewItem(OptimizerViewType.SENSITIVITY, "Sensitivity");

	/** Panel to insert plots */
	private TabView plotTab;

	/** Pool panel */
	private PoolViewPanel pool;

	/** Logging panel */
	private EventView event;
	/** Console */
	private Console console;
	/** Python console */
	public PythonConsole pythonConsole;

	/** Plot panels */
	public PanelPlots plots = null;

	/** Mapping panel */
	private PanelMapping mapping = null;

	/** Sensitivity panel */
	private PanelSensitivity sensitivity = null;

	/** Compare panel */
	private PanelCompare compare;

	/**
	 * Create view-manager for optimizer
	 * 
	 * @param controller
	 *            Controller
	 */
	public OptimizerViewManager(OptimizerController controller) {
		super(controller, new OptimizerMenuBar(controller), new OptimizerToolBar(controller),
				new OptimizerStatusBar(controller));

		SplitView root, split;
		TabView tab;

		// create view-items
		this.viewItems.add(this.POOL);
		this.viewItems.add(this.EVENT);
		this.viewItems.add(this.CONSOLE);
		this.viewItems.add(this.PYTHON);
		// this.viewItems.add(this.MAPPING);
		// this.viewItems.add(this.SENSITIVITY);
		this.viewItems.add(this.COMPARE);

		// create views
		this.pool = new PoolViewPanel();

		this.event = new EventView();
		this.console = new Console();
		System.setOut(this.console.getOutputPrintStream());
		System.setErr(this.console.getErrorPrintStream());
		try {
			this.pythonConsole = new PythonConsole();
		} catch (Throwable ex) {
		}
		this.compare = new PanelCompare();

		// define view-structure
		root = new SplitView(this);
		this.views = root;
		root.setHorizontal(false);
		root.setDivider(0.7);

		split = new SplitView(this);
		split.setHorizontal(true);
		split.setDivider(0.25);
		root.setRightView(split);

		// top
		tab = new TabView(this);
		root.setLeftView(tab);
		this.plotTab = tab;
		tab.addView(new OptimizerViewItem(OptimizerViewType.EMPTY, "        "));
		tab.addView(this.COMPARE);

		// bottom left
		tab = new TabView(this);
		split.setLeftView(tab);
		tab.addView(this.POOL);

		// bottom right
		tab = new TabView(this);
		split.setRightView(tab);
		tab.addView(this.EVENT);
		tab.addView(this.CONSOLE);
		tab.addView(this.PYTHON);

		this.refresh();
	}

	@Override
	public OptimizerController getController() {
		return (OptimizerController) super.getController();
	}

	/**
	 * Set controller
	 * 
	 * @param controller
	 *            Controller
	 */
	public void setController(OptimizerController controller) {
		super.setController(controller);
	}

	@Override
	public void refresh(OptimizerViewItem item) {
		if (item != null) {
			switch (item.getType()) {

			case EVENT:
				if ((this.getController() != null) && (this.getController().getOptimization() != null)) {
					// FIXME IEventLog
					this.event.setEventLog(this.getController().getOptimization().getEventLog());
				} else {
					this.event.setEventLog(null);
				}
				break;
			case CONSOLE:
				break;
			case PYTHON:
				break;
			case POOL:
				// FIXME IProject
				if ((this.getController() != null) && (this.getController().getOptimization() != null)) {
					this.pool.set(this.getController().getOptimization().getProjectControl());
				} else {
					this.pool.set(null);
				}
				break;
			case PLOTS:
				// if (this.plots != null) {
				// this.plots.refresh();
				// }
				break;
			case COMPARE:
				if (this.getController() != null) {
					this.compare.addCompareObject(this.getController().getProblem());
				}
				break;
			case MAPPING:
				// FIXME
				if (this.mapping != null) {
					this.mapping.refresh();
				}
				break;
			case SENSITIVITY:
				// FIXME
				if (this.sensitivity != null) {
					this.sensitivity.refresh();
				}
				break;
			default:
				break;
			}
		}
	}

	@Override
	public void refresh() {
		String title;
		int fix, var;
		OptimizerViewItem plot;
		Iterator<OptimizerViewItem> it;

		// gui title
		if (this.getController().getGuiComponent() instanceof JFrame) {
			title = "SyMSpace Optimizer";
			if (this.getController().getProblem() != null) {
				title += " - " + this.getController().getProblem().getArchive().getMop().getName() + " : "
						+ this.getController().getProblem().getArchive().getName() + " : "
						+ this.getController().getProblem().getName() + " ("
						+ this.getController().getProblem().getArchive().getMop().getParent() + ")";
			}
			((JFrame) this.getController().getGuiComponent()).setTitle(title);
		}

		// plots
		if ((this.getController() != null) && (this.getController().getOptimization() != null)) {
			if (this.plots == null) {
				try {
					this.plots = new PanelPlots(this.getController().getOptimization());
				} catch (Exception e) {
					e.printStackTrace();
				}
			}

			fix = this.plots.getNumFixPlots();
			var = this.plots.getNumUserPlots();

			// update fix plots
			for (int k = 0; k < fix; k++) {
				plot = null;
				for (OptimizerViewItem item : this.viewItems) {
					if ((item.getType() == OptimizerViewType.PLOTS) && (item.getId() == k)) {
						plot = item;
					}
				}

				if ((plot != null) && !plot.getName().equals(this.plots.getFixPlotName(k))) {
					this.viewItems.remove(plot);
					this.plotTab.removeView(plot);
				}
				if (plot == null) {
					plot = new OptimizerViewItem(OptimizerViewType.PLOTS, this.plots.getFixPlotName(k), k);
					this.viewItems.add(plot);
					this.plotTab.addView(plot, k);
					this.plotTab.selectIndex(1);
				}
			}

			// update userplots
			for (int k = fix; k < (fix + var); k++) {
				plot = null;
				for (OptimizerViewItem item : this.viewItems) {
					if ((item.getType() == OptimizerViewType.PLOTS) && (item.getId() == k)) {
						plot = item;
					}
				}

				if ((plot != null) && !plot.getName().equals(this.plots.getUserPlotName(k - fix))) {
					this.viewItems.remove(plot);
					this.plotTab.removeView(plot);
				}

				if (plot == null) {
					plot = new OptimizerViewItem(OptimizerViewType.PLOTS, this.plots.getUserPlotName(k - fix), k);
					this.viewItems.add(plot);
					this.plotTab.addView(plot, k);
				}
			}

			// remove other plots
			it = this.viewItems.iterator();
			while (it.hasNext()) {
				plot = it.next();
				if (plot.getType() == OptimizerViewType.PLOTS) {
					if (plot.getId() >= (fix + var)) {
						it.remove();
						this.plotTab.removeView(plot);
					}
				}
			}

		} else {
			this.plots = null;
			// remove all plots
			it = this.viewItems.iterator();
			while (it.hasNext()) {
				plot = it.next();
				if (plot.getType() == OptimizerViewType.PLOTS) {
					it.remove();
					this.plotTab.removeView(plot);
				}
			}
		}

		// mapping and sensitivity
		if ((this.getController() != null) && (this.getController().getOptimization() != null)) {
			if (this.mapping == null) {
				try {
					this.mapping = new PanelMapping(this.getController().getOptimization());
					this.viewItems.add(this.MAPPING);
					this.plotTab.addView(this.MAPPING);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			if (this.sensitivity == null) {
				try {
					this.sensitivity = new PanelSensitivity(this.getController().getOptimization());
					this.viewItems.add(this.SENSITIVITY);
					this.plotTab.addView(this.SENSITIVITY);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} else {
			this.mapping = null;
			if (this.viewItems.contains(this.MAPPING)) {
				this.viewItems.remove(this.MAPPING);
				this.plotTab.removeView(this.MAPPING);
			}

			this.sensitivity = null;
			if (this.viewItems.contains(this.SENSITIVITY)) {
				this.viewItems.remove(this.SENSITIVITY);
				this.plotTab.removeView(this.SENSITIVITY);
			}
		}

		super.refresh();
	}

	@Override
	public JComponent getComponent(OptimizerViewItem view) {
		JComponent comp = null;

		switch (view.getType()) {
		case POOL:
			comp = new DisablePanel(this.pool);
			break;
		case EVENT:
			comp = this.event;
			break;
		case CONSOLE:
			comp = this.console;
			break;
		case PYTHON:
			comp = new DisablePanel(this.pythonConsole);
			break;
		case PLOTS:
			if (view.getId() < this.plots.getNumFixPlots()) {
				comp = new DisablePanel(this.plots.getFixPlot(view.getId()).getPanel());
			} else if (view.getId() < (this.plots.getNumFixPlots() + this.plots.getNumUserPlots())) {
				comp = new DisablePanel(this.plots.getUserPlot(view.getId() - this.plots.getNumFixPlots()).getPanel());
			}
			break;
		case MAPPING:
			comp = new DisablePanel(new JScrollPane(this.mapping));
			break;
		case SENSITIVITY:
			comp = new DisablePanel(new JScrollPane(this.sensitivity));
			break;
		case COMPARE:
			comp = new DisablePanel(this.compare);
			break;
		case EMPTY:
			comp = new JPanel();
			comp.setSize(1800, 1600);
			break;
		}

		return comp;
	}

	@Override
	public void parameterChanged() {
		this.refresh();
	}
}
