/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10212 $
 * $Date: 2021-05-20 14:45:22 +0200 (Thu, 20 May 2021) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import org.json.simple.JSONObject;
import org.json.simple.JSONValue;

import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.FillTexReport;
import at.lcm.bu21.magopt.structure.client.CComponent;
import at.lcm.bu21.magopt.structure.rmi.RComponent;

/**
 * Component object for a container. Component contain linked files and linked
 * help-files.
 * 
 * @author Werner Koppelsttter
 */
public class Component implements IComponent {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Folder for component data */
	public static final String ComponentFolder = "_Component";

	/** Parent container of component */
	private ParamContainer container;

	/** Path of used project */
	private String path = null;

	/** Parameter identifier of used container */
	private String pid = null;

	/** List of linked-files (relative path) */
	private ArrayList<String> linkedFiles = null;

	/**
	 * Clean files-directory (remove non component-files)
	 * 
	 * @param proj Project to clean
	 */
	public static void cleanup(Project proj) {
		File folder, compFolder;
		boolean exist;
		File[] list;
		LinkedList<File> files;
		Component comp;

		folder = proj.getAllFilesDirectory(false);

		if (folder.exists()) {
			Parameter[] containers = proj.getParameters(ParamContainer.class);

			// component files
			compFolder = new File(folder, Component.ComponentFolder);
			if (compFolder.exists()) {
				for (File f : compFolder.listFiles()) {
					exist = false;
					if (f.isDirectory()) {
						for (Parameter c : containers) {
							exist |= f.getName().equals(c.getPId()) && (((ParamContainer) c).getComponent() != null);
						}
					}
					if (!exist) {
						FileUtils.delete(f);
					}
				}
			}

			// TODO Clean all projects ???
			if (proj.getMopFile().isComponent()) {
				// linked files
				files = new LinkedList<File>();
				// collect file list
				for (File f : folder.listFiles()) {
					if (f.isDirectory() && f.getName().startsWith("_")
							&& !f.getName().equals(Component.ComponentFolder)) {
						files.addAll(Arrays.asList(FileUtils.listFiles(f)));
					}
				}
				// clear file list
				for (Parameter c : containers) {
					comp = ((ParamContainer) c).getComponent();
					if (comp != null) {
						for (File f : comp.listLinkedFiles()) {
							files.remove(f);
						}
					}
				}
				// delete files
				for (File f : files) {
					if (f.exists()) {
						list = f.listFiles();
						if ((list == null) || (list.length == 0)) {
							f.delete();
						}
					}
				}
			}

		}
	}

	/**
	 * Create component for container
	 * 
	 * @param container Parent container
	 */
	public Component(ParamContainer container) {
		this.container = container;
		this.refresh();
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Component other = (Component) obj;
		if (this.linkedFiles == null) {
			if (other.linkedFiles != null) {
				return false;
			}
		} else if (!this.linkedFiles.equals(other.linkedFiles)) {
			return false;
		}

		// TODO path , pid

		return true;
	}

	/**
	 * Clone content of this component to the other component.
	 * 
	 * @param other Component to fill
	 * @throws CloneException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	void cloneComponent(Component other) throws CloneException, InstantiationException, IllegalAccessException {

		if (this.linkedFiles == null) {
			other.linkedFiles = null;
		} else {
			if (other.linkedFiles == null) {
				other.linkedFiles = new ArrayList<String>();
			} else {
				other.linkedFiles.clear();
			}

			for (int k = 0; k < this.linkedFiles.size(); k++) {
				other.linkedFiles.add(this.linkedFiles.get(k));
			}
		}

		other.path = this.path;
		other.pid = this.pid;
	}

	@Override
	public ParamContainer getContainer() {
		return this.container;
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("Component\n");
		str.append("\n\tLinkedFiles : ");

		for (File f : this.listLinkedFiles()) {
			str.append("\n\t\t" + f.getAbsolutePath());
		}

		str.append("\n\tComponent-folder : " + this.getFolder().getAbsolutePath());

		return str.toString();
	}

	/**
	 * Read content form xml-file
	 * 
	 * @param xmlReader XML-reader
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("LinkedFiles")) {
				xmlReader.next();
				this.linkedFiles = new ArrayList<String>();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					xmlReader.next();
					tmp = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						tmp += xmlReader.getText();
						xmlReader.next();
					}
					this.linkedFiles.add(tmp);
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write content to xml-file
	 * 
	 * @param xmlWriter XML-Writer
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("Component");

		// Linked files
		xmlWriter.writeStartElement("LinkedFiles");
		if (this.linkedFiles != null) {
			for (String p : this.linkedFiles) {
				xmlWriter.writeStartElement("entry");
				xmlWriter.writeCharacters(p);
				xmlWriter.writeEndElement();
			}
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public String getRelLinkedFile(File file) {
		return FileUtils.getRelativePath(this.path, file.getPath());
	}

	@Override
	public File[] listLinkedFiles() {
		File[] files;
		int offset = 0;

		if (this.linkedFiles == null) {
			files = new File[0];
		} else {
			if (!this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{')) {
				offset = 1;
			}
			files = new File[this.linkedFiles.size() - offset];
			for (int k = offset; k < this.linkedFiles.size(); k++) {
				files[k - offset] = FileUtils.trimPath(new File(this.path, this.linkedFiles.get(k)));
			}
		}

		return files;
	}

	/**
	 * Add file to linked-files
	 * 
	 * @param file File to add
	 */
	public void addLinkedFile(String file) {
		this.addLinkedFile(new File(file));
	}

	/**
	 * Add file to linked-files
	 * 
	 * @param file File to add
	 */
	public void addLinkedFile(File file) {
		String f;

		f = this.getRelLinkedFile(file);

		if (this.linkedFiles == null) {
			this.linkedFiles = new ArrayList<String>();
			this.linkedFiles.add(f);
		} else {
			if (!this.linkedFiles.contains(f)) {
				this.linkedFiles.add(f);
			}
		}
	}

	/**
	 * Remove file from linked-files
	 * 
	 * @param file File to remove
	 */
	public void removeLinkedFile(File file) {
		String f;

		if (this.linkedFiles != null) {
			f = this.getRelLinkedFile(file);
			this.linkedFiles.remove(f);
		}

		if (this.linkedFiles.isEmpty()) {
			this.linkedFiles = null;
		}
	}

	/**
	 * Get linked-file of index
	 * 
	 * @param index Index of file
	 * @return Linked-file
	 */
	public File getLinkedFile(int index) {
		File file = null;

		if (this.linkedFiles != null) {

			if (!this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{')) {
				index++;
			}

			if ((0 <= index) && (index < this.linkedFiles.size())) {
				file = new File(this.path, this.linkedFiles.get(index));
			}
		}

		return file;
	}

	public Object getJSONValue(String key) {
		JSONObject json;
		Object obj = null;

		if ((this.linkedFiles != null) && !this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{')) {
			try {
				json = (JSONObject) JSONValue.parseWithException(this.linkedFiles.get(0));
				obj = json.get(key);
			} catch (Exception e) {
			}
		}
		return obj;
	}

	public void setJSONValue(String key, Object value) {
		JSONObject json = null;
		boolean exists = false;

		if (this.linkedFiles == null) {
			this.linkedFiles = new ArrayList<String>();
		}

		exists = !this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{');

		if (exists) {
			try {
				json = (JSONObject) JSONValue.parseWithException(this.linkedFiles.get(0));
			} catch (Exception e) {
			}
		} else {
			json = new JSONObject();
		}

		if (json != null) {
			json.put(key, value);
			if (exists) {
				this.linkedFiles.set(0, json.toJSONString());
			} else {
				this.linkedFiles.add(0, json.toJSONString());
			}
		}

	}

	public void removeJSONValue(String key) {
		JSONObject json = null;

		if ((this.linkedFiles != null) && !this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{')) {
			try {
				json = (JSONObject) JSONValue.parseWithException(this.linkedFiles.get(0));
				json.remove(key);
				if (json.isEmpty()) {
					this.linkedFiles.remove(0);
				} else {
					this.linkedFiles.set(0, json.toJSONString());
				}
				if (this.linkedFiles.isEmpty()) {
					this.linkedFiles = null;
				}
			} catch (Exception e) {
			}
		}
	}

	public String[] getJSONKeys() {
		JSONObject json = null;
		String[] keys = new String[0];

		if ((this.linkedFiles != null) && !this.linkedFiles.isEmpty() && (this.linkedFiles.get(0).charAt(0) == '{')) {
			try {
				json = (JSONObject) JSONValue.parseWithException(this.linkedFiles.get(0));
				keys = (String[]) json.keySet().toArray(keys);
			} catch (Exception e) {
			}
		}
		return keys;
	}

	@Override
	public File getHelpFile() {
		return new File(this.getHelpFolder(), "index.html");
	}

	@Override
	public File getImage() {
		return new File(this.getHelpFolder(), "Component.png");
	}

	/**
	 * Get help-file as absolute path
	 * 
	 * @return Help-file
	 */
	public File getHelpFolder() {
		return FileUtils.getFile(this.path, Component.ComponentFolder, this.pid, "Help");
	}

	@Override
	public File getFolder() {
		return FileUtils.getFile(this.path, Component.ComponentFolder, this.pid);
	}

	@Override
	public void refresh() {
		Parameter p;
		String newPath;
		File oldFile, newFile, file;
		Project proj;

		p = this.container.getRoot();
		if (p instanceof Project) {
			proj = (Project) p;
			if (!proj.getMopFile().getName().equals("error.mop")) {
				newPath = proj.getAllFilesDirectory().getPath();
				if (this.path != null) {
					if (!newPath.equals(this.path)) {
						if (this.linkedFiles != null) {
							// copy linked files
							for (String f : this.linkedFiles) {
								if (f.charAt(0) != '{') {
									if (this.path != null) {
										oldFile = new File(this.path, f);
										if (oldFile.exists()) {
											newFile = new File(newPath, f);
											try {
												FileUtils.copy(oldFile, newFile, true);
											} catch (IOException e) {
												this.container.warning("Unable to copy component file '" + f + "'!");
											}
										} else {
											this.container.warning("Component file doesn't exist!");
										}
									}
								}
							}
						}
					}

					// copy component files
					oldFile = FileUtils.trimPath(FileUtils.getFile(this.path, Component.ComponentFolder, this.pid));
					newFile = FileUtils.getFile(newPath, Component.ComponentFolder, this.container.getPId());
					if (oldFile.exists()) {
						if (!newFile.equals(oldFile)) {
							FileUtils.delete(newFile);
							try {
								FileUtils.copy(oldFile, newFile);
							} catch (IOException e) {
								this.container.warning("Unable to copy component files!");
							}
						}
					}

					if (!newPath.equals(this.path)) {
						this.path = newPath;
					}
					if (!this.container.getPId().equals(this.pid)) {
						this.pid = this.container.getPId();
					}
				} else {
					this.path = newPath;
					this.pid = this.container.getPId();
				}

				file = this.getHelpFolder();
				if (!file.exists()) {
					file.mkdirs();
				}
			}
		}
	}

	/**
	 * Function for generating documentation-files
	 * 
	 * @param directory Directory for export-files
	 */
	public void generateDocFiles(File directory) {
		File docu, filled, doc;
		String name;

		docu = new File(this.getFolder(), "Docu");
		if (docu.exists()) {
			for (File f : docu.listFiles()) {
				if (f.isFile() && f.getName().toLowerCase().endsWith(".tex")
						&& !f.getName().toLowerCase().endsWith("_filled.tex")) {
					name = f.getName().substring(0, f.getName().length() - 4);

					FillTexReport fill = new FillTexReport(f, (Project) this.container.getRoot(),
							this.container.getIdentifier());
					filled = fill.getFile();
					if (filled.exists()) {
						if (!fill.toString().contains("\\begin{document}")) {

							// extract body
							doc = this.container.getUniqueDocFileName(new File(directory, name + ".tex"));
							try {
								FileUtils.move(filled, doc, true);
								this.container.registerDocFile(name + ".tex", doc.getPath(),
										"Component tex template " + name);
							} catch (Exception e) {
								this.container.warning(e.getMessage(), e.getStackTrace());
							}
						} else {

							// compile
							int ret = -1;
							try {
								ret = fill.compile(false);
							} catch (IOException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}

							if (ret == 0) {
								File report = new File(
										filled.getPath().substring(0, filled.getPath().length() - 4) + ".pdf");
								DocFile docfile = this.container.getDocFile(name);
								if (docfile != null) {
									doc = new File(directory, new File(docfile.getFile()).getName());
								} else {
									doc = this.container.getUniqueDocFileName(new File(directory, report.getName()));
								}
								try {
									FileUtils.move(report, doc, true);
									this.container.registerDocFile(name, doc.getPath(), "Component template " + name);
								} catch (Exception e) {
									this.container.warning(e.getMessage(), e.getStackTrace());
								}
								filled.delete();
							} else {
								this.container.warning("Unable to compile DocFile '" + name + "'");
							}
						}
					} else {
						this.container.warning("Unable to create DocFile '" + name + "'");
					}
				}
			}
		}
	}

	/**
	 * Get RMI-adapter of component
	 * 
	 * @return RMI-adapter
	 * @throws RemoteException
	 */
	public IComponent getRMIImpl() throws RemoteException {
		return new RComponent(this);
	}

	@Override
	public IComponent getClientImpl() {
		return new CComponent(this);
	}
}
