/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10108 $
 * $Date: 2021-02-10 10:07:16 +0100 (Wed, 10 Feb 2021) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.util.LinkedList;

/**
 * Location of magopt-directories
 * 
 * @author Werner Koppelsttter
 */
public enum Location {
	/** Relative path to parameter-directories */
	PARAM("Rel. path to parameter-directories", false),
	/** Relative path to dataset-directory */
	DATASET("Rel. path to dataset-directory", false),
// TODO File version
//	/** Relative path to dataset-directory */
//	COMPONENT("Rel. path to component-directory", true),
	/** Relative path to project */
	PROJECT_FILES("Rel. path to project-files", true),
	/** Relative path to project */
	PROJECT("Rel. path to project", true),
	/** Relative path to SyMSpace-library */
	LIBRARY("Rel. path to SyMSpace-library", true),
	/** Relative path to SyMSpace */
	MAGOPT("Rel. path to SyMSpace", true),
	/** Absolute path */
	ABSOLUTE("Absolute path", true);

	/** Description of location */
	private String desc;

	/** Is location for constant possible */
	private boolean constantPossible;

	/**
	 * List of locations for functions
	 * 
	 * @return List of function-location
	 */
	public static Location[] functionValues() {
		LinkedList<Location> list;

		// filter only constant
		list = new LinkedList<Location>();
		for (Location loc : Location.values()) {
			if (loc.isConstantPossible()) {
				list.add(loc);
			}
		}

		// reorder PROJECT
		list.remove(Location.PROJECT);
		list.add(list.size() - 1, Location.PROJECT);

		return list.toArray(new Location[0]);
	}

	/**
	 * List of locations for files
	 * 
	 * @return List of file-location
	 */
	public static Location[] fileValues() {
		return Location.values();
	}

	/**
	 * Construct location
	 * 
	 * @param desc             Description
	 * @param constantPossible Is constant location is possible
	 */
	private Location(String desc, boolean constantPossible) {
		this.desc = desc;
		this.constantPossible = constantPossible;
	}

	/**
	 * Get path of location
	 * 
	 * @param param Parameter for location
	 * @return Path of location
	 */
	public String getPath(Parameter param) {
		String path = null;
		Parameter root;
// TODO File version
//		ParamContainer cont;

		switch (this) {
		case PARAM:
			if (param != null) {
				root = param.getRoot();
				if (root instanceof Project) {
					path = new File(((Project) root).getFilesDirectory(), "param").getPath();
				}
			}
			break;
		case DATASET:
			if (param != null) {
				root = param.getRoot();
				if (root instanceof Project) {
					path = ((Project) root).getFilesDirectory().getPath();
				}
			}
			break;
// TODO File version
//		case COMPONENT:
//			if (param != null) {
//				if (param instanceof ParamContainer) {
//					cont = (ParamContainer) param;
//				} else {
//					cont = param.getParent();
//				}
//				while ((cont != null) && (cont.getComponent() == null)) {
//					cont = cont.getParent();
//				}
//				if (cont != null) {
//					path = cont.getComponent().getFolder().getPath();
//				}
//			}
//			break;
		case PROJECT_FILES:
			if (param != null) {
				param = param.getRoot();
				if (param instanceof Project) {
					path = ((Project) param).getAllFilesDirectory().getPath();
				}
			}
			break;
		case PROJECT:
			if (param != null) {
				param = param.getRoot();
				if (param instanceof Project) {
					path = ((Project) param).getPath();
				}
			}
			break;
		case LIBRARY:
			path = Project.getLibraryPath();
			break;
		case MAGOPT: // TODO change with File-Verison
			path = Project.getInstallPath();
			break;
		case ABSOLUTE:
			path = "";
			break;
		}

		return path;
	}

	/**
	 * Is constant location is possible
	 * 
	 * @return Is constant
	 */
	public boolean isConstantPossible() {
		return this.constantPossible;
	}

	@Override
	public String toString() {
		return this.desc;
	}
}
