/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.compare;

import java.util.LinkedList;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.magopt.structure.Depend;
import at.lcm.bu21.magopt.structure.DependItem;
import at.lcm.bu21.magopt.structure.Dependency;
import at.lcm.bu21.magopt.structure.Function;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.ParamFormula;
import at.lcm.bu21.magopt.structure.ParamFunction;
import at.lcm.bu21.magopt.structure.ParamMaterial;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Parameter.ParameterPath;
import at.lcm.bu21.magopt.structure.Project;

/**
 * Compare parameters to get differences
 * 
 * @author Werner Koppelsttter
 */
public class CompareParameter {

	/** List of different */
	private LinkedList<CompareResult> diff = new LinkedList<CompareResult>();

	/** Source 1 */
	private String source1 = "";

	/** Source 2 */
	private String source2 = "";

	/** Compare results */
	private boolean compareResults;

	/**
	 * Compare parameter 1 with parameter 2
	 * 
	 * @param p1
	 *            Parameter 1
	 * @param index1
	 *            Index of parameter 1
	 * @param p2
	 *            Parameter 2
	 * @param index2
	 *            Index of parameter 2
	 */
	public void compare(Parameter p1, int index1, Parameter p2, int index2) {
		this.compare(p1, index1, p2, index2, false);
	}

	/**
	 * Compare parameter 1 with parameter 2
	 * 
	 * @param p1
	 *            Parameter 1
	 * @param index1
	 *            Index of parameter 1
	 * @param p2
	 *            Parameter 2
	 * @param index2
	 *            Index of parameter 2
	 * @param compareResults
	 *            Compare results
	 */
	public void compare(Parameter p1, int index1, Parameter p2, int index2, boolean compareResults) {

		this.compareResults = compareResults;

		this.diff.clear();

		if (p1.getRoot() instanceof Project) {
			this.source1 = ((Project) p1).getMopFile().toString();
		} else {
			this.source1 = "";
		}
		this.source1 += ":" + p1.getIdentifier() + ":" + index1;

		if (p2.getRoot() instanceof Project) {
			this.source2 = ((Project) p2).getMopFile().toString();
		} else {
			this.source2 = "";
		}
		this.source2 += ":" + p2.getIdentifier() + ":" + index2;

		this.compareParameter(p1, index1, p2, index2);
	}

	/**
	 * Get result of compare
	 * 
	 * @return List of different
	 */
	public CompareResult[] getResult() {
		return this.diff.toArray(new CompareResult[0]);
	}

	/**
	 * Get source of parameter 1
	 * 
	 * @return Source string
	 */
	public String getSource1() {
		return this.source1;
	}

	/**
	 * Get source of parameter 2
	 * 
	 * @return Source string
	 */
	public String getSource2() {
		return this.source2;
	}

	/**
	 * Compare parameter
	 * 
	 * @param p1
	 *            Parameter 1
	 * @param index1
	 *            Index of parameter 1
	 * @param p2
	 *            Parameter 2
	 * @param index2
	 *            Index of parameter 2
	 */
	private void compareParameter(Parameter p1, int index1, Parameter p2, int index2) {

		// Check names
		if (!p1.getName().equals(p2.getName())) {
			this.diff.add(new CompareResult(p1.getIdentifier(), "Different Name", p1.getName() + "->" + p2.getName()));
		}

		// check compareable types
		if (!p1.getClass().equals(p2.getClass())) {
			this.diff.add(new CompareResult(p1.getIdentifier(), "Different Type",
					p1.getClass().getName() + "->" + p2.getClass().getName()));
		}

		// Compare data
		if (this.compareResults || (!p1.getReadOnly() && (!(p1 instanceof Function)))) {
			if (!p1.getData(index1).equals(p2.getData(index2))) {
				String rel = "";
				TData data1, data2;
				data1 = p1.getData(index1);
				data2 = p2.getData(index2);
				if ((data1 instanceof TNumeric) && (data2 instanceof TNumeric)) {
					TNumeric num1, num2;
					num1 = (TNumeric) data1;
					num2 = (TNumeric) data2;
					if (num1.isScalar() && num1.isReal() && num2.isScalar() && num2.isReal()) {
						rel = " (" + ((((num2.getReal(0, 0) / num1.getReal(0, 0)) - 1.0) * 1e2)) + "% difference)";
					}
				}
				this.diff.add(new CompareResult(p1.getIdentifier(), "Different Value", data1 + "->" + data2 + rel));
			}
		}

		// compare fields
		if ((p1 instanceof ParamField) && (p2 instanceof ParamField)) {
			ParamField f1, f2;

			f1 = (ParamField) p1;
			f2 = (ParamField) p2;
			if (!f1.getUnit().equals(f2.getUnit())) {
				this.diff.add(
						new CompareResult(p1.getIdentifier(), "Different Unit", f1.getUnit() + "->" + f2.getUnit()));
			}
			if (!f1.getTexString().equals(f2.getTexString())) {
				this.diff.add(new CompareResult(p1.getIdentifier(), "Different Tex",
						f1.getTexString() + "->" + f2.getTexString()));
			}
		}

		// compare expression of formula
		if ((p1 instanceof ParamFormula) && (p2 instanceof ParamFormula)) {
			ParamFormula f1, f2;

			f1 = (ParamFormula) p1;
			f2 = (ParamFormula) p2;
			if (!f1.getExpression().equals(f2.getExpression())) {
				this.diff.add(new CompareResult(p1.getIdentifier(), "Different Expression",
						f1.getExpression() + "->" + f2.getExpression()));
			}
		}

		// compare functions
		if ((p1 instanceof ParamFunction) && (p2 instanceof ParamFunction)) {
			ParamFunction f1, f2;

			f1 = (ParamFunction) p1;
			f2 = (ParamFunction) p2;

			if (!f1.getClassPath().equals(f2.getClassPath())) {
				this.diff.add(new CompareResult(p1.getIdentifier(), "Different Classpath",
						f1.getClassPath() + "->" + f2.getClassPath()));
			}
			if (!f1.getClassName().equals(f2.getClassName())) {
				this.diff.add(new CompareResult(p1.getIdentifier(), "Different Classname",
						f1.getClassName() + "->" + f2.getClassName()));
			}
		}

		// compare imports/exports
		if ((p1 instanceof Dependency) && (p2 instanceof Dependency)) {
			Depend d1, d2;

			d1 = ((Dependency) p1).getDepend();
			d2 = ((Dependency) p2).getDepend();

			this.compareDepend("Imports", p1, d1.getImports(), d2.getImports());
			this.compareDepend("Exports", p1, d1.getExports(), d2.getExports());
		}

		// compare sub-parameter
		if ((p1 instanceof ParamContainer) && (p2 instanceof ParamContainer)) {
			this.compareContainer((ParamContainer) p1, index1, (ParamContainer) p2, index2);
		}

	}

	/**
	 * Compare sub-parameter of container
	 * 
	 * @param cont1
	 *            Container 1
	 * @param index1
	 *            Dataset-index of container 1
	 * @param cont2
	 *            Container 2
	 * @param index2
	 *            Dataset-index of container 2
	 */
	private void compareContainer(ParamContainer cont1, int index1, ParamContainer cont2, int index2) {
		Parameter[] params1, params2;
		ParameterPath path;

		params1 = cont1.getParam();
		params2 = cont2.getParam();

		// compare parameter-list of container 1
		for (Parameter p : params1) {
			path = cont2.getParamByUniqueIdentifier(cont2.getName() + "." + p.getName(), true);
			if (path == null) {
				this.diff.add(new CompareResult(p.getIdentifier(), "Exist only in 1", ""));
			} else {
				this.compareParameter(p, index1, path.getParameter(), index2);
			}
		}

		// compare parameter-list of container 2
		for (Parameter p : params2) {
			path = cont1.getParamByUniqueIdentifier(cont1.getName() + "." + p.getName(), true);
			if (path == null) {
				this.diff.add(new CompareResult(p.getIdentifier(), "Exist only in 2", ""));
			}
		}
	}

	/**
	 * Compare import/export lists
	 * 
	 * @param depend
	 *            Name of depend
	 * @param param
	 *            Parameter
	 * @param items1
	 *            List of depends 1
	 * @param items2
	 *            List of depends 2
	 */
	private void compareDepend(String depend, Parameter param, DependItem[] items1, DependItem[] items2) {
		boolean found;
		Parameter p1, p2;

		if (!(param instanceof ParamMaterial)) {
			for (DependItem i1 : items1) {
				found = false;
				for (DependItem i2 : items2) {
					if (i1.getVarName().equals(i2.getVarName())) {
						found = true;
						// if (!i1.getParam().equals(i2.getParam())) {
						// this.diff.add(new
						// CompareResult(param.getIdentifier(),
						// depend + " " + i1.getVarName()
						// + " with different parameter names", i1
						// .getParam() + "->" + i2.getParam()));
						// }

						p1 = i1.getHandle();
						p2 = i2.getHandle();
						if ((p1 == null) && (p2 == null)) {
						} else if (p1 == null) {
							this.diff.add(new CompareResult(param.getIdentifier(),
									depend + " " + i1.getVarName() + " is only assigned in 2", ""));
						} else if (p2 == null) {
							this.diff.add(new CompareResult(param.getIdentifier(),
									depend + " " + i1.getVarName() + " is only assigned in 1", ""));
						} else if (!this.equalIdentifier(p1.getIdentifier(), p2.getIdentifier())) {
							this.diff.add(new CompareResult(param.getIdentifier(),
									depend + " " + i1.getVarName() + " with different parameter",
									p1.getIdentifier() + "->" + p2.getIdentifier()));
						}
					}
				}
				if (!found) {
					this.diff.add(
							new CompareResult(param.getIdentifier(), depend + " exist only in 1", i1.getVarName()));
				}
			}

			for (DependItem i2 : items2) {
				found = false;
				for (DependItem i1 : items1) {
					if (i1.getVarName().equals(i2.getVarName())) {
						found = true;
					}
				}
				if (!found) {
					this.diff.add(
							new CompareResult(param.getIdentifier(), depend + " exist only in 2", i2.getVarName()));
				}
			}
		}
	}

	/**
	 * Are Identifier equal except root-name
	 * 
	 * @param ident1
	 *            Identifier 1
	 * @param ident2
	 *            Identifier 2
	 * @return Are Identifier equal
	 */
	private boolean equalIdentifier(String ident1, String ident2) {
		if (ident1.contains(".") && ident2.contains(".")) {
			ident1 = ident1.substring(ident1.indexOf("."));
			ident2 = ident2.substring(ident2.indexOf("."));
		}
		return ident1.equals(ident2);
	}
}
