/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_25.
 * 
 * - Rename dataset-directories <br>
 * - ReRename parameter-directories
 * 
 * @author Werner Koppelsttter
 */
public class Converter_25 {

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @param file
	 *            Project file
	 * @return Changed data node
	 */
	static XMLNode convert(XMLNode node, File file) {
		File dir;

		dir = new File(file.getPath() + ".Files");
		Converter_25.convertDSDirectories(node, dir);
		Converter_25.convertParamDirectories(node, dir);

		return node;
	}

	/**
	 * Rename dataset directories and remove files-entry
	 * 
	 * @param node
	 *            Node to update
	 * @param dir
	 *            Files-direcotry
	 */
	private static void convertDSDirectories(XMLNode node, File dir) {
		XMLNode dsid, files;
		File oldDir, newDir;

		dsid = node.getNode(6);
		files = node.getNode(8);

		if ((dsid.getName().compareTo("DSID") == 0)
				&& (files.getName().compareTo("FilesDir") == 0)) {
			if (dir.exists()) {
				for (int k = 0; k < files.getNodeSize(); k++) {
					oldDir = new File(dir, files.getNode(k).getText());
					newDir = new File(dir, "_" + dsid.getNode(k).getText());
					if (oldDir.exists()) {
						oldDir.renameTo(newDir);
					}
				}
				for (int k = 0; k < files.getNodeSize(); k++) {
					oldDir = new File(dir, "_" + dsid.getNode(k).getText());
					newDir = new File(dir, dsid.getNode(k).getText());
					if (oldDir.exists()) {
						oldDir.renameTo(newDir);
					}
				}
			}
			node.rmNode(8);
		}
	}

	/**
	 * Rename parameter directories
	 * 
	 * @param node
	 *            Node to update
	 * @param dir
	 *            Files-direcotry
	 */
	private static void convertParamDirectories(XMLNode node, File dir) {
		XMLNode dsid;
		Hashtable<String, String> map;

		// collect pid
		map = new Hashtable<String, String>();
		Converter_25.collectPID(node, "", map);

		// rename directory
		dsid = node.getNode(6);
		if (dsid.getName().compareTo("DSID") == 0) {
			dsid = node.getNode(6);
			if (dir.exists()) {
				for (int k = 0; k < dsid.getNodeSize(); k++) {
					Converter_25.convertDirectories(
							new File(dir, dsid.getNode(k).getText()), map);
				}
			}
		}
	}

	/**
	 * Collect parameter-id of all parameter
	 * 
	 * @param node
	 *            Node to search
	 * @param path
	 *            Path in structure
	 * @param map
	 *            Map of parameter-id
	 */
	private static void collectPID(XMLNode node, String path,
			Hashtable<String, String> map) {
		String name, pid;
		int k = 0;

		if (node.getName().compareTo("Parameter") == 0) {
			// add parameter
			pid = node.getNode(0).getText();
			name = node.getNode(1).getText();
			map.put(path + "." + name, pid);
		} else if (node.getName().compareTo("ParamContainer") == 0) {
			// add container
			name = node.getNode(0).getNode(1).getText();
			Converter_25.collectPID(node.getNode(0), path, map);
			path += "." + name;
			k = 1;
		}

		// next Node
		for (; k < node.getNodeSize(); k++) {
			Converter_25.collectPID(node.getNode(k), path, map);
		}
	}

	/**
	 * Rename parameter directories
	 * 
	 * @param dir
	 *            Dataset-directory to convert
	 * @param map
	 *            Map with parameter-id
	 */
	private static void convertDirectories(File dir,
			Hashtable<String, String> map) {
		String pid;
		File subdir, paramdir;

		if (dir.exists()) {
			paramdir = new File(dir, "param");

			// ParameterID directory
			for (File f : dir.listFiles()) {
				if (f.isDirectory()) {
					if (map.containsValue(f.getName())) {
						try {
							FileUtils.move(f, new File(paramdir, f.getName()));
						} catch (IOException e) {
						}
					}
				}
			}

			// Femag directory
			for (File f : dir.listFiles()) {
				if (f.isDirectory()) {
					pid = map.get("." + f.getName());
					if (pid != null) {
						try {
							FileUtils.move(f, new File(paramdir, pid));
						} catch (IOException e) {
						}
					}
				}
			}

			// LTSpice directory
			subdir = new File(dir, "LTSpice");
			if (subdir.exists()) {
				for (File f : subdir.listFiles()) {
					if (f.isDirectory()) {
						pid = map.get("." + f.getName());
						if (pid != null) {
							try {
								FileUtils.move(f, new File(paramdir, pid));
							} catch (IOException e) {
							}
						}
					}
				}
				FileUtils.delete(subdir);
			}

			// Femm42 directory
			subdir = new File(dir, "femm42");
			if (subdir.exists()) {
				for (File f : subdir.listFiles()) {
					if (f.isDirectory()) {
						pid = map.get("." + f.getName());
						if (pid != null) {
							try {
								FileUtils.move(f, new File(paramdir, pid));
							} catch (IOException e) {
							}
						}
					}
				}
				FileUtils.delete(subdir);
			}

		}
	}
}
