/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.event;

import java.io.Serializable;
import java.rmi.Remote;
import java.rmi.RemoteException;

/**
 * Interface of event-log
 * 
 * @author Werner Koppelsttter
 */
public interface IEventLog extends Remote, Serializable, Cloneable {

	/**
	 * Add a event-object to the event-buffer
	 * 
	 * @param evObj
	 *            Event to add
	 * @throws RemoteException
	 */
	public void add(Event evObj) throws RemoteException;

	/**
	 * Clear event-buffer
	 * 
	 * @throws RemoteException
	 */
	public void clear() throws RemoteException;

	/**
	 * Get the event-buffer of the event-log
	 * 
	 * @param info
	 *            Get info-events
	 * @param warning
	 *            Get warning-events
	 * @param error
	 *            Get error-events
	 * 
	 * @return Event-array
	 * @throws RemoteException
	 */
	public Event[] getBuffer(boolean info, boolean warning, boolean error)
			throws RemoteException;

	/**
	 * Get the maximal Size of the event-buffer
	 * 
	 * @return Maximal Size
	 * @throws RemoteException
	 */
	public int getMaxSize() throws RemoteException;

	/**
	 * Set the maximal size of the event-buffer
	 * 
	 * @param size
	 *            Maximal size
	 * @throws RemoteException
	 */
	public void setMaxSize(int size) throws RemoteException;

	/**
	 * Insert info-event with calling-stack to event-log
	 * 
	 * @param identifier
	 *            Identifier where the event occurs
	 * @param message
	 *            Description of the event
	 * @param stack
	 *            Calling stack
	 * @throws RemoteException
	 */
	public void info(String identifier, String message,
			StackTraceElement[] stack) throws RemoteException;

	/**
	 * Insert warning-event with calling-stack to event-log
	 * 
	 * @param identifier
	 *            Identifier where the event occurs
	 * @param message
	 *            Description of the event
	 * @param stack
	 *            Calling stack
	 * @throws RemoteException
	 */
	public void warning(String identifier, String message,
			StackTraceElement[] stack) throws RemoteException;

	/**
	 * Insert error-event with calling-stack to event-log
	 * 
	 * @param identifier
	 *            Identifier where the event occurs
	 * @param message
	 *            Description of the event
	 * @param stack
	 *            Calling stack
	 * @throws RemoteException
	 */
	public void error(String identifier, String message,
			StackTraceElement[] stack) throws RemoteException;

	/**
	 * Create a full string representation
	 * 
	 * @return String representation
	 * @throws RemoteException
	 */
	public String toFullString() throws RemoteException;

	/**
	 * To string for interface
	 * 
	 * @return String representation of string
	 * @throws RemoteException
	 */
	public String toStringI() throws RemoteException;

	/**
	 * Get client-implementation of interface
	 * 
	 * @return Client-implementation
	 * @throws RemoteException
	 * @throws RemoteException
	 */
	public IEventLog getClientImpl() throws RemoteException;

}
