/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure.ipython;

import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Collections;

import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

/**
 * 
 * 
 *
 * @author gweidenholzer
 */
public class ConnectionFile implements Comparable<ConnectionFile> {

	File file;

	Long shell_port;
	Long stdin_port;
	Long iopub_port;
	Long control_port;
	String transport;
	String ip;
	String keyString;

	/**
	 * Create connection file
	 * 
	 * @param f Connection file
	 * @throws Exception
	 */
	public ConnectionFile(File f) throws Exception {

		this.file = f;

		// read kernel connection file
		JSONParser parser = new JSONParser();
		FileReader cf_reader = new FileReader(f);
		JSONObject cf = (JSONObject) parser.parse(cf_reader);
		cf_reader.close();

		// get connection data
		this.keyString = (String) cf.get("key");
		this.transport = (String) cf.get("transport");
		this.ip = (String) cf.get("ip");
		this.shell_port = (Long) cf.get("shell_port");
		this.iopub_port = (Long) cf.get("iopub_port");
		this.stdin_port = (Long) cf.get("stdin_port");
		this.control_port = (Long) cf.get("control_port");

	}

	/**
	 * Get represented string
	 * 
	 * @return Connection id
	 */
	public String getID() {
		String id;

		id = this.file.getName();
		id = id.substring(7, id.length() - 5);

		return id;
	}

	/**
	 * Get connection file
	 * 
	 * @return File
	 */
	public File getFile() {
		return this.file;
	}

	/**
	 * String representation
	 */
	@Override
	public String toString() {
		return "Kernel connection file id: " + this.getID();

	}

	/**
	 * compare creation dates
	 */
	@Override
	public int compareTo(ConnectionFile o) {
		int comp = 0;
		if (o.file.lastModified() != this.file.lastModified()) {
			if (o.file.lastModified() > this.file.lastModified()) {
				comp = 1;
			} else {
				comp = -1;
			}
		}

		return comp;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ConnectionFile other = (ConnectionFile) obj;
		if (this.file == null) {
			if (other.file != null) {
				return false;
			}
		} else if (!this.file.equals(other.file)) {
			return false;
		}
		return true;
	}

	/**
	 * return a list of connection files of active kernels
	 * 
	 * @throws Exception
	 */
	public static ArrayList<ConnectionFile> findConnectionFiles() throws Exception {
		File[] folders;
		FilenameFilter filter;
		File userdir;
		ArrayList<ConnectionFile> list = new ArrayList<ConnectionFile>();

		filter = new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return name.toLowerCase().matches("kernel-[\\w-]+\\.json");
			}
		};

		userdir = new File(System.getProperty("user.home"));
		folders = new File[] { new File(userdir, ".ipython\\profile_default\\security"),
				new File(userdir, "AppData\\Roaming\\jupyter\\runtime") };

		// get files
		for (File folder : folders) {
			if (folder.exists()) {
				for (File f : folder.listFiles(filter)) {
					list.add(new ConnectionFile(f));
				}
			}
		}

		Collections.sort(list);

		return list;
	}
}
