/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure.python;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;

import javax.xml.stream.XMLStreamException;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import net.razorvine.pyro.PyroException;

/**
 * Interface to Python
 * 
 * @author Werner Koppelsttter
 */
public class PythonConnection {

	/** hmac-key for connection */
	private static String hMacKey = "MagOptServer";

	/** remote python object (console) */
	private static PythonProxy remoteObject = null;

	private static boolean isAlive(PythonProxy rObj) {
		boolean alive = false;

		if (rObj != null) {
			try {
				alive = ((String) rObj.call("echo", "abc")).equals("abc");
			} catch (Exception e) {
			}
		}

		return alive;
	}

	private static PythonProxy getConnection() throws Exception {
		int port;
		LinkedList<String> cmd;
		ProcessBuilder pb;
		File pythonServer, generalPath;
		boolean console = false;
		String python;
		int count = 5;

		try {
			port = Integer.parseInt(MagOptConfig.getConfig("PythonServerPort"));
		} catch (Exception ex) {
			port = 7845;
			ex.printStackTrace();
		}
		try {
			console = Boolean.parseBoolean(MagOptConfig.getConfig("PythonConsole"));
		} catch (Exception ex) {
		}

		if ((PythonConnection.remoteObject == null) || (PythonConnection.remoteObject.getPort() != port)
				|| !PythonConnection.isAlive(PythonConnection.remoteObject)) {
			try {
				PythonConnection.remoteObject = new PythonProxy("localhost", port, "magopt.pythonserver");
				PythonConnection.remoteObject.setHmacKey(PythonConnection.hMacKey.getBytes("utf-8"));
			} catch (Exception e) {
				PythonConnection.remoteObject = null;
				e.printStackTrace();
			}
		}

		while ((count > 0) && (PythonConnection.remoteObject != null)
				&& !PythonConnection.isAlive(PythonConnection.remoteObject)) {

			try {
				pythonServer = FileUtils.trimPath(FileUtils.getFile(
						PythonConnection.class.getProtectionDomain().getCodeSource().getLocation().getFile(), "..",
						"..", "Python", "magoptserver.py"));
				generalPath = FileUtils.trimPath(
						FileUtils.getFile(TData.class.getProtectionDomain().getCodeSource().getLocation().getFile(),
								"..", "..", "Python"));

				cmd = new LinkedList<String>();
				if (console) {
					cmd.add("cmd");
					cmd.add("/c");
					cmd.add("start");
					cmd.add("/wait");
					cmd.add("Python connection"); // name of window
					cmd.add("cmd");
					cmd.add("/c");
				}
				python = MagOptConfig.getConfig("PythonPath");
				if (python == null) {
					python = "";
				} else {
					python += File.separator;
				}
				python += "python.exe";
				cmd.add(python);
				cmd.add(pythonServer.getPath());

				pb = new ProcessBuilder(cmd);
				pb.directory(PythonConnection.getWorkingDirectory());

				// set environment
				pb.environment().put("PYTHONPATH", generalPath.getPath());
				pb.environment().put("PythonServerPort", Integer.toString(port));
				pb.environment().put("PythonHMacKey", PythonConnection.hMacKey);

				pb.start();
				Thread.sleep(500);
				count--;
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		if ((PythonConnection.remoteObject == null) || !PythonConnection.isAlive(PythonConnection.remoteObject)) {
			throw new Exception("Unable to open Python connection!");
		}

		return PythonConnection.remoteObject;
	}

	/**
	 * Get working directory of PreeCADServer
	 * 
	 * @return Working directory
	 */
	public static File getWorkingDirectory() {
		File wd;

		wd = new File(System.getProperty("java.io.tmpdir"), "PythonServer");
		if (!wd.exists()) {
			wd.mkdirs();
		}

		return wd;
	}

	/**
	 * Connect to application (open id no instance is running)
	 * 
	 * @throws Exception
	 */
	public static void connect() throws Exception {

		PythonConnection.getConnection();
	}

	/**
	 * Disconnect application
	 * 
	 * @throws Exception
	 */
	public static void disconnect() throws Exception {
		PythonProxy conn;

		if (PythonConnection.isAlive(PythonConnection.remoteObject)) {
			conn = PythonConnection.getConnection();
			try {
				conn.call("shutdown");
			} catch (Exception e) {
				e.printStackTrace();
			}
			PythonConnection.remoteObject = null;
		}
	}

	public static boolean isValidScript(File pyFile) throws Exception {
		PythonProxy conn;
		boolean valid = false;

		conn = PythonConnection.getConnection();
		try {
			valid = (Boolean) conn.call("isValidScript", pyFile.getPath());
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return valid;
	}

	public static boolean hasHtml(File pyFile) throws Exception {
		PythonProxy conn;
		boolean valid = false;

		conn = PythonConnection.getConnection();
		try {
			valid = (Boolean) conn.call("hasHtml", pyFile.getPath());
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return valid;
	}

	public static String getFunctionID(File pyFile) throws Exception {
		PythonProxy conn;
		String id;

		conn = PythonConnection.getConnection();
		try {
			id = (String) conn.call("functionID", pyFile.getPath());
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return id;
	}

	@SuppressWarnings("unchecked")
	public static ArrayList<String> imports(File pyFile) throws Exception {
		PythonProxy conn;
		ArrayList<String> list;

		conn = PythonConnection.getConnection();
		try {
			list = (ArrayList<String>) conn.call("imports", pyFile.getPath());
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return list;
	}

	@SuppressWarnings("unchecked")
	public static ArrayList<String> exports(File pyFile) throws Exception {
		PythonProxy conn;
		ArrayList<String> list;

		conn = PythonConnection.getConnection();
		try {
			list = (ArrayList<String>) conn.call("exports", pyFile.getPath());
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return list;
	}

	private static HashMap<String, Object> convertImports(Map<String, TData> map)
			throws XMLStreamException, IOException {
		HashMap<String, Object> dict;
		ByteArrayOutputStream bos;

		dict = new HashMap<String, Object>();
		dict.put("__class__", "magopt.tdata");

		bos = new ByteArrayOutputStream();
		TData.writeMap(bos, map, false);
		bos.close();
		dict.put("xml", bos.toString("UTF-8"));

		return dict;
	}

	@SuppressWarnings("unchecked")
	private static Map<String, TData> convertExports(HashMap<String, Object> dict)
			throws TDataException, XMLStreamException, IOException {
		Map<String, TData> map;
		ByteArrayInputStream bis;
		String xml;

		// FIXME Python dictornary
		map = new Hashtable<String, TData>();
		for (String key : dict.keySet()) {
			xml = (String) ((HashMap<String, Object>) dict.get(key)).get("xml");
			bis = new ByteArrayInputStream(xml.getBytes("UTF-8"));
			map.put(key, TData.read(bis));
			bis.close();
		}

		// xml = (String) dict.get("xml");
		// bis = new ByteArrayInputStream(xml.getBytes("UTF-8"));
		// map = TData.readMap(bis);
		// bis.close();

		return map;
	}

	@SuppressWarnings("unchecked")
	public static Map<String, TData> update(File pyFile, Map<String, TData> map) throws Exception {
		PythonProxy conn;
		HashMap<String, Object> res;
		Map<String, TData> exp;

		conn = PythonConnection.getConnection();
		try {
			res = (HashMap<String, Object>) conn.call("update", pyFile.getPath(), PythonConnection.convertImports(map));
			exp = PythonConnection.convertExports(res);
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}

		return exp;
	}

	public static void html(File pyFile, File folder, Map<String, TData> map) throws Exception {
		PythonProxy conn;

		conn = PythonConnection.getConnection();
		try {
			conn.call("html", pyFile.getPath(), folder.getPath(), PythonConnection.convertImports(map));
		} catch (PyroException ex) {
			PythonConnection.modifyPyroException(ex);
			throw ex;
		}
	}

	/**
	 * Modify StackTrace list from PyroException (add python stack)
	 * 
	 * @param ex Pyro-exception
	 * @throws PyroException
	 */
	private static void modifyPyroException(PyroException ex) {
		LinkedList<StackTraceElement> stack;
		String[] stackline, elems;
		File file;
		String str, methode, cl;
		int line;
		int k = 1;
		StackTraceElement st;

		stack = new LinkedList<StackTraceElement>(Arrays.asList(ex.getStackTrace()));
		stackline = ex._pyroTraceback.split("\n");
		while ((k < stackline.length) && stackline[k].trim().startsWith("File")) {
			elems = stackline[k].split(",");

			// file
			str = elems[0].trim();
			str = str.substring(6, str.length() - 1);
			file = new File(str);
			// class
			cl = file.getName();
			cl = cl.substring(0, cl.lastIndexOf("."));

			// line
			str = elems[1].trim();
			str = str.substring(5);
			line = Integer.parseInt(str);

			// methode
			if (elems.length > 2) {
				str = elems[2].trim();
				methode = str.substring(3);
			} else {
				methode = "";
			}

			st = new StackTraceElement(cl, methode, file.getName(), line);
			stack.addFirst(st);
			k += 2;
		}
		ex.setStackTrace(stack.toArray(new StackTraceElement[0]));
	}
}
