/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10127 $
 * $Date: 2021-03-09 13:18:45 +0100 (Tue, 09 Mar 2021) $
 */
package magopt.gui.console;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.magopt.structure.jep.JepProxy;
import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Display input, output and error of python console
 *
 * @author Werner Koppelsttter
 */
public class PythonConsole extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Console */
	private Console console;

	/** Python connection */
	private JepProxy jep;

	/** Command line */
	private JTextField cl;

	/** Commend history */
	private LinkedList<String> history = new LinkedList<String>();

	/** Index of selected history-command */
	private int historyIndex = 0;

	/** Current directory for run-script-dialog */
	private File scriptDirectory = null;

	/** Command before tabulatuor was used */
	private String tabText = "";

	/**
	 * Get initalCommands executed at startup
	 * 
	 * @return List of commands
	 * @throws Exception
	 */
	public static List<String> initCommands() throws Exception {
		LinkedList<String> commands;

		commands = new LinkedList<String>();

		commands.add("from symspace.wrapper.symwrapper import SymProject, SymContainer, "
				+ "SymField, SymFormula, SymComboBox, SymLinkedField, " + "SymMaterial, SymDirectory, SymFile, "
				+ "SymJavaFunction, SymPythonFunction, SymMatlabFunction");
		commands.add("from symspace.wrapper.symplot import SymPlot2Bar, SymPlot2Hist, SymPlot2Line");
		commands.add("from symspace.wrapper.complete import completeTab");

		// eval preamble
		String path = MagOptConfig.getConfig("PythonConsolePreamble");
		if (path != null) {
			File file = new File(path);
			if (file.exists() && file.isFile()) {
				commands.add("exec(open(r'" + file.getPath() + "').read())");
			}
		}

		return commands;
	}

	/**
	 * Create console to print input, output and errors of python
	 */
	public PythonConsole() {
		this.setLayout(new BorderLayout());

		// console
		this.console = new Console();
		this.add(this.console, BorderLayout.CENTER);

		JMenuItem item = new JMenuItem("Restart python");
		this.console.popup.add(item);
		item.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				PythonConsole.this.console.getOutputPrintStream().println("Python restart");
				PythonConsole.this.restart();
			}
		});

		item = new JMenuItem("Run script");
		this.console.popup.add(item);
		item.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;
				File dir;

				dir = PythonConsole.this.scriptDirectory;
				if ((dir == null) || !dir.exists()) {
					String path;
					try {
						path = (String) PythonConsole.this.jep.evalGet("_cwd_", "_cwd_ = os.getcwd()");
						PythonConsole.this.jep.eval("del _cwd_");
						dir = new File(path);
					} catch (Exception e1) {
						dir = new File(MagOptConfig.getHome());
					}
				}

				chooser = new JFileChooser(dir);
				chooser.setFileFilter(new FileNameExtensionFilter("Python script (*.py)", "py"));
				chooser.setDialogTitle("Select python script");
				if (chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
					PythonConsole.this.runScript(chooser.getSelectedFile());
				}

			}
		});

		// commandline
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.setBorder(new LineBorder(Color.GRAY));
		this.add(panel, BorderLayout.SOUTH);

		this.cl = new JTextField();
		this.cl.setEnabled(false);
		this.cl.setBorder(new EmptyBorder(5, 5, 5, 5));
		this.cl.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent action) {
				String txt = PythonConsole.this.cl.getText();
				PythonConsole.this.cl.setText("");
				PythonConsole.this.eval(txt);
			}

		});
		this.cl.addKeyListener(new KeyAdapter() {

			@Override
			public void keyPressed(KeyEvent e) {
				switch (e.getKeyCode()) {
				case KeyEvent.VK_UP:
					PythonConsole.this.historyIndex--;
					if (PythonConsole.this.historyIndex < 0) {
						PythonConsole.this.historyIndex = 0;
					} else {
						PythonConsole.this.cl.setText(PythonConsole.this.history.get(PythonConsole.this.historyIndex));
					}
					PythonConsole.this.tabText = "";
					break;
				case KeyEvent.VK_DOWN:
					PythonConsole.this.historyIndex++;
					if (PythonConsole.this.historyIndex < PythonConsole.this.history.size()) {
						PythonConsole.this.cl.setText(PythonConsole.this.history.get(PythonConsole.this.historyIndex));
					} else {
						PythonConsole.this.historyIndex = PythonConsole.this.history.size() - 1;
					}
					PythonConsole.this.tabText = "";
					break;
				case KeyEvent.VK_ESCAPE:
					PythonConsole.this.historyIndex = PythonConsole.this.history.size();
					PythonConsole.this.cl.setText("");
					PythonConsole.this.tabText = "";
					break;
				case KeyEvent.VK_TAB:
					if (PythonConsole.this.tabText.isEmpty()) {
						PythonConsole.this.tabText = PythonConsole.this.cl.getText();
					}
					try {
						PythonConsole.this.cl.setText((String) PythonConsole.this.jep.invoke("completeTab",
								PythonConsole.this.tabText, !e.isShiftDown()));
					} catch (Exception e1) {
					}

					e.consume();
					break;
				case KeyEvent.VK_SHIFT:
					break;
				default:
					PythonConsole.this.tabText = "";
				}
			}
		});
		this.cl.setFocusTraversalKeysEnabled(false);
		panel.add(this.cl, BorderLayout.CENTER);

		// python connection
		this.restart();

	}

	/**
	 * Get console
	 * 
	 * @return Console
	 */
	public Console getConsole() {
		return this.console;
	}

	/**
	 * Get python connection
	 * 
	 * @return Jep
	 */
	public JepProxy getJep() {
		return this.jep;
	}

	/**
	 * Clear console
	 */
	public void clear() {
		this.console.clear();
	}

	/**
	 * Restart python console
	 */
	public void restart() {
		this.restart(null);
	}

	/**
	 * Restart python console
	 * 
	 * @param jep python connection, if is null then create new one
	 */
	public void restart(final JepProxy jep) {
		this.cl.setEnabled(false);

//		new Thread() {
//			@Override
//			public void run() {
		try {
			if (jep == null) {
				PythonConsole.this.jep = new JepProxy();
			} else {
				PythonConsole.this.jep = jep;
			}

			PythonConsole.this.jep.redirectStream(PythonConsole.this.console.getOutputPrintStream(),
					PythonConsole.this.console.getErrorPrintStream());

			PythonConsole.this.jep.setInteractive(true);

			for (String command : PythonConsole.initCommands()) {
				PythonConsole.this.jep.eval(command);
			}

			PythonConsole.this.cl.setEnabled(true);
		} catch (Throwable e) {
			this.console.writeError(e.getMessage());
			e.printStackTrace(this.console.getErrorPrintStream());
		}
//	}
//		}.start();
	}

	/**
	 * Run script-file
	 * 
	 * @param file Python script
	 */
	public void runScript(final File file) {
		if ((file != null) && file.exists()) {
			PythonConsole.this.console.writeInput("Run script: " + file.getName() + " (" + file.getParent() + ")\n");
			new Thread() {
				@Override
				public void run() {
					try {
						PythonConsole.this.jep.eval("exec(open(r'" + file.getAbsolutePath() + "').read())");
					} catch (Exception e) {
						PythonConsole.this.console.getErrorPrintStream().println(e.getMessage());
						e.printStackTrace(PythonConsole.this.console.getErrorPrintStream());
					}
				}
			}.start();
		}
	}

	/**
	 * Eval command
	 * 
	 * @param txt Command
	 */
	public void eval(final String txt) {
		if (this.jep != null) {
			new Thread() {
				@Override
				public void run() {
					try {
						PythonConsole.this.console.writeInput(txt + "\n");
						if (PythonConsole.this.history.isEmpty() || !PythonConsole.this.history.getLast().equals(txt)) {
							PythonConsole.this.history.add(txt);
						}
						PythonConsole.this.historyIndex = PythonConsole.this.history.size();
						PythonConsole.this.jep.eval(txt);
					} catch (Exception e) {
						PythonConsole.this.console.writeError(e.getMessage());
						e.printStackTrace(PythonConsole.this.console.getErrorPrintStream());
					}
				}
			}.start();
		}
	}

	/**
	 * Set Current directory for run-script-dialog
	 * 
	 * @param dir Script directory
	 */
	public void setScriptDirectory(File dir) {
		this.scriptDirectory = dir;
	}
}
