/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9749 $
 * $Date: 2020-04-10 16:00:50 +0200 (Fri, 10 Apr 2020) $
 */
package magopt.gui.view.summary;

import java.rmi.RemoteException;

import javax.swing.JTable;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.magopt.structure.IParamComboBox;
import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.IParamField;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.Parameter;

/**
 * Table model for container summary table.
 * 
 * @author Aigner Markus
 * @author Werner Koppelsttter
 */
public class SummaryModel extends AbstractTableModel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * List of parameters
	 */
	private IParamContainer container = null;

	/**
	 * Set items to display depends
	 * 
	 * @param container
	 *            ParamContainer
	 */
	public void setContainer(IParamContainer container) {
		this.container = container;

		// refresh table of model
		for (TableModelListener listener : this.getListeners(TableModelListener.class)) {
			if ((listener instanceof JTable) && this.equals(((JTable) listener).getModel())) {
				listener.tableChanged(new TableModelEvent(this));
			}
		}
	}

	@Override
	public int getColumnCount() {
		return SummaryColumn.values().length;
	}

	@Override
	public int getRowCount() {
		int len = 0;

		if (this.container != null) {
			try {
				len = this.container.getParam().length;
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		return len;
	}

	@Override
	public Object getValueAt(int row, int column) {
		Object obj = null;
		IParameter param;

		if (this.container != null) {
			try {
				param = this.container.getParam()[row];
				if (column == SummaryColumn.NAME.getValue()) {
					obj = param.getName();
				} else if (column == SummaryColumn.DESCRIPTION.getValue()) {
					obj = param.getDescription();
				} else if (column == SummaryColumn.VALUE.getValue()) {
					if (param instanceof IParamField) {
						obj = param.getData().toString();
					} else {
						obj = "";
					}
				} else if (column == SummaryColumn.UNIT.getValue()) {
					if (param instanceof IParamField) {
						obj = ((IParamField) param).getUnit();
					} else {
						obj = "";
					}
				} else {
					obj = param;
				}
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		return obj;
	}

	@Override
	public String getColumnName(int column) {
		for (SummaryColumn e : SummaryColumn.values()) {
			if (e.getValue() == column) {
				return e.getName();
			}
		}
		return null;
	}

	@Override
	public boolean isCellEditable(int row, int column) {
		boolean edit = false;
		IParameter param;
		TData curData;

		if (this.container != null) {
			try {
				param = this.container.getParam()[row];

				if (column == SummaryColumn.VALUE.getValue()) {
					if ((param instanceof IParamField) && !(param instanceof IParamComboBox) && !param.getReadOnly()) {
						curData = param.getData();
						if ((curData instanceof TNumeric) || (curData instanceof TString)
								|| (curData instanceof TBoolean) || (curData instanceof TNull)) {
							edit = true;
						}
					}
				} else if (column == SummaryColumn.DESCRIPTION.getValue()) {
					edit = true;
				} else if (column == SummaryColumn.UNIT.getValue()) {
					if (param instanceof IParamField) {
						edit = true;
					}
				}
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		return edit;
	}

	@Override
	public Class<?> getColumnClass(int column) {
		Class<?> type = String.class;

		if (column == SummaryColumn.TYPE.getValue()) {
			// TODO RMI
			type = Parameter.class;
		}

		return type;
	}

	@Override
	public void setValueAt(Object value, int row, int col) {
		String str;
		IParameter param;
		TData data = null;

		if (this.container != null) {
			try {
				param = this.container.getParam()[row];

				if (col == SummaryColumn.VALUE.getValue()) {
					// FIXME parse string
					str = (String) value;
					str.trim();
					if (str.isEmpty() || (str.startsWith("[") && str.endsWith("]")
							&& str.substring(1, str.length() - 1).trim().isEmpty())) {
						data = new TNull();
					} else if (str.startsWith("'") && str.endsWith("'")) {
						data = new TString(str.substring(1, str.length() - 1));
					} else if ((str.compareToIgnoreCase("false") == 0) || (str.compareToIgnoreCase("true") == 0)) {
						data = new TBoolean(Boolean.parseBoolean(str));
					} else {
						try {
							data = new TNumeric(str);
						} catch (TDataException e) {
						}
					}

					if (data != null) {
						try {
							param.setData(data);
							this.fireTableDataChanged();
						} catch (Exception e) {
							param.warning("Unable to set data to: " + str);
						}
					} else {
						param.warning("Unable to set data to: " + str);
					}
				} else if (col == SummaryColumn.DESCRIPTION.getValue()) {
					param.setDescription((String) value);
					this.fireTableDataChanged();
				} else if (col == SummaryColumn.UNIT.getValue()) {
					if (param instanceof IParamField) {
						((IParamField) param).setUnit((String) value);
						this.fireTableDataChanged();
					}
				}
			} catch (RemoteException e1) {
				e1.printStackTrace();
			} catch (Exception ex) {
				try {
					this.container.error(ex.getMessage(), ex.getStackTrace());
				} catch (RemoteException e) {
					e.printStackTrace();
				}
			}
		}
	}

}
