/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.LinkedList;

/**
 * Convert PS-graphic to remove field-lines and change font
 * 
 * @author Werner Koppelsttter
 */
public class PSConverter {

	/** Size of the buffer */
	static final int BUFFER_SIZE = 10;

	/** System new-line characters */
	static final String NEW_LINE = System.getProperty("line.separator");

	/**
	 * Start converter from command-line
	 * 
	 * @param args
	 *            String arguments
	 */
	public static void main(String[] args) {
		boolean rmFieldlines = true;

		if (args.length > 1) {
			try {
				rmFieldlines = Boolean.parseBoolean(args[0]);
				args = Arrays.copyOfRange(args, 1, args.length);
			} catch (Exception e) {
			}
		}
		new PSConverter().convert(rmFieldlines, args);
	}

	/**
	 * Convert a list of files or subfiles of directories
	 * 
	 * @param files
	 *            Files/Directories to convert
	 */
	public void convert(String... files) {
		this.convert(true, files);
	}

	/**
	 * Convert a list of files or subfiles of directories
	 * 
	 * @param rmFieldlines
	 *            Remove field-lines
	 * @param files
	 *            Files/Directories to convert
	 */
	public void convert(boolean rmFieldlines, String... files) {
		for (String file : files) {
			this.convert(rmFieldlines, new File(file));
		}
	}

	/**
	 * Convert a list of files or subfiles of directories
	 * 
	 * @param files
	 *            Files/Directories to convert
	 */
	public void convert(File... files) {
		this.convert(true, files);
	}

	/**
	 * Convert a list of files or subfiles of directories
	 * 
	 * @param rmFieldlines
	 *            Remove field-lines
	 * @param files
	 *            Files/Directories to convert
	 */
	public void convert(boolean rmFieldlines, File... files) {
		for (File file : files) {
			if (file.exists()) {
				if (file.isDirectory()) {
					for (File f1 : file.listFiles()) {
						this.convert(rmFieldlines, f1);
					}
				} else {
					this.convertFile(file, rmFieldlines);
				}
			}
		}
	}

	/**
	 * Convert ps-file
	 * 
	 * @param file
	 *            File to convert
	 * @param rmFieldlines
	 *            Remove field-lines
	 */
	protected void convertFile(File file, boolean rmFieldlines) {
		File tmp;
		FileInputStream in;
		FileOutputStream out;

		if (file.exists() && file.isFile()
				&& file.getName().toLowerCase().endsWith(".ps")) {
			tmp = new File(file.getAbsolutePath() + ".tmp");
			file.renameTo(tmp);

			try {
				in = new FileInputStream(tmp);
				out = new FileOutputStream(file);
				this.convertStream(in, out, rmFieldlines);
				in.close();
				out.close();
			} catch (IOException e) {
			}

			if (file.exists()) {
				tmp.delete();
			} else {
				tmp.renameTo(file);
			}
		}
	}

	/**
	 * Convert file from inputstream to outputstream
	 * 
	 * @param inStream
	 *            Stream to read file
	 * @param outStream
	 *            Stream to write file
	 * @param rmFieldlines
	 *            Remove field-lines
	 * @throws IOException
	 */
	protected void convertStream(InputStream inStream, OutputStream outStream,
			boolean rmFieldlines) throws IOException {
		BufferedReader in;
		BufferedWriter out;
		LinkedList<String> buffer;
		boolean isFieldlines = false;
		String line;

		in = new BufferedReader(new InputStreamReader(inStream));
		out = new BufferedWriter(new OutputStreamWriter(outStream));
		buffer = new LinkedList<String>();

		while ((line = in.readLine()) != null) {
			if (isFieldlines) {
				if (line.contains("setlinewidth")) {
					// end of field-lines
					isFieldlines = false;
					buffer.add(line);
				}
			} else {
				if (rmFieldlines && line.contains("setlinewidth")
						&& (buffer.getLast().compareTo(line) == 0)) {
					// start of field-lines
					isFieldlines = true;
					buffer.removeLast();
				} else if (line.contains("scalefont")
						&& buffer.getLast().contains("setlinewidth")) {
					// rescale font
					line = line.substring(0, 9) + "1" + line.substring(11);
					buffer.add(line);
				} else {
					buffer.add(line);
				}
			}

			// limit buffer
			while (buffer.size() > PSConverter.BUFFER_SIZE) {
				out.write(buffer.removeFirst() + PSConverter.NEW_LINE);
			}
		}

		// flush
		while (buffer.size() > 0) {
			out.write(buffer.removeFirst() + PSConverter.NEW_LINE);
		}
		out.flush();
	}
}
