/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping;

import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.optim.algorithm.Solver;
import at.lcm.bu21.optim.algorithm.SolverList;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Memory;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.mapping.utils.AWSAThread;
import at.lcm.bu21.optim.core.mapping.utils.MappingHolder;
import at.lcm.bu21.optim.core.mapping.utils.MappingOptimizationThread;
import at.lcm.bu21.optim.core.mapping.utils.SensitivityAnalysisHolder;
import at.lcm.bu21.optim.core.mapping.utils.SensitivityAnalysisThread;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.THolderState;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.TThreadState;
import magopt.utils.Base64;

public class MappingManager {

	/** Problem of optimizer */
	private final Problem problem;

	private final Memory mappingMemory;

	private final Optimization optim;

	private Solver solv = null;

	private String masterFolderName;

	private MappingOptimizationThread optimThread;

	private SensitivityAnalysisThread sensitivityThread;

	private final AtomicBoolean stopOptimization = new AtomicBoolean(false);

	private final AtomicBoolean stopSensitivityAnalysis = new AtomicBoolean(false);

	private AtomicInteger optProgress = new AtomicInteger(0);

	private AtomicInteger saProgress = new AtomicInteger(0);

	private AtomicInteger awsaProgress = new AtomicInteger(0);

	private List<String> foldersToDelete = new ArrayList<String>();

	private Set<String> targetNames = new TreeSet<String>();

	private List<String> sensitivityTargetNames = new ArrayList<String>();

	private Map<String, Boolean> fastTrainingEnabled = new HashMap<String, Boolean>();

	private Map<String, Integer> reducedTrainingSetPreferances = new HashMap<String, Integer>();

	private Map<String, Integer> maxTrainingDurationPreferances = new HashMap<String, Integer>();

	private boolean useErrorsClassifier;

	private Individual currentBaseIndividual = null;
	private SensitivityAnalysisHolder currentSensitivityHolder = null;
	private List<SensitivityAnalysisHolder> savedSensitivityHolders = new ArrayList<SensitivityAnalysisHolder>();

	private int currentSensitivityAnalysisDisplayIndex;

	private AtomicBoolean isPerformingAWSA = new AtomicBoolean(false);

	private boolean mappingColoredLSM = true; // Local Sensitivity Matrix is
												// colored according to
												// mapping-based SA
	private boolean feColoredLSM = true; // Local Sensitivity Matrix is colored
											// according to FE-based SA

	public boolean isUsingSavedAnalysis() {
		return (this.currentSensitivityAnalysisDisplayIndex >= 0);
	}

	public Set<String> getTargetNames() {
		return this.targetNames;
	}

	public List<String> getSensitivityTargetNames() {
		return this.sensitivityTargetNames;
	}

	/**
	 * Map that stores the mappings associated with the targets (optimization
	 * objectives and/or constraints) of the current problem
	 */
	private final Map<String, MappingHolder> mappingHolders = new HashMap<String, MappingHolder>();

	public MappingManager(Optimization optim) throws Exception {

		this.masterFolderName = System.getenv("TEMP") + "/MappingManager_" + UUID.randomUUID().toString();

		this.problem = optim.getProblem();
		this.mappingMemory = new Memory(this.problem.getArchive());
		this.optim = optim;
		this.useErrorsClassifier = false;

		// String outcome = OpenCVDynamicBinder.loadLibrary();
		// this.addLog(outcome);

		this.solv = SolverList.getSolverListEntry(this.problem.getSolver().getOptimizerName()).getSolver(this.problem);
		this.solv.setSettings(this.problem.getSolver().getSettings());
		this.solv.setState(this.problem.getSolver().getState());

		this.extractTargetNames();
		this.extractSensitivityTargetNames();

		this.createMappingHolders();
		this.loadSavedSensitivityHolders();
	}

	public void addFolderToDelete(String folderName) {
		this.foldersToDelete.add(folderName);
	}

	public void addLog(String message) {
		this.optim.addLog(message);
	}

	public void addError(String message, StackTraceElement[] stack) {
		this.optim.addLogError(message, stack);
	}

	public boolean areAllMappingHoldersReady() {
		for (String key : this.mappingHolders.keySet()) {
			if (this.targetNames.contains(key)) {
				if (!this.mappingHolders.get(key).getState().equals(THolderState.READY)) {
					return false;
				}
			}
		}
		return true;
	}

	public boolean areAllMappingHoldersExceptErrorsClsReady() {
		for (String key : this.mappingHolders.keySet()) {
			if (this.targetNames.contains(key)) {
				if (!(this.mappingHolders.get(key).getState().equals(THolderState.READY))
						&& !(key.equals(TMappingConstants.ERRORS_CLASSIFIER))) {
					return false;
				}
			}
		}
		return true;
	}

	public boolean areAnyMappingHoldersBusy() {
		for (String key : this.mappingHolders.keySet()) {
			if (this.targetNames.contains(key)) {
				if (this.mappingHolders.get(key).getState().equals(THolderState.BUSY)) {
					return true;
				}
			}
		}
		return false;
	}

	public void clearMappingMemory() {
		this.mappingMemory.clear();
		this.optimThread.setOptimizationState(TThreadState.IDLE);

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	private void createMappingHolders() {
		int validMappingStrings = 0;
		int i = -1;
		MappingHolder mappingHolder = null;
		for (ResultDesc rDesc : this.problem.getResults()) {
			i++;
			String mappingString = this.problem.getSerializedMappings()[i];
			String targetName = rDesc.getName();
			if (!mappingString.equals("")) {
				validMappingStrings++;
				try {
					mappingHolder = this.loadMappingHolderFromString(mappingString);
					mappingHolder.performPostDeserializationActions(this);
				} catch (Exception e) {
					this.optim.addLogError("ERROR: could not load pre-existing mapping for target " + targetName + " !",
							e.getStackTrace());
					// addLog(e.getMessage());
					// this.controller
					// .addLog("ERROR: could not load pre-existing mapping for
					// target "
					// + targetName + " !");
					validMappingStrings--;
					mappingHolder = new MappingHolder(targetName, this, "REG");
				}
			} else {
				mappingHolder = new MappingHolder(targetName, this, "REG");
			}
			this.mappingHolders.put(targetName, mappingHolder);
		}

		String classifierString = this.problem.getSerializedErrorsClassifier();
		if (!classifierString.equals("")) {
			validMappingStrings++;
			try {
				mappingHolder = this.loadMappingHolderFromString(classifierString);
				mappingHolder.performPostDeserializationActions(this);
			} catch (Exception e) {
				e.printStackTrace();
				this.optim.addLog("ERROR: could not load pre-existing errors classifier!");
				validMappingStrings--;
				mappingHolder = new MappingHolder(TMappingConstants.ERRORS_CLASSIFIER, this, "CLS");
			}
		} else {
			mappingHolder = new MappingHolder(TMappingConstants.ERRORS_CLASSIFIER, this, "CLS");
		}
		this.mappingHolders.put(TMappingConstants.ERRORS_CLASSIFIER, mappingHolder);

		for (String targetName : this.mappingHolders.keySet()) {
			this.maxTrainingDurationPreferances.put(targetName,
					this.mappingHolders.get(targetName).getMaxTrainingDuration());
			this.reducedTrainingSetPreferances.put(targetName,
					this.mappingHolders.get(targetName).getReducedTrainingSetSize());
		}

		this.optim.addLog("The MappingManager has detected " + this.getNumMappingTargets() + " targets and "
				+ validMappingStrings + " pre-existing mappings.");
	}

	private void loadSavedSensitivityHolders() {
		List<String> serializedSensitivityHolders = this.problem.getSerializedSensitivityAnalyses();

		int validSensitivityStrings = 0;
		int loadedSensitivityStrings = 0;

		this.savedSensitivityHolders.clear();

		if (serializedSensitivityHolders != null) {
			for (String sensitivityString : serializedSensitivityHolders) {
				SensitivityAnalysisHolder saHolder = null;
				if (!sensitivityString.equals("")) {
					validSensitivityStrings++;
					loadedSensitivityStrings++;
					try {
						saHolder = this.loadSensitivityHolderFromString(sensitivityString);

					} catch (Exception e) {
						this.optim.addLogError("ERROR: could not load pre-existing sensitivity !", e.getStackTrace());
						loadedSensitivityStrings--;
					}
				}
				if (saHolder != null) {
					saHolder.performPostDeserializationActions(this);
					this.savedSensitivityHolders.add(saHolder);
				}
			}
		}
		this.optim.addLog("The MappingManager has detected " + validSensitivityStrings + " and loaded "
				+ loadedSensitivityStrings + " existing sensitivity analyses!");

		this.currentSensitivityAnalysisDisplayIndex = -1;
		if (loadedSensitivityStrings > 0) {

			this.selectSavedSensitvityAnalysis(this.savedSensitivityHolders.size() - 1);
			// System.out.println(this.currentSensitivityHolder.getDisplayName());
		}

		// this.getProblem().clearSensitivityAnalyses();
	}

	public boolean isIndividualLikelyToBeErroneous(Individual ind) {
		boolean outcome = false;
		try {
			List<Double> sample = new ArrayList<Double>();
			for (VariableDesc vdesc : this.problem.getVariedVariables()) {
				sample.add(ind.getVariable(vdesc.getName()));
			}
			// sample.add(Double.parseDouble(TMappingConstants.NO_ERROR));

			// predict if the design is likely to be erroneous using the errors
			// classifier
			Double clsLabel = this.mappingHolders.get(TMappingConstants.ERRORS_CLASSIFIER).getMappingPrediction(sample);

			if (clsLabel.equals(Double.parseDouble(TMappingConstants.ERROR))) {
				// System.out.println("ERRONEOUS individual detected !");
				outcome = true;
			}

		} catch (Exception e) {
			e.printStackTrace();
		}

		return outcome;
	}

	public boolean evaluateRegressionTargetsOfIndividual(Individual ind) {
		boolean outcome = true;

		try {
			List<Double> sample = new ArrayList<Double>();
			for (VariableDesc vdesc : this.problem.getVariedVariables()) {
				sample.add(ind.getVariable(vdesc.getName()));
			}
			// sample.add(0.0);

			// predict the objective and constraint values using the
			// corresponding regression
			// models
			for (ResultDesc rDesc : this.problem.getResults()) {
				if (rDesc.isObjective() || rDesc.isConstraint()) {
					double prediction = this.mappingHolders.get(rDesc.getName()).getMappingPrediction(sample);

					// this.controller.addLog(rDesc.getName() +
					// " - predicted : "
					// + prediction);

					ind.setResult(rDesc.getName(), prediction);
				} else {
					ind.setResult(rDesc.getName(), 0.0);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}

	private void extractTargetNames() {
		for (ResultDesc rDesc : this.problem.getResults()) {
			if (rDesc.isObjective() || rDesc.isConstraint()) {
				this.targetNames.add(rDesc.getName());
				this.setFastTrainingEnabled(rDesc.getName(), false);
			}
		}
		this.targetNames.add(TMappingConstants.ERRORS_CLASSIFIER);
		this.setFastTrainingEnabled(TMappingConstants.ERRORS_CLASSIFIER, false);
	}

	private void extractSensitivityTargetNames() {
		for (ResultDesc rDesc : this.problem.getResults()) {
			if (rDesc.isObjective() || rDesc.isConstraint()) {
				this.sensitivityTargetNames.add(rDesc.getName());
			}
		}
	}

	public Memory getMappingMemory() {
		return this.mappingMemory;
	}

	public String getMappingQuality(String targetName) {
		String value = "-";
		MappingHolder mappingHolder = this.mappingHolders.get(targetName);
		if ((mappingHolder != null) && (mappingHolder.getState().equals(THolderState.READY))) {
			value = mappingHolder.getModelQuality();
		}
		return value;
	}

	public String getNrOfTrainingSamples(String targetName) {
		String value = "-";
		MappingHolder mappingHolder = this.mappingHolders.get(targetName);
		if ((mappingHolder != null) && (mappingHolder.getState().equals(THolderState.READY))) {
			value = mappingHolder.getNrOfUsedTrainingSamples();
		}
		return value;
	}

	public int getNumMappingTargets() {
		return this.targetNames.size();
	}

	public int getOptimizationProgressValue() {
		return this.optProgress.get();
	}

	public int getSensitivityAnalysisProgressValue() {
		return this.saProgress.get();
	}

	public int getAWSAsProgressValue() {
		return this.awsaProgress.get();
	}

	public TThreadState getOptimizationState() {
		if (this.optimThread == null) {
			return TThreadState.IDLE;
		} else {
			return this.optimThread.getOptimizationState();
		}
	}

	public Problem getProblem() {
		return this.problem;
	}

	public void handleHolderStateChange(MappingHolder mappingHolder) {
		if (mappingHolder.getState().equals(THolderState.READY)) {
			this.persistMappingHolder(mappingHolder);
		}

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	public void handleOptimizationStateChange() {
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	public boolean hasSufficientSamples() {
		for (String targetName : this.targetNames) {
			return this.hasSufficientSamples(targetName);
		}
		return false;
	}

	public boolean hasSufficientSamples(String targetName) {
		return this.mappingHolders.get(targetName).hasSufficientSamples();
	}

	public boolean isHolderBusy(String targetName) {
		return this.mappingHolders.get(targetName).getState().equals(THolderState.BUSY);
	}

	public boolean isMappingCreated(String targetName) {
		return this.mappingHolders.get(targetName).isMappingCreated();
	}

	public boolean isOptimizationIdle() {
		if (this.optimThread != null) {
			if (this.optimThread.getOptimizationState().equals(TThreadState.IDLE)) {
				return true;
			} else {
				return false;
			}
		}

		return true;
	}

	public boolean isOptimizationRunning() {
		if (this.optimThread != null) {
			if (this.optimThread.getOptimizationState().equals(TThreadState.RUNNING)) {
				return true;
			}
		}

		return false;
	}

	public MappingHolder loadMappingHolderFromString(String serializationString)
			throws IOException, ClassNotFoundException {
		byte[] data = Base64.decode(serializationString);
		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(data));
		MappingHolder mh = (MappingHolder) ois.readObject();
		ois.close();
		return mh;
	}

	private SensitivityAnalysisHolder loadSensitivityHolderFromString(String serializationString)
			throws IOException, ClassNotFoundException {
		byte[] data = Base64.decode(serializationString);
		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(data));
		SensitivityAnalysisHolder sh = (SensitivityAnalysisHolder) ois.readObject();
		ois.close();
		return sh;
	}

	public boolean persistMappingHolder(MappingHolder mappingHolder) {
		boolean outcome = true;

		try {
			if (mappingHolder.getTargetName().equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				this.problem.setSerializedErrorsClassifier(this.saveMappingHolderToString(mappingHolder));
			} else {
				this.problem.setSerializedMapping(mappingHolder.getTargetName(),
						this.saveMappingHolderToString(mappingHolder));
				this.optim.addLog("Mapping holder for: " + mappingHolder.getTargetName() + " was persisted!");
			}

			this.problem.getArchive().save();
		} catch (Exception e) {
			e.printStackTrace();
			this.optim.addLog("ERROR: could not persist the mapping holder for target " + mappingHolder.getTargetName()
					+ " to file!");
			outcome = false;
		}

		return outcome;
	}

	public boolean persistSensitivityHolder(SensitivityAnalysisHolder sensitivityHolder, boolean update) {
		boolean outcome = true;

		try {
			if (update) {
				this.problem.updateSerializedSensitivityAnalyses(this.saveSensitivityHolderToString(sensitivityHolder),
						this.currentSensitivityAnalysisDisplayIndex);
			} else {
				this.problem.addSerializedSensitivityAnalyses(this.saveSensitivityHolderToString(sensitivityHolder));
			}
			this.problem.getArchive().save();
		} catch (Exception e) {
			e.printStackTrace();
			this.optim.addLog("ERROR: could not persist the sensitivity holder to file!");
			outcome = false;
		}

		return outcome;
	}

	/**
	 * Order the creation of new mappings (i.e., the regression models) for every
	 * target. The mappings will be created in parallel and stored in the problem
	 * object.
	 */
	public void runGeneralMappingCreation() {
		this.optim.addLog("Creating mappings for all targets!");

		/** Train new mappings */
		for (String targetName : this.targetNames) {
			MappingHolder mappingHolder = this.mappingHolders.get(targetName);
			mappingHolder.setMaxTrainingDuration(this.maxTrainingDurationPreferances.get(targetName));
			mappingHolder.setReducedTrainingSetSize(this.reducedTrainingSetPreferances.get(targetName));
			mappingHolder.startFastCrossValidationExperiment(this.fastTrainingEnabled.get(targetName));
		}

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		// System.out.println("Extracting savepoint info:");
		// try {
		// this.storeSavepointInformationOnDisk("d:\\\\IEMDCData\\Scenario5New");
		// } catch (Exception e) {
		// e.printStackTrace();
		// }
	}

	private void storeSavepointInformationOnDisk(String directoryName) throws Exception {
		int currentSavePointNr = this.getProblem().getNumSavePoints();

		int actualSavePointNr = 0;
		for (int i = 1; i < this.getProblem().getNumSavePoints(); i++) {

			// if (this.problem.getSavePoint(i).getType()
			// .equals(SavePointType.INTERVAL)) {
			// actualSavePointNr++;
			// } else {
			// break;
			// }
			actualSavePointNr++;

			String fileName2 = directoryName + "\\" + actualSavePointNr + ".archive";

			File f = new File(fileName2);
			File dir = new File(f.getParent());
			if (!dir.exists() && !dir.mkdirs()) {
				throw new IOException("Could not create directory path: " + f.getParent());
			}

			if (!f.exists()) {
				f.createNewFile();
			}

			BufferedWriter bw = new BufferedWriter(new FileWriter(fileName2, true));
			for (Individual ind : this.problem.getSavePoint(i).getLocalPareto()) {
				// for (Individual ind : this.problem.getSolver().getArchive())
				// {
				String line = "";
				for (ResultDesc rDesc : this.problem.getResults()) {
					if (rDesc.isObjective()) {
						Individual archivedInd = this.problem.getArchive().getMemory().getIndividual(ind);
						Double value = archivedInd.getResult(rDesc.getName());
						line += " " + String.format("%.6f", value);
					}
				}
				bw.write(line + "\n");
			}
			System.out.println("for SavePoint nr. " + actualSavePointNr);
			System.out.println("save time: " + this.problem.getSavePoint(i).getDate());
			bw.close();
		}
	}

	public void runIndividualMappingCreation(String targetName) {
		this.optim.addLog("Creating mapping for target: " + targetName + " Fast training="
				+ this.fastTrainingEnabled.get(targetName));

		MappingHolder mappingHolder = this.mappingHolders.get(targetName);
		mappingHolder.setMaxTrainingDuration(this.maxTrainingDurationPreferances.get(targetName));
		mappingHolder.setReducedTrainingSetSize(this.reducedTrainingSetPreferances.get(targetName));
		mappingHolder.startFastCrossValidationExperiment(this.fastTrainingEnabled.get(targetName));

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	public void runIndividualMappingRetraining(String targetName) {
		this.optim.addLog("Retraining mapping for target: " + targetName);

		MappingHolder mappingHolder = this.mappingHolders.get(targetName);
		mappingHolder.setMaxTrainingDuration(this.maxTrainingDurationPreferances.get(targetName));
		mappingHolder.setReducedTrainingSetSize(this.reducedTrainingSetPreferances.get(targetName));
		mappingHolder.startEnsRetrainingExperiment();

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	public void runMappingBasedOptimization() {
		this.optim.addLog("Mapping based optimization started!");

		this.stopOptimization.set(false);
		this.optProgress.set(0);
		// this.managedPanel.refresh();
		this.optimThread = new MappingOptimizationThread(this, this.stopOptimization);
		this.optimThread.start();
	}

	private String saveMappingHolderToString(Serializable mappingHolder) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(mappingHolder);
		oos.close();
		return new String(Base64.encodeBytes(baos.toByteArray()));
	}

	private String saveSensitivityHolderToString(Serializable sensitivityHolder) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(sensitivityHolder);
		oos.close();
		return new String(Base64.encodeBytes(baos.toByteArray()));
	}

	public void shutdown() {
		this.addFolderToDelete(this.masterFolderName);

		/** remove all running jobs */
		for (ResultDesc rDesc : this.problem.getResults()) {
			if (rDesc.isObjective() || rDesc.isConstraint()) {
				MappingHolder mappingHolder = this.mappingHolders.get(rDesc.getName());
				mappingHolder.stopMappingCreationThread();
			}
		}

		boolean allJobsRemoved = false;
		while (!allJobsRemoved) {
			// System.out.println("Waiting for all jobs to be removed ...");
			for (ResultDesc rDesc : this.problem.getResults()) {
				if (rDesc.isObjective() || rDesc.isConstraint()) {
					MappingHolder mappingHolder = this.mappingHolders.get(rDesc.getName());
					allJobsRemoved = true;
					if (mappingHolder.hasUnremovedJobs()) {
						allJobsRemoved = false;
					}
				}
			}
		}

		/** delete temporary folders */
		for (String folderName : this.foldersToDelete) {
			try {
				// System.out.println("Deleting folder: " + folderName);
				FileUtils.delete(folderName);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public void stopIndividualMappingCreation(String targetName) {
		this.optim.addLog("Stopping mapping creation for target: " + targetName);

		this.mappingHolders.get(targetName).stopMappingCreationThread();

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	public void stopMappingBasedOptimization() {
		this.stopOptimization.set(true);
	}

	public void updateOptimizationProgress(double optimizationProgress) {
		int val = (int) (optimizationProgress * 100);
		int oldVal = this.optProgress.getAndSet(val);
		if (oldVal < val) {
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		}
	}

	public void updateSensitivityAnalysisProgress(double sensitivityAnalysisProgress) {
		int val = (int) (sensitivityAnalysisProgress * 100);
		int oldVal = this.saProgress.getAndSet(val);
		if (oldVal < val) {
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
		}
	}

	public void updateAWSAProgress(double awsaProgress) {
		int val = (int) (awsaProgress * 100);
		int oldVal = this.awsaProgress.getAndSet(val);
		if (oldVal < val) {
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
		}
	}

	public void verifyBestIndividuals(int count) {
		List<Individual> indsToVerify = this.optimThread.extractBestIndividuals(count);
		for (Individual ind : indsToVerify) {
			ind.setFromMapping(true);
		}

		this.optim.add2Buffer(indsToVerify.toArray(new Individual[indsToVerify.size()]));

		this.clearMappingMemory();

	}

	public void setFastTrainingEnabled(String targetName, Boolean fastTrainingEnabled) {
		this.fastTrainingEnabled.put(targetName, fastTrainingEnabled);
	}

	public void setReducedTrainingSetPreference(String targetName, String preferenceValue) {
		// this.fastTrainingEnabled.put(targetName, fastTrainingEnabled);
		int value = TMappingConstants.DEFAULT_NR_OF_TRAINING_SAMPLES;
		try {
			value = Integer.parseInt(preferenceValue);
		} catch (Exception e) {
			this.addLog("Training set size preference (" + preferenceValue
					+ ") is not a valid integer. The default value (" + value + ") shall be used!");
		}
		if (value < TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES) {
			value = TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES;
			this.addLog("Training set size preference (" + preferenceValue + ") is smaller than the min allowed value ("
					+ TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES + "). The min value (" + value + ") shall be used!");
		}
		int currentSamples = this.mappingHolders.get(targetName).getNrOfCurrentTrainingSamples();
		if (value > currentSamples) {
			value = currentSamples;
			this.addLog("Training set size preference (" + preferenceValue
					+ ") is larger than the current number of available samples (" + currentSamples
					+ "). The current number of samples (" + value + ") shall be used!");
		}

		this.reducedTrainingSetPreferances.put(targetName, value);
	}

	public void setMaxTrainingDurationPreference(String targetName, String preferenceValue) {
		int value = TMappingConstants.DEFAULT_MAPPING_CREATION_TIME_LIMIT;
		try {
			value = Integer.parseInt(preferenceValue);
		} catch (Exception e) {
			this.addLog("Training duration preference (" + preferenceValue
					+ ") is not a valid integer. The default value (" + value + ") shall be used!");
		}
		if (value < TMappingConstants.MIN_MAPPING_CREATION_TIME_LIMIT) {
			value = TMappingConstants.MIN_MAPPING_CREATION_TIME_LIMIT;
			this.addLog("Training duration preference (" + preferenceValue + ") is smaller than the min allowed value ("
					+ TMappingConstants.MIN_MAPPING_CREATION_TIME_LIMIT + "). The min value (" + value
					+ ") shall be used!");
		}
		if (value > TMappingConstants.MAX_MAPPING_CREATION_TIME_LIMIT) {
			value = TMappingConstants.MAX_MAPPING_CREATION_TIME_LIMIT;
			this.addLog("Training duration preference (" + preferenceValue + ") is larger than the max allowed value ("
					+ TMappingConstants.MAX_MAPPING_CREATION_TIME_LIMIT + "). The max value (" + value
					+ ") shall be used!");
		}
		this.maxTrainingDurationPreferances.put(targetName, value);
	}

	public Double getPlusVariableRange(String variableName) {
		if (this.currentSensitivityHolder != null) {
			return this.currentSensitivityHolder.getPlusVariableRange(variableName);
		}
		return 0.0;
	}

	public Double getMinusVariableRange(String variableName) {
		if (this.currentSensitivityHolder != null) {
			return this.currentSensitivityHolder.getMinusVariableRange(variableName);
		}
		return 0.0;
	}

	public Integer getReducedTrainingSetPreference(String targetName) {
		return this.reducedTrainingSetPreferances.get(targetName);
	}

	public Integer getMaxTrainingDurationPreference(String targetName) {
		return this.maxTrainingDurationPreferances.get(targetName);
	}

	public Individual getcurrentBaseIndividual() {
		return this.currentBaseIndividual;
	}

	public boolean useErrorsClassifier() {
		return this.useErrorsClassifier;
	}

	public void setUseErrorsClassifier(boolean useErrorsClassifier) {
		this.useErrorsClassifier = useErrorsClassifier;
	}

	public void initializeSensitivityAnalysis(Individual chosenIndividual) {

		// this.currentBaseIndividual = this.problem.getSavePoint(
		// this.problem.getNumSavePoints() - 1).getFullPareto()[chosenID];

		this.currentBaseIndividual = chosenIndividual;

		this.currentSensitivityHolder = new SensitivityAnalysisHolder(this, this.currentBaseIndividual);

		this.currentSensitivityAnalysisDisplayIndex = -1;

		if (this.sensitivityThread != null) {
			this.sensitivityThread.setSensitivityAnalysisState(TThreadState.IDLE);
		}

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));

		// String v = "";
		// String o = "";
		// for (VariableDesc vdesc : this.problem.getVariedVariables()) {
		// v += this.indToAnalyse.getVariable(vdesc.getName()) + ",";
		// }
		// for (ResultDesc rDesc : this.problem.getResults()) {
		// if (rDesc.isObjective() || rDesc.isConstraint()) {
		// o += this.indToAnalyse.getResult(rDesc.getName()) + ",";
		// }
		// }
		//
		// System.out.println("Variables:" + v + " - Objectives: " + o);

	}

	public void runMappingBasedSensitivityAnalysis() {
		this.optim.addLog("Mapping-based sensitivity analysis started!");

		if (!this.currentSensitivityHolder.getIsWaitingForFEResults()
				&& !this.currentSensitivityHolder.getContainsFEResults()) {
			this.currentSensitivityHolder.createLocalSensitivityIndividuals();
			this.currentSensitivityHolder.createGlobalSensitivityIndividuals();
		}

		this.currentSensitivityHolder.updateIndividualsCount();

		this.stopSensitivityAnalysis.set(false);
		this.saProgress.set(0);
		// this.managedPanel.refresh();
		this.sensitivityThread = new SensitivityAnalysisThread(this, this.stopSensitivityAnalysis,
				this.currentSensitivityHolder);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		this.sensitivityThread.start();
	}

	public void runFEBasedSensitivityAnalysis() {
		this.optim.addLog("FE-based sensitivity analysis started!");

		this.currentSensitivityHolder.setIsWaitingForFEResults(true);
		this.saveMappingBasedSensitivityInformation();
		this.currentSensitivityHolder.addSensitivityIndividualsToSensitivityMemory();

		// this.persistSensitivityHolder(this.currentSensitivityHolder,
		// this.isUsingSavedAnalysis());
		// this.loadSavedSensitivityHolders();

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
		// this.controller.fireOptmizerPerformed(new OptimizerEvent(
		// OptimizerEvent.MAPPING));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	public void stopSensitivityAnalysis() {
		this.stopSensitivityAnalysis.set(true);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	public void saveMappingBasedSensitivityInformation() {
		if (this.currentSensitivityHolder != null) {

			int cIndex = this.currentSensitivityAnalysisDisplayIndex;

			this.currentSensitivityHolder.saveIndVarValues();
			this.persistSensitivityHolder(this.currentSensitivityHolder, this.isUsingSavedAnalysis());
			this.loadSavedSensitivityHolders();

			if (cIndex < 0) {
				cIndex = this.savedSensitivityHolders.size() - 1;
			}
			this.selectSavedSensitvityAnalysis(cIndex);

			if (this.sensitivityThread != null) {
				this.sensitivityThread.setSensitivityAnalysisState(TThreadState.IDLE);
			}
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		}

	}

	public void saveFEBasedSensitivityInformation() {
		if (this.currentSensitivityHolder != null) {
			int cIndex = this.currentSensitivityAnalysisDisplayIndex;

			this.persistSensitivityHolder(this.currentSensitivityHolder, true);
			this.loadSavedSensitivityHolders();

			if (cIndex < 0) {
				cIndex = this.savedSensitivityHolders.size() - 1;
			}
			this.selectSavedSensitvityAnalysis(cIndex);

			if (this.sensitivityThread != null) {
				this.sensitivityThread.setSensitivityAnalysisState(TThreadState.IDLE);
			}
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		}
	}

	public void discardMappingBasedSensitivityInformation() {

		int cIndex = this.currentSensitivityAnalysisDisplayIndex;

		if (cIndex < 0) {
			this.currentSensitivityHolder = new SensitivityAnalysisHolder(this, this.currentBaseIndividual);
		} else {
			SensitivityAnalysisHolder saHolder = null;
			try {
				saHolder = this
						.loadSensitivityHolderFromString(this.problem.getSerializedSensitivityAnalyses().get(cIndex));

			} catch (Exception e) {
				this.optim.addLogError("ERROR: could not load pre-existing sensitivity !", e.getStackTrace());
			}
			if (saHolder != null) {
				saHolder.performPostDeserializationActions(this);
				this.savedSensitivityHolders.set(cIndex, saHolder);
			}

			this.selectSavedSensitvityAnalysis(cIndex);

		}

		if (this.sensitivityThread != null) {
			this.sensitivityThread.setSensitivityAnalysisState(TThreadState.IDLE);
		}

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
	}

	public TThreadState getSensitivityRunState() {
		if (this.sensitivityThread == null) {
			return TThreadState.IDLE;
		} else {
			return this.sensitivityThread.getSensitivityAnalysisState();
		}
	}

	public void setPlusVariableRangesPreference(String variableName, String preferenceValue) {
		double value = this.currentBaseIndividual.getVariable(variableName)
				* TMappingConstants.DEFAULT_SENSITIVITY_RANGE;
		try {
			value = Double.parseDouble(preferenceValue);
			value = Math.abs(value);
		} catch (Exception e) {
			this.addLog("Plus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is not a valid double. The default value (" + value + ") shall be used!");
		}
		if (value < 0.0) {
			value = TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES;
			this.addLog("Plus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is smaller than the min allowed value (0.0). The min value (0.0) shall be used!");
		}

		double maxValue = this.currentBaseIndividual.getVariable(variableName)
				* TMappingConstants.MAX_SENSITIVITY_RANGE;

		if (maxValue == 0.0) {
			maxValue = TMappingConstants.MAX_SENSITIVITY_RANGE * (this.getProblem().getVariable(variableName).getUB()
					- this.getProblem().getVariable(variableName).getLB());
		}

		if (value > maxValue) {
			value = maxValue;
			this.addLog("Plus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is larger than the max allowed value (" + maxValue + "). The max value (" + value
					+ ") shall be used!");
		}

		this.currentSensitivityHolder.setPlusVariableRange(variableName, value);
	}

	public void setMinusVariableRangesPreference(String variableName, String preferenceValue) {
		double value = this.currentBaseIndividual.getVariable(variableName)
				* TMappingConstants.DEFAULT_SENSITIVITY_RANGE;
		try {
			value = Double.parseDouble(preferenceValue);
			value = Math.abs(value);
		} catch (Exception e) {
			this.addLog("Minus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is not a valid double. The default value (" + value + ") shall be used!");
		}
		if (value < 0.0) {
			value = TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES;
			this.addLog("Minus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is smaller than the min allowed value (0.0). The min value (0.0) shall be used!");
		}

		double maxValue = this.currentBaseIndividual.getVariable(variableName)
				* TMappingConstants.MAX_SENSITIVITY_RANGE;

		if (maxValue == 0.0) {
			maxValue = TMappingConstants.MAX_SENSITIVITY_RANGE * (this.getProblem().getVariable(variableName).getUB()
					- this.getProblem().getVariable(variableName).getLB());
		}

		if (value > maxValue) {
			value = maxValue;
			this.addLog("Minus variable range preference for " + variableName + " (" + preferenceValue
					+ ") is larger than the max allowed value (" + maxValue + "). The max value (" + value
					+ ") shall be used!");
		}

		this.currentSensitivityHolder.setMinusVariableRange(variableName, value);
	}

	public void handleSensitivityAnalysisRunErronousEnd() {

		this.discardMappingBasedSensitivityInformation();

		this.optim.addLog("Mapping-based sensitivity analysis ended with errors");
	}

	public void handleSensitivityAnalysisSucessfulRun() {
		if (this.sensitivityThread != null) {

			this.currentSensitivityHolder.postProcessMappingBasedSensitivityInformation();

			this.sensitivityThread.setSensitivityAnalysisState(TThreadState.READY);
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		}
		this.optim.addLog("Mapping-based sensitivity analysis ended sucessfully");
	}

	public String getLocalSensitivityValueToDisplay(String variableName, String targetName) {
		if (this.currentSensitivityHolder == null) {
			return "-";
		}
		return this.currentSensitivityHolder.getLocalSensitivityValueToDisplay(variableName, targetName);
	}

	public String getGlobalSensitivityRelativeValueToDisplay(String targetName) {
		if (this.currentSensitivityHolder == null) {
			return "-";
		}
		return this.currentSensitivityHolder.getGlobalSensitivityRelativeValueToDisplay(targetName);
	}

	public String getGlobalSensitivityAbsoluteMinValueToDisplay(String targetName) {
		if (this.currentSensitivityHolder == null) {
			return "-";
		}
		return this.currentSensitivityHolder.getGlobalSensitivityAbsoluteMinValueToDisplayStr(targetName);
	}

	public String getGlobalSensitivityAbsolutePlusValueToDisplay(String targetName) {
		if (this.currentSensitivityHolder == null) {
			return "-";
		}
		return this.currentSensitivityHolder.getGlobalSensitivityAbsoluteMaxValueToDisplayStr(targetName);
	}

	public Double getNormalizedSensitivity(String variableName, String targetName) {
		if (this.currentSensitivityHolder == null) {
			return -1.0;
		}
		return this.currentSensitivityHolder.getNormalizedSensitivity(variableName, targetName, this.mappingColoredLSM,
				this.feColoredLSM);
	}

	public String getUsedAnalysisSavedName() {
		if (this.isUsingSavedAnalysis()) {
			return this.currentSensitivityHolder.getDisplayName();
		} else {
			return null;
		}
	}

	public Integer getNrOfSavedAnalyses() {
		return this.savedSensitivityHolders.size();
	}

	public String[] getSavedAnalysisNames() {
		String[] result = new String[this.savedSensitivityHolders.size()];
		int i = 0;
		for (SensitivityAnalysisHolder sah : this.savedSensitivityHolders) {
			result[i] = sah.getDisplayName() + " - SA nr. " + i;
			i++;
		}
		return result;
	}

	public int getCurrentSensitivityAnalysisDisplayIndex() {
		return this.currentSensitivityAnalysisDisplayIndex;
	}

	public void selectSavedSensitvityAnalysis(int selectedIndex) {
		this.currentSensitivityAnalysisDisplayIndex = selectedIndex;
		this.currentBaseIndividual = this.savedSensitivityHolders.get(this.currentSensitivityAnalysisDisplayIndex)
				.getBaseIndividual("FE");
		this.currentSensitivityHolder = this.savedSensitivityHolders.get(this.currentSensitivityAnalysisDisplayIndex);

		if (this.currentSensitivityHolder.getContainsFEResults()) {
			this.switchToFEColors();
		} else {
			this.switchToMappingColors();
		}

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
	}

	public boolean isSensitivityAnalysisIdle() {

		if (this.sensitivityThread != null) {
			if (this.sensitivityThread.getSensitivityAnalysisState().equals(TThreadState.IDLE)) {
				return true;
			} else {
				return false;
			}
		}

		return true;
	}

	public boolean isUsingSAWithFEResults() {
		// TODO investigate error
		// return this.currentSensitivityHolder.getContainsFEResults();
		return false;
	}

	public boolean isUsingSAWithMappingResults() {
		// TODO investigate error
		// return this.currentSensitivityHolder.getContainsMappingResults();
		return false;
	}

	public boolean isUsingSAWaitingForFEResults() {
		return this.currentSensitivityHolder.getIsWaitingForFEResults();
	}

	public void checkFEProgress() {
		if (this.currentSensitivityHolder != null) {
			this.currentSensitivityHolder.checkFEProgress();

			this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		}
	}

	public String getFESensitivityProgress() {
		if (this.currentSensitivityHolder != null) {

			return this.currentSensitivityHolder.getFESensitivityProgress();
		} else {
			return "ERROR: invalid holder!";
		}
	}

	public boolean isPerformingAWSA() {
		return this.isPerformingAWSA.get();
	}

	public void isPerformingAWSA(boolean isPerformingSA) {
		this.isPerformingAWSA.set(isPerformingSA);
	}

	public void runAWSA(String filePrefix) {
		this.awsaProgress.set(0);
		this.isPerformingAWSA(true);
		this.optim.addLog("Memory-wide sensitivity analysis started!");

		AWSAThread awsaThread = new AWSAThread(this, filePrefix);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
		awsaThread.start();
	}

	public void handleAWSASucessfulRun() {
		this.isPerformingAWSA(false);
		this.optim.addLog("Memory-wide sensitivity analysis finished!");
	}

	public void switchToMappingColors() {
		this.mappingColoredLSM = true;
		this.feColoredLSM = false;
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
	}

	public void switchToFEColors() {
		this.mappingColoredLSM = false;
		this.feColoredLSM = true;
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.SENSITIVITY));
	}
}
