/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.jobs;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTask;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTaskBatch;
import jobmanager.JobInterface;
import main.ocvwrap.java.OpenCVDataSource;
import main.ocvwrap.java.OpenCVDynamicBinder;
import main.ocvwrap.java.OpenCVMLPPredictor;
import main.ocvwrap.java.OpenCVUtils;

/**
 * Runs openCV training/parameter optimization tasks
 * 
 * @author Ciprian Zavoianu
 */
public class MappingJob implements JobInterface {

	private int nrOfFolds;
	private final Set<String> targetNames = new TreeSet<String>();
	private final TMappingTaskBatch uncompletedTaskBatch = new TMappingTaskBatch();
	private final TMappingTaskBatch completedTaskBatch = new TMappingTaskBatch();
	private final Map<String, List<OpenCVDataSource>> partitionedDataset = new HashMap<String, List<OpenCVDataSource>>();

	private Logger logger;

	private OpenCVMLPPredictor createBasePredictionModel(TMappingTask mappingTask) throws Exception {

		Double learningRate = Double
				.parseDouble(mappingTask.getParameters().get(TMappingTask.TParameter.LEARN_RATE.getValue()));
		Double momentum = Double
				.parseDouble(mappingTask.getParameters().get(TMappingTask.TParameter.MOMENTUM.getValue()));
		int nrOfHiddenUnits = Integer
				.parseInt(mappingTask.getParameters().get(TMappingTask.TParameter.HIDDEN_UNITS.getValue()));
		int maxSecondsForTraining = Integer
				.parseInt(mappingTask.getParameters().get(TMappingTask.TParameter.TRAINING_TIME.getValue())) * 60;

		OpenCVMLPPredictor newPredictor = new OpenCVMLPPredictor(learningRate, momentum, nrOfHiddenUnits, 10000, 25,
				250, maxSecondsForTraining);
		return newPredictor;

	}

	private void extractTargetNames() {
		for (TMappingTask mappingTask : this.uncompletedTaskBatch.getTaskList()) {

			this.logger.log(Level.INFO, "" + mappingTask.getParameters().size());
			this.logger.log(Level.INFO, "" + TMappingTask.TParameter.TARGET.getValue());

			this.targetNames.add(mappingTask.getParameters().get(TMappingTask.TParameter.TARGET.getValue()));
		}
	}

	private void loadPartitionedDataSet() throws Exception {
		this.extractTargetNames();
		for (String target : this.targetNames) {
			List<OpenCVDataSource> targetInstances = new ArrayList<OpenCVDataSource>();
			List<OpenCVDataSource> testInstances = new ArrayList<OpenCVDataSource>();
			for (int i = 0; i < this.nrOfFolds; i++) {

				OpenCVDataSource trainingSet = new OpenCVDataSource(
						target + TMappingConstants.TRAINING_PREFIX + i + ".arff");
				trainingSet.setAsTarget(trainingSet.getAttributesCount() - 1);

				OpenCVDataSource testSet = new OpenCVDataSource(target + TMappingConstants.TEST_PREFIX + i + ".arff");
				testSet.setAsTarget(testSet.getAttributesCount() - 1);

				targetInstances.add(trainingSet);
				testInstances.add(testSet);
			}
			targetInstances.addAll(testInstances);

			OpenCVDataSource allSamples = new OpenCVDataSource(
					target + TMappingConstants.TRAINING_PREFIX + 0 + ".arff");
			allSamples.addSamplesFromSource(targetInstances.get(this.nrOfFolds), 0,
					targetInstances.get(this.nrOfFolds).getNrOfOpenCVSamples() - 1);
			allSamples.setAsTarget(allSamples.getAttributesCount() - 1);

			targetInstances.add(allSamples);
			this.partitionedDataset.put(target, targetInstances);
		}
	}

	private String getElapsedTime(long startTime, long endTime) {
		Long elapsedTime = (endTime - startTime) / 1000;
		String seconds = String.format("%02d", elapsedTime % 60);
		String minutes = String.format("%02d", (elapsedTime % 3600) / 60);
		String hours = String.format("%03d", elapsedTime / 3600);
		return hours + ":" + minutes + ":" + seconds;
	}

	private void doTraining(TMappingTask mappingTask, OpenCVDataSource trainingSet, OpenCVMLPPredictor predictionModel)
			throws Exception {

		predictionModel.train(trainingSet, false);
		mappingTask.getResults().put(TMappingTask.TResult.TRAIN_CORR.getValue(),
				String.format(Locale.ENGLISH, "%.10f", predictionModel.getTrainingR2()));
		mappingTask.getResults().put(TMappingTask.TResult.TRAIN_RRSE.getValue(),
				String.format(Locale.ENGLISH, "%.10f", predictionModel.getTrainingRMSE()));
		mappingTask.getResults().put(TMappingTask.TResult.TRAIN_RMSE.getValue(),
				String.format(Locale.ENGLISH, "%.10f", predictionModel.getTrainingRMSE()));
		mappingTask.getRunInfo().put(TMappingTask.TRunInfo.TRAIN_SIZE.getValue(),
				"" + trainingSet.getNrOfOpenCVSamples());
		mappingTask.getResults().put(TMappingTask.TResult.CORRECTLY_CLASSIFIED_INSTANCES.getValue(),
				String.format(Locale.ENGLISH, "%.10f", predictionModel.getTrainingBACC()));
	}

	private void executeModelTrainingTask(TMappingTask mappingTask) throws Exception {
		OpenCVDataSource trainingSet = this.partitionedDataset
				.get(mappingTask.getParameters().get(TMappingTask.TParameter.TARGET.getValue()))
				.get(2 * this.nrOfFolds);

		OpenCVMLPPredictor predictor = this.createBasePredictionModel(mappingTask);
		this.doTraining(mappingTask, trainingSet, predictor);
		this.doCV10Training(mappingTask); // in order to have the CV10 quality
											// info

		Thread.sleep(30000);

		this.writeModelToFile(predictor, TMappingConstants.REMOTE_MODEL_FILE_NAME);
	}

	private void doCV10Training(TMappingTask mappingTask) throws Exception {

		Double[] valuesCORR = new Double[this.nrOfFolds];
		Double[] valuesRRSE = new Double[this.nrOfFolds];
		Double[] valuesRMSE = new Double[this.nrOfFolds];
		Double[] valuesBACC = new Double[this.nrOfFolds];
		double meanCORR = 0;
		double meanRRSE = 0;
		double meanRMSE = 0;
		double meanBACC = 0;
		double stdevCORR = 0;
		double stdevRRSE = 0;
		double stdevRMSE = 0;
		double stdevBACC = 0;

		for (int i = 0; i < this.nrOfFolds; i++) {

			OpenCVMLPPredictor foldPredictor = this.createBasePredictionModel(mappingTask);
			foldPredictor.train(
					this.partitionedDataset
							.get(mappingTask.getParameters().get(TMappingTask.TParameter.TARGET.getValue())).get(i),
					false);
			foldPredictor.predict(this.partitionedDataset
					.get(mappingTask.getParameters().get(TMappingTask.TParameter.TARGET.getValue()))
					.get(i + this.nrOfFolds), true);

			valuesCORR[i] = foldPredictor.getTestR2();
			valuesRRSE[i] = foldPredictor.getTestRMSE();
			valuesRMSE[i] = foldPredictor.getTestRMSE();
			valuesBACC[i] = foldPredictor.getTestBACC();
			meanCORR += valuesCORR[i];
			meanRRSE += valuesRMSE[i];
			meanRMSE += valuesRMSE[i];
			meanBACC += valuesBACC[i];
		}

		meanCORR /= this.nrOfFolds;
		meanRRSE /= this.nrOfFolds;
		meanRMSE /= this.nrOfFolds;
		meanBACC /= this.nrOfFolds;

		for (int i = 0; i < this.nrOfFolds; i++) {
			valuesCORR[i] = Math.pow((valuesCORR[i] - meanCORR), 2);
			valuesRRSE[i] = Math.pow((valuesRRSE[i] - meanRRSE), 2);
			valuesRMSE[i] = Math.pow((valuesRMSE[i] - meanRMSE), 2);
			valuesBACC[i] = Math.pow((valuesBACC[i] - meanBACC), 2);
		}

		for (int i = 0; i < this.nrOfFolds; i++) {
			stdevCORR += valuesCORR[i];
			stdevRRSE += valuesRRSE[i];
			stdevRMSE += valuesRMSE[i];
			stdevBACC += valuesBACC[i];
		}

		stdevCORR = Math.sqrt(stdevCORR / this.nrOfFolds);
		stdevRRSE = Math.sqrt(stdevRRSE / this.nrOfFolds);
		stdevRMSE = Math.sqrt(stdevRMSE / this.nrOfFolds);
		stdevBACC = Math.sqrt(stdevBACC / this.nrOfFolds);

		mappingTask.getResults().put(TMappingTask.TResult.CV_CORR_MEAN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", meanCORR));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RRSE_MEAN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", meanRRSE));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RMSE_MEAN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", meanRMSE));
		mappingTask.getResults().put(TMappingTask.TResult.CV_ACC_MEAN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", meanBACC));
		mappingTask.getResults().put(TMappingTask.TResult.CV_CORR_STDEV.getValue(),
				String.format(Locale.ENGLISH, "%.10f", stdevCORR));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RRSE_STDEV.getValue(),
				String.format(Locale.ENGLISH, "%.10f", stdevRRSE));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RMSE_STDEV.getValue(),
				String.format(Locale.ENGLISH, "%.10f", stdevRMSE));
		mappingTask.getResults().put(TMappingTask.TResult.CV_ACC_STDEV.getValue(),
				String.format(Locale.ENGLISH, "%.10f", stdevBACC));
		mappingTask.getResults().put(TMappingTask.TResult.CV_CORR_MIN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", (meanCORR - stdevCORR)));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RRSE_MIN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", (meanRRSE - stdevRMSE)));
		mappingTask.getResults().put(TMappingTask.TResult.CV_RMSE_MIN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", (meanRMSE - stdevRMSE)));
		mappingTask.getResults().put(TMappingTask.TResult.CV_ACC_MIN.getValue(),
				String.format(Locale.ENGLISH, "%.10f", (meanBACC - stdevBACC)));
	}

	private void executeCrossValidationEvaluationTask(TMappingTask mappingTask) throws Exception {
		OpenCVDataSource trainingSet = this.partitionedDataset
				.get(mappingTask.getParameters().get(TMappingTask.TParameter.TARGET.getValue()))
				.get(2 * this.nrOfFolds);
		OpenCVMLPPredictor predictor;
		predictor = this.createBasePredictionModel(mappingTask);
		this.doTraining(mappingTask, trainingSet, predictor);
		this.doCV10Training(mappingTask);
	}

	private void writeModelToFile(OpenCVMLPPredictor predictor, String fileName) throws Exception {
		File f = new File(fileName);
		if (f.getParent() != null) {
			File dir = new File(f.getParent());
			if (!dir.exists() && !dir.mkdirs()) {
				throw new IOException("Could not create directory path: " + f.getParent());
			}
		}

		FileOutputStream os = new FileOutputStream(f);
		os.write(OpenCVUtils.savePredictorToString(predictor).getBytes("ISO-8859-1"));
		os.close();
	}

	@Override
	public void runJob(String... arg0) throws Exception {
		this.logger = Logger.getLogger("OpenCV log " + arg0[0]);
		FileHandler handler = new FileHandler("localOpenCVJob.log", true);
		this.logger.addHandler(handler);
		this.logger.setLevel(java.util.logging.Level.ALL);

		String inputFileName = arg0[0];
		this.nrOfFolds = Integer.parseInt(arg0[1]);

		this.uncompletedTaskBatch.appendFromFile(inputFileName);
		this.logger.log(Level.INFO, "" + this.uncompletedTaskBatch.getTaskList().size());

		String outcome = "[opencv library not loaded]";

		try {
			outcome = OpenCVDynamicBinder.loadLibrary();
		} catch (Exception e) {
			this.logger.log(Level.SEVERE, "Error: could load opencv library! " + outcome, e);
			return;
		}
		try {
			this.loadPartitionedDataSet();
		} catch (Exception e) {
			this.logger.log(Level.SEVERE, "Error: could not read base dataset!", e);
			return;
		}
		for (int i = 0; i < this.uncompletedTaskBatch.getTaskList().size(); i++) {
			TMappingTask currentTask = new TMappingTask(this.uncompletedTaskBatch.getTaskList().get(i));
			boolean testFinished = true;
			try {
				long testStart = System.currentTimeMillis();
				if (currentTask.getParameters().get(TMappingTask.TParameter.EX_TYPE.getValue()).toUpperCase()
						.equals(TMappingConstants.TExperimentType.MODEL_TRAINING_REG.getValue())
						|| currentTask.getParameters().get(TMappingTask.TParameter.EX_TYPE.getValue()).toUpperCase()
								.equals(TMappingConstants.TExperimentType.MODEL_TRAINING_CLS.getValue())) {
					this.executeModelTrainingTask(currentTask);
				} else if (currentTask.getParameters().get(TMappingTask.TParameter.EX_TYPE.getValue()).toUpperCase()
						.equals(TMappingConstants.TExperimentType.CV_EVALUATION_REG.getValue())
						|| currentTask.getParameters().get(TMappingTask.TParameter.EX_TYPE.getValue()).toUpperCase()
								.equals(TMappingConstants.TExperimentType.CV_EVALUATION_CLS.getValue())) {
					this.executeCrossValidationEvaluationTask(currentTask);
				} else {
					throw new Exception(
							"ERROR: unsupported experiment type: " + TMappingTask.TParameter.EX_TYPE.getValue());
				}

				long testEnd = System.currentTimeMillis();
				String elapsedTime = this.getElapsedTime(testStart, testEnd);
				currentTask.getRunInfo().put(TMappingTask.TRunInfo.EVAL_TIME.getValue(), elapsedTime);
				currentTask.getRunInfo().put(TMappingTask.TRunInfo.EVAL_SEC.getValue(),
						"" + ((testEnd - testStart) / 1000));
			} catch (Exception e) {
				this.logger.log(Level.SEVERE, "ERROR: could not perform OPENCV test: " + i, e);
				testFinished = false;
			}

			if (testFinished) {
				try {
					this.completedTaskBatch.getTaskList().add(currentTask);
					this.completedTaskBatch.saveRecordToFile(TMappingConstants.REMOTE_TASK_BATCH_FILE_NAME, false, i);
				} catch (Exception e) {
					this.logger.severe("ERROR: could not save OPENCV test result for test: " + i);
					e.printStackTrace();
				}
			}
		} // end for

	}// end method
}
