/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.project;

import java.io.File;
import java.io.FileOutputStream;

import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.AreaReference;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Name;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.CellReference;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;

/**
 * Write pareto-front to excel
 * 
 * @author Gnther Weidenholzer
 */
public class XLSWrite {
	/**
	 * Write pareto-front to excel
	 * 
	 * @param problem
	 *            Problem with pareto-front
	 * @param file
	 *            File to write
	 * @return File
	 * @throws Exception
	 */
	public static File writeXLSReport(Problem problem, File file) throws Exception {
		Individual[] ind = problem.getSavePoint(problem.getNumSavePoints() - 1).getFullPareto();
		return XLSWrite.writeXLSReport(problem, file, ind);
	}

	/**
	 * Write pareto-front to excel
	 * 
	 * @param problem
	 *            Problem with pareto-front
	 * @param file
	 *            File to write
	 * @param ind
	 *            List of individuals
	 * @return File
	 * @throws Exception
	 */
	public static File writeXLSReport(Problem problem, File file, Individual[] ind) throws Exception {

		Workbook wb;
		if (file.getName().toLowerCase().endsWith(".xls")) {
			wb = new HSSFWorkbook();
		} else if (file.getName().toLowerCase().endsWith(".xlsx")) {
			wb = new XSSFWorkbook();
		} else {
			wb = new XSSFWorkbook();
			file = new File(file.getPath() + ".xlsx");
		}
		wb.createSheet();

		Project project = problem.getArchive().getProject();
		// write Parameters
		VariableDesc[] varDesc = problem.getVariables();
		int nVars = varDesc.length;
		for (int k = 0; k < nVars; k++) {
			// name
			String varname = project.getObject(varDesc[k].getName()).getShortIdentifier();
			XLSWrite.getXlCell(wb, 0, null, 0, k).setCellValue(varname);

			// values
			for (int q = 0; q < ind.length; q++) {
				XLSWrite.getXlCell(wb, 0, null, q + 1, k).setCellValue(ind[q].getVariable(k));
			}
		}

		// write Objectives
		ResultDesc[] resDesc = problem.getResults();
		int nObj = resDesc.length;
		for (int k = 0; k < nObj; k++) {
			// name
			String objname = project.getObject(resDesc[k].getName()).getShortIdentifier();
			if (resDesc[k].isObjective()) {
				objname += " (" + resDesc[k].getObjective().name() + ")";
			}
			XLSWrite.getXlCell(wb, 0, null, 0, nVars + k).setCellValue(objname);

			// values
			for (int q = 0; q < ind.length; q++) {
				XLSWrite.getXlCell(wb, 0, null, q + 1, nVars + k).setCellValue(ind[q].getResult(k));

			}
		}

		// Write the output to a file
		FileOutputStream fileOut = new FileOutputStream(file);
		wb.write(fileOut);
		fileOut.close();

		return file;

	}

	/**
	 * Get cell of workbook
	 * 
	 * @param wb
	 *            Workbook
	 * @param sheetIdx
	 *            Index of sheet
	 * @param ref
	 *            Reference
	 * @param rowOffset
	 *            Row index
	 * @param colOffset
	 *            Column index
	 * @return Selected cell
	 */
	private static Cell getXlCell(Workbook wb, int sheetIdx, String ref, int rowOffset, int colOffset) {

		int row = rowOffset;
		int col = colOffset;

		if (ref != null) {

			Name range = wb.getName(ref);
			CellReference cref = new AreaReference(range.getRefersToFormula()).getFirstCell();
			row += cref.getRow();
			col += cref.getCol();
		}

		Sheet sheet = wb.getSheetAt(sheetIdx);

		Row xlRow = sheet.getRow(row);
		if (xlRow == null) {
			xlRow = sheet.createRow(row);
		}

		Cell xlCell = xlRow.getCell(col);

		if (xlCell == null) {
			xlCell = xlRow.createCell(col);
		}

		return xlCell;
	}

}
