/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.rmi.RemoteException;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;

import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.bu21.magopt.utils.MatlabControl;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.event.OptimizerListener;
import at.lcm.bu21.optim.core.project.ComputeProject;
import at.lcm.bu21.optim.gui.setting.ChooseProblem;
import at.lcm.bu21.optim.gui.setting.ChooseResults;
import at.lcm.bu21.optim.gui.setting.ChooseVariables;
import at.lcm.bu21.optim.gui.setting.GeneralSettings;
import at.lcm.bu21.optim.gui.setting.SolverSettings;
import jobmanager.JobDesc;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;

/**
 * Optimizer gui
 * 
 * @author Werner Koppelsttter
 */
public class OptimizerGui extends JFrame {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Disable sensitivity for stable version */
	public static boolean enableSensitivity = false;

	/** Optimization */
	private Optimization optim;

	/** Views of optimizer */
	private OptimizerViews views;

	// display panels
	/** Tabbed pane for different optimizer */
	private JTabbedPane tabbedPane;

	/** Tabbed pane for plots */
	private JTabbedPane plots;

	// project menu entry
	/** Run project optimization */
	private JMenuItem mntmRunProject;
	/** Finish project optimization */
	private JMenuItem mntmFinishProject;
	/** Stop project optimization */
	private JMenuItem mntmStopProject;
	/** Init project buffer */
	private JMenuItem mntmInitProjectBuffer;
	/** Clear project buffer */
	private JMenuItem mntmClearProjectBuffer;
	/** Write pareto to excel */
	private JMenuItem mntmWriteToExel;
	/** Select pareto */
	private JMenuItem mntmSelectPareto;

	// mapping menu entry
	/** Create mapping optimization */
	private JMenuItem mntmCreateMapping;
	/** Run mapping optimization */
	private JMenuItem mntmRunMapping;

	// debug menu
	/** Debug menu */
	private JMenu mnDebug;

	/**
	 * Create the frame.
	 * 
	 * @param optim
	 *            Optimization
	 * @throws Exception
	 */
	public OptimizerGui(Optimization optim) throws Exception {
		JPanel contentPane, panelProj, panelMap, panelSensitivity, panelLog, panelComp;
		Problem problem;
		JMenuBar menuBar;

		this.optim = optim;

		// enable sensitivity
		String str = MagOptConfig.getConfig("OptimizerSensitivity");
		if (str != null) {
			try {
				OptimizerGui.enableSensitivity = Boolean.parseBoolean(str);
			} catch (Exception ex) {
			}
		}

		menuBar = new JMenuBar();
		this.setJMenuBar(menuBar);

		this.initMenuFile(menuBar);
		this.initMenuDisplay(menuBar);
		this.initMenuProject(menuBar);
		this.initMenuMapping(menuBar);
		this.initMenuSetting(menuBar);
		this.initMenuDebug(menuBar);

		this.views = new OptimizerViews(this.optim);
		this.optim.addOptimizerListener(new OptimizerListener() {
			@Override
			public void optimizerPerformed(OptimizerEvent event) {
				String name;

				switch (event.getType()) {
				case OptimizerEvent.STATUS:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// Update project status
							OptimizerGui.this.refreshMenu();
						}
					});
					break;
				case OptimizerEvent.PLOT:
					if (event.getMessage().startsWith("Add: ")) {
						// add plot to tabbed-pane
						name = event.getMessage().substring(5);
						OptimizerGui.this.plots.addTab(name, OptimizerGui.this.views.getPlot(name));
						// OptimizerGui.this.plots.insertTab(name, null,
						// OptimizerGui.this.views.getPlot(name), null,
						// OptimizerGui.this.plots.getTabCount() - 1);
						OptimizerGui.this.plots.setSelectedIndex(OptimizerGui.this.views.getNumPlots() - 1);
					} else if (event.getMessage().startsWith("Remove: ")) {
						// remove plot from tabbed-pane
						name = event.getMessage().substring(8);
						for (int k = OptimizerGui.this.plots.getTabCount() - 1; k >= 0; k--) {
							if (OptimizerGui.this.plots.getTitleAt(k).compareTo(name) == 0) {
								OptimizerGui.this.plots.removeTabAt(k);
							}
						}
						// if (OptimizerGui.this.plots.getSelectedIndex() ==
						// (OptimizerGui.this.plots
						// .getTabCount() - 1)) {
						// OptimizerGui.this.plots
						// .setSelectedIndex(OptimizerGui.this.plots
						// .getTabCount() - 2);
						// }
					}
					break;
				}
			}
		});

		problem = this.optim.getProblem();
		this.setTitle("Optimize - " + problem.getArchive().getMop().getPath() + ":" + problem.getArchive().getName()
				+ ":" + problem.getName());
		this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				new Thread() {
					@Override
					public void run() {
						OptimizerGui.this.setVisible(false);
						OptimizerGui.this.views.shutdown();
						OptimizerGui.this.dispose();
						if (!MatlabControl.isInsideMatlab()) {
							System.exit(0);
						}
					}
				}.start();
			}
		});
		this.setBounds(100, 100, 1000, 800);

		contentPane = new JPanel();
		contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
		this.setContentPane(contentPane);
		contentPane.setLayout(new BorderLayout(0, 0));

		this.tabbedPane = new JTabbedPane(SwingConstants.TOP);
		this.tabbedPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		contentPane.add(this.tabbedPane, BorderLayout.CENTER);

		panelProj = new JPanel();
		panelProj.setLayout(new BorderLayout(0, 0));
		panelProj.add(this.initProjectPanel());
		this.tabbedPane.addTab("Project", null, panelProj, null);

		panelMap = new JPanel();
		panelMap.setLayout(new BorderLayout(0, 0));
		JScrollPane mapScrollPane = new JScrollPane(this.views.getMappingPanel(),
				ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		// mapScrollPane.setMinimumSize(new Dimension(200, 200));
		// mapScrollPane.setPreferredSize(new Dimension(200, 200));
		mapScrollPane.setViewportView(this.views.getMappingPanel());

		panelMap.add(mapScrollPane, BorderLayout.CENTER);
		this.tabbedPane.addTab("Mappings", null, panelMap, null);

		panelSensitivity = new JPanel();
		panelSensitivity.setLayout(new BorderLayout(0, 0));

		JScrollPane sensScrollPane = new JScrollPane(this.views.getSensitivityPanel(),
				ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		// sensScrollPane.setMinimumSize(new Dimension(200, 200));
		// sensScrollPane.setPreferredSize(new Dimension(200, 200));
		sensScrollPane.setViewportView(this.views.getSensitivityPanel());

		panelSensitivity.add(sensScrollPane, BorderLayout.CENTER);
		if (OptimizerGui.enableSensitivity) {
			this.tabbedPane.addTab("Sensitivity", null, panelSensitivity, null);
		}

		panelLog = new JPanel();
		panelLog.setLayout(new BorderLayout(0, 0));
		panelLog.add(this.views.getLogging(), BorderLayout.CENTER);
		this.tabbedPane.addTab("Logging", null, panelLog, null);

		panelComp = new JPanel();
		panelComp.setLayout(new BorderLayout(0, 0));
		panelComp.add(this.views.getComparePanel());
		this.tabbedPane.addTab("Compare", null, panelComp, null);

		contentPane.add(this.views.getControl(), BorderLayout.NORTH);

		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

		this.setVisible(true);
	}

	/**
	 * Init panel for project view
	 * 
	 * @return Panel for project
	 */
	private JPanel initProjectPanel() {
		JPanel projectPanel;

		projectPanel = new JPanel();
		projectPanel.setLayout(new BorderLayout(0, 0));
		projectPanel.add(this.views.getPool(), BorderLayout.WEST);

		this.plots = new JTabbedPane();
		this.plots.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		projectPanel.add(this.plots, BorderLayout.CENTER);

		for (int k = 0; k < this.views.getNumPlots(); k++) {
			this.plots.addTab(this.views.getPlotName(k), this.views.getPlot(k));
		}

		// FIXME __delete
		// this.plots.addTab("+", null);
		//
		// // add tab with user-defined plot
		// this.plots.addChangeListener(new ChangeListener() {
		// @Override
		// public void stateChanged(ChangeEvent event) {
		// System.out.println("Changed");
		// String name;
		// int index = ((JTabbedPane) event.getSource())
		// .getSelectedIndex();
		// if ((index != -1)
		// && (((JTabbedPane) event.getSource()).getTitleAt(index)
		// .compareTo("+") == 0)) {
		// name = JOptionPane.showInputDialog(OptimizerGui.this,
		// "Name for new plot", "Add plot",
		// JOptionPane.QUESTION_MESSAGE);
		//
		// if ((name != null) && !name.isEmpty()) {
		// OptimizerGui.this.views.addPlot(name);
		// }
		// }
		// }
		// });
		//
		// // remove tab with userdefined plot
		// this.plots.addMouseListener(new MouseAdapter() {
		// @Override
		// public void mouseReleased(MouseEvent event) {
		// if (event.isPopupTrigger()) {
		// JPopupMenu menu = new JPopupMenu();
		// JMenuItem item = new JMenuItem("Close");
		//
		// item.setEnabled(OptimizerGui.this.views.getPlotPanel()
		// .getUserPlot(
		// OptimizerGui.this.plots
		// .getTitleAt(OptimizerGui.this.plots
		// .getSelectedIndex())) != null);
		// item.addActionListener(new ActionListener() {
		// @Override
		// public void actionPerformed(ActionEvent event) {
		// OptimizerGui.this.views
		// .rmPlot(OptimizerGui.this.plots
		// .getTitleAt(OptimizerGui.this.plots
		// .getSelectedIndex()));
		// }
		// });
		// menu.add(item);
		// menu.show(event.getComponent(), event.getX(), event.getY());
		// }
		// }
		// });

		return projectPanel;
	}

	/**
	 * Init display menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuDisplay(JMenuBar menuBar) {
		JMenu mnDisplay;
		JMenuItem mntmRefresh, mntmJobManager;

		mnDisplay = new JMenu("Display");
		menuBar.add(mnDisplay);

		mntmRefresh = new JMenuItem("Refresh");
		mntmRefresh.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.refreshDisplay();
			}
		});
		mnDisplay.add(mntmRefresh);

		mntmJobManager = new JMenuItem("JobManager");
		mntmJobManager.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JobDesc.setVisible(true);
			}
		});
		mnDisplay.add(mntmJobManager);

	}

	/**
	 * Init file menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuFile(JMenuBar menuBar) {
		JMenu mnFile;
		JMenuItem mntmExit;

		mnFile = new JMenu("File");
		menuBar.add(mnFile);

		mntmExit = new JMenuItem("Exit");
		mntmExit.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				new Thread() {
					@Override
					public void run() {
						OptimizerGui.this.views.shutdown();
						OptimizerGui.this.dispose();
						if (!MatlabControl.isInsideMatlab()) {
							System.exit(0);
						}
					}
				}.start();
			}
		});
		mnFile.add(mntmExit);
	}

	/**
	 * Init mapping menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuMapping(JMenuBar menuBar) {
		JMenu mnMapping;

		mnMapping = new JMenu("Mapping");
		menuBar.add(mnMapping);

		this.mntmCreateMapping = new JMenuItem("Create");
		this.mntmCreateMapping.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.createMapping();
			}
		});
		mnMapping.add(this.mntmCreateMapping);

		this.mntmRunMapping = new JMenuItem("Run");
		this.mntmRunMapping.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.runMapping();
			}
		});
		mnMapping.add(this.mntmRunMapping);
	}

	/**
	 * Init project menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuProject(JMenuBar menuBar) {
		JMenu mnProject;

		mnProject = new JMenu("Project");
		menuBar.add(mnProject);

		this.mntmRunProject = new JMenuItem("Run");
		this.mntmRunProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.runProject();
			}
		});
		mnProject.add(this.mntmRunProject);

		this.mntmFinishProject = new JMenuItem("Finish");
		this.mntmFinishProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.finishProject();
			}
		});
		mnProject.add(this.mntmFinishProject);

		this.mntmStopProject = new JMenuItem("Stop");
		this.mntmStopProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				OptimizerGui.this.views.stopProject();
			}
		});
		mnProject.add(this.mntmStopProject);

		mnProject.addSeparator();

		this.mntmInitProjectBuffer = new JMenuItem("Init buffer");
		this.mntmInitProjectBuffer.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.initProjectBuffer();
			}
		});
		mnProject.add(this.mntmInitProjectBuffer);

		this.mntmClearProjectBuffer = new JMenuItem("Clear buffer");
		this.mntmClearProjectBuffer.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.clearProjectBuffer();
			}
		});
		mnProject.add(this.mntmClearProjectBuffer);

		mnProject.addSeparator();

		this.mntmWriteToExel = new JMenuItem("Write pareto-front to excel");
		this.mntmWriteToExel.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.writeProjectToExcel();
			}
		});
		mnProject.add(this.mntmWriteToExel);

		this.mntmSelectPareto = new JMenuItem("Select pareto-front");
		this.mntmSelectPareto.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				OptimizerGui.this.views.selectPareto(true);
			}
		});
		mnProject.add(this.mntmSelectPareto);

	}

	/**
	 * Init setting menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuSetting(JMenuBar menuBar) {
		JMenu mnSettings;
		JMenuItem mntmProblem, mntmVariables, mntmResults, mntmGeneral, mntmSolver;

		mnSettings = new JMenu("Settings");
		menuBar.add(mnSettings);

		mntmProblem = new JMenuItem("Problem");
		mntmProblem.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JDialog dia;
				ChooseProblem panel;
				ComputeProject cProject;

				cProject = OptimizerGui.this.optim.getProjectControl();
				dia = new JDialog(OptimizerGui.this, "Problem", false);
				panel = new ChooseProblem(cProject.getProblem());
				panel.setSettingEnabled(false);
				dia.getContentPane().add(panel);
				dia.setSize(670, 500);
				dia.setLocationByPlatform(true);
				dia.setVisible(true);

			}
		});
		mnSettings.add(mntmProblem);

		mntmVariables = new JMenuItem("Variables");
		mntmVariables.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JDialog dia;
				ChooseVariables panel;
				ComputeProject cProject;

				cProject = OptimizerGui.this.optim.getProjectControl();
				dia = new JDialog(OptimizerGui.this, "Variables", false);
				panel = new ChooseVariables(cProject.getProblem());
				panel.setSettingEnabled(false);
				dia.getContentPane().add(panel);
				dia.setSize(670, 500);
				dia.setLocationByPlatform(true);
				dia.setVisible(true);
			}
		});
		mnSettings.add(mntmVariables);

		mntmResults = new JMenuItem("Results");
		mntmResults.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JDialog dia;
				ChooseResults panel;
				ComputeProject cProject;

				cProject = OptimizerGui.this.optim.getProjectControl();
				dia = new JDialog(OptimizerGui.this, "Results", false);
				panel = new ChooseResults(cProject.getProblem());
				panel.setSettingEnabled(false);
				dia.getContentPane().add(panel);
				dia.setSize(670, 500);
				dia.setLocationByPlatform(true);
				dia.setVisible(true);
			}
		});
		mnSettings.add(mntmResults);

		mntmGeneral = new JMenuItem("General");
		mntmGeneral.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JDialog dia;
				GeneralSettings panel;

				dia = new JDialog(OptimizerGui.this, "General settings", false);
				panel = new GeneralSettings(OptimizerGui.this.optim.getProjectControl().getProblem());
				panel.setSettingEnabled(false);
				dia.getContentPane().add(panel);
				dia.setSize(670, 500);
				dia.setLocationByPlatform(true);
				dia.setVisible(true);

			}
		});
		mnSettings.add(mntmGeneral);

		mntmSolver = new JMenuItem("Solver");
		mntmSolver.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JDialog dia;
				SolverSettings panel;

				dia = new JDialog(OptimizerGui.this, "Solver settings", false);
				panel = new SolverSettings(OptimizerGui.this.optim.getProjectControl().getProblem().getSolver());
				panel.setSettingEnabled(false);
				dia.getContentPane().add(panel);
				dia.setSize(670, 500);
				dia.setLocationByPlatform(true);
				dia.setVisible(true);
			}
		});
		mnSettings.add(mntmSolver);

	}

	/**
	 * Init debug menu
	 * 
	 * @param menuBar
	 *            Parent menu-bar
	 */
	private void initMenuDebug(JMenuBar menuBar) {
		JMenuItem mntmAssignProject, mntmAssignConroller;
		JMenu mnPlotObjectives, mnPlotVariableHist, mnPlotVariable;

		this.mnDebug = new JMenu("Debug");
		this.mnDebug.setVisible(false);
		menuBar.add(this.mnDebug);

		mntmAssignProject = new JMenuItem("Assign Project to Workspace");
		mntmAssignProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				MatlabProxy proxy;

				proxy = MatlabControl.getProxy();
				try {
					if (proxy.isRunningInsideMatlab()) {
						proxy.setVariable("OptProject", OptimizerGui.this.optim.getProjectControl().getProject());
					} else {
						try {
							proxy.setVariable("OptProject",
									OptimizerGui.this.optim.getProjectControl().getProject().getRMIImpl());
						} catch (RemoteException e) {
							// TODO RMI
							e.printStackTrace();
						}
					}
				} catch (MatlabInvocationException e) {
					e.printStackTrace();
				}
			}
		});
		this.mnDebug.add(mntmAssignProject);

		mntmAssignConroller = new JMenuItem("Assign Controller to Workspace");
		mntmAssignConroller.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				try {
					MatlabControl.getProxy().setVariable("OptController", OptimizerGui.this.optim);
				} catch (MatlabInvocationException e) {
					e.printStackTrace();
				}
			}
		});
		this.mnDebug.add(mntmAssignConroller);

		// this.mnDebug.addSeparator();
		//
		// mnPlotObjectives = new JMenu("Plot objectives");
		// this.mnDebug.add(mnPlotObjectives);
		// mnPlotObjectives.add(new JCheckBoxMenuItem(""));
		//
		// mnPlotVariableHist = new JMenu("Plot variable histrogram");
		// this.mnDebug.add(mnPlotVariableHist);
		// mnPlotVariableHist.add(new JCheckBoxMenuItem(""));
		//
		// mnPlotVariable = new JMenu("Plot variable");
		// this.mnDebug.add(mnPlotVariable);
		// mnPlotVariable.add(new JCheckBoxMenuItem(""));
	}

	/**
	 * Refresh user-interfaces
	 */
	private void refreshMenu() {
		String str;
		Boolean visible = false;

		// project
		this.mntmRunProject.setEnabled(this.views.isRunProjectEnabled());
		this.mntmFinishProject.setEnabled(this.views.isFinishProjectEnabled());
		this.mntmStopProject.setEnabled(this.views.isStopProjectEnabled());

		this.mntmClearProjectBuffer.setEnabled(this.views.isClearBufferEnabled());
		this.mntmWriteToExel.setEnabled(this.views.isWriteExcelEnabled());
		this.mntmSelectPareto.setEnabled(this.views.isSelectEnabled());

		// mapping
		// this.mntmCreateMapping.setEnabled(this.views.isCreateMappingEnabled());
		// this.mntmRunMapping.setEnabled(this.views.isRunMappingEnabled());

		// debug
		str = MagOptConfig.getConfig("OptimizerDebug");
		if (str != null) {
			try {
				visible = Boolean.parseBoolean(str);
			} catch (Exception ex) {
			}
		}
		this.mnDebug.setVisible(visible);
	}
}
