/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui;

import java.awt.BorderLayout;
import java.awt.Desktop;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.general.plot.Plot;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.utils.MatlabControl;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.event.OptimizerListener;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.project.ComputeProject;
import at.lcm.bu21.optim.gui.mapping.PanelMapping;
import at.lcm.bu21.optim.gui.mapping.PanelSensitivity;
import at.lcm.bu21.optim.gui.project.DialogStopCondition;
import at.lcm.bu21.optim.gui.project.PanelCompare;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import at.lcm.bu21.optim.gui.project.PanelProjectOverview;
import magopt.gui.event.EventView;
import magopt.utils.MopFile;

/**
 * View-control for the optimizer
 * 
 * @author Werner Koppelsttter
 */
public class OptimizerViews {

	/** Optimization */
	private Optimization optim;

	/** Event view */
	private EventView logging = null;

	/** Project pool panel */
	private PanelProjectOverview pool = null;

	/** Control panel */
	private PanelControl control = null;

	/** Plot panels */
	private PanelPlots plots = null;

	/** Mapping panel */
	private PanelMapping mapping = null;

	/** Sensitivity panel */
	private PanelSensitivity sensitivity = null;

	/** Compare panel */
	private PanelCompare compare;

	/**
	 * Dialog for number of individual to initialize
	 * 
	 * @author Werner Koppelsttter
	 */
	public class DialogInit extends JDialog {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Spinner for number to initialize */
		JSpinner spinnerInit;

		/**
		 * Create the dialog
		 */
		public DialogInit() {
			JButton btnApply;
			JPanel spinnerPanel, buttonPanel;

			this.setTitle("Initialize buffer");
			this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			this.getContentPane().setLayout(new BorderLayout());
			this.setModal(true);

			// spinner
			this.spinnerInit = new JSpinner();
			this.spinnerInit.setModel(
					new SpinnerNumberModel(new Integer(OptimizerViews.this.optim.getProblem().getAutosaveInterval()),
							new Integer(1), null, new Integer(10)));
			spinnerPanel = new JPanel(new BorderLayout());
			spinnerPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
			spinnerPanel.add(this.spinnerInit);
			this.getContentPane().add(spinnerPanel, BorderLayout.CENTER);

			// button
			btnApply = new JButton("Apply");
			btnApply.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					OptimizerViews.this.optim.initProjectBuffer((Integer) DialogInit.this.spinnerInit.getValue());
					DialogInit.this.setVisible(false);
					DialogInit.this.dispose();
				}
			});
			buttonPanel = new JPanel(new BorderLayout());
			buttonPanel.setBorder(new EmptyBorder(0, 5, 5, 5));
			buttonPanel.add(btnApply);
			this.getContentPane().add(buttonPanel, BorderLayout.SOUTH);

			// this.setSize(100, 100);
			this.pack();
			this.setLocationByPlatform(true);
			this.setVisible(true);
		}
	}

	/**
	 * Create view-control for the optimizer
	 * 
	 * @param controller
	 *            Controller of the optimizer
	 * @throws Exception
	 */
	public OptimizerViews(final Optimization controller) throws Exception {
		this.optim = controller;

		this.compare = new PanelCompare(this);

		controller.addOptimizerListener(new OptimizerListener() {
			@Override
			public void optimizerPerformed(final OptimizerEvent event) {
				switch (event.getType()) {
				case OptimizerEvent.LOG:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							if (OptimizerViews.this.logging != null) {
								OptimizerViews.this.getLogging().refresh();
							}
						}
					});
					break;
				case OptimizerEvent.PROBLEM:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// refresh overview panel
							controller.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
							// refresh objective plots
							if (OptimizerViews.this.plots != null) {
								OptimizerViews.this.getPlotPanel().refresh();
							}
						}
					});
					break;
				case OptimizerEvent.STATUS:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// Update project status
							if (OptimizerViews.this.pool != null) {
								OptimizerViews.this.getPool().refresh();
							}
							if (OptimizerViews.this.control != null) {
								OptimizerViews.this.getControl().refresh();
							}
							// if (OptimizerViews.this.mapping != null) {
							// OptimizerViews.this.getMappingPanel().refresh();
							// }
							// if (OptimizerViews.this.sensitivity != null) {
							// OptimizerViews.this.getSensitivityPanel()
							// .refresh();
							// }
						}
					});
					break;
				case OptimizerEvent.POOL:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// Update project status
							if (OptimizerViews.this.pool != null) {
								OptimizerViews.this.getPool().refresh();
							}
						}
					});
					break;
				case OptimizerEvent.PLOT:
					if (OptimizerViews.this.plots != null) {
						OptimizerViews.this.plots.performEvent(event);
					}
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// refresh plots
							controller.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
						}
					});
					break;
				case OptimizerEvent.MAPPING:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// refresh mapping panel
							if (OptimizerViews.this.mapping != null) {
								OptimizerViews.this.getMappingPanel().refresh();
							}
						}
					});
					break;
				case OptimizerEvent.SENSITIVITY:
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							// refresh mapping panel
							if (OptimizerViews.this.sensitivity != null) {
								OptimizerViews.this.getSensitivityPanel().refresh();
							}
						}
					});
					break;
				}
			}
		});
	}

	/**
	 * Get controller of the optimizer
	 * 
	 * @return Controller
	 */
	public Optimization getOptimization() {
		return this.optim;
	}

	/**
	 * Get the pool-panel of the optimizer
	 * 
	 * @return Pool-panel
	 */
	public PanelProjectOverview getPool() {
		if (this.pool == null) {
			this.pool = new PanelProjectOverview();
			this.pool.set(this.optim.getProjectControl());
		}

		return this.pool;
	}

	/**
	 * Get the control-panel of the optimizer
	 * 
	 * @return Control-panel
	 */
	public PanelControl getControl() {
		if (this.control == null) {
			try {
				this.control = new PanelControl(this);
			} catch (Exception e) {
			}
		}

		return this.control;
	}

	/**
	 * Get the logging-panel of the optimizer
	 * 
	 * @return Logging-panel
	 */
	public EventView getLogging() {
		if (this.logging == null) {
			this.logging = new EventView();
			this.logging.setEventLog(this.optim.getEventLog());
		}

		return this.logging;
	}

	/**
	 * List types of plots
	 * 
	 * @return List of names
	 */
	public String[] listPlotTypes() {
		PanelPlots plots;

		plots = this.getPlotPanel();

		return PanelPlots.listPlotTypes();
	}

	/**
	 * Get type of plot
	 * 
	 * @return Plot type
	 */
	public String getPlotType() {
		PanelPlots plots;

		plots = this.getPlotPanel();

		return plots.getPlotType();

	}

	/**
	 * Set type of plot
	 * 
	 * @param plotType
	 *            Plot type
	 */
	public void setPlotType(String plotType) {
		PanelPlots plots;

		plots = this.getPlotPanel();

		plots.setPlotType(plotType);
	}

	/**
	 * Get number of existing plots
	 * 
	 * @return Number of plots
	 */
	public int getNumPlots() {
		PanelPlots plots;

		plots = this.getPlotPanel();

		return plots.getNumFixPlots() + plots.getNumUserPlots();
	}

	/**
	 * Get name of plot
	 * 
	 * @param index
	 *            Plot index
	 * @return Name
	 */
	public String getPlotName(int index) {
		String name;
		PanelPlots plots;

		plots = this.getPlotPanel();

		if (index < plots.getNumFixPlots()) {
			name = plots.getFixPlotName(index);
		} else {
			name = plots.getUserPlotName(index - plots.getNumFixPlots());
		}

		return name;
	}

	// FIMXE Plot-panel
	/**
	 * Get plot-panel
	 * 
	 * @return Plot-panels
	 */
	public PanelPlots getPlotPanel() {
		if (this.plots == null) {
			try {
				this.plots = new PanelPlots(this.optim);
			} catch (Exception e) {
			}
		}
		return this.plots;
	}

	/**
	 * Get plot-panel of index
	 * 
	 * @param index
	 *            Plot index
	 * @return Panel of plot
	 */
	public JPanel getPlot(int index) {
		JPanel panel;
		PanelPlots plots;

		plots = this.getPlotPanel();

		if (index < plots.getNumFixPlots()) {
			panel = plots.getFixPlot(index).getPanel();
		} else {
			panel = plots.getUserPlot(index - plots.getNumFixPlots()).getPanel();
		}

		return panel;
	}

	/**
	 * Get plot-panel for name
	 * 
	 * @param name
	 *            Name of plot
	 * @return Panel of plot
	 */
	public JPanel getPlot(String name) {
		JPanel panel = null;
		Plot plot;
		PanelPlots plots;

		plots = this.getPlotPanel();

		plot = plots.getFixPlot(name);
		if (plot == null) {
			plot = plots.getUserPlot(name);
		}
		if (plot != null) {
			panel = plot.getPanel();
		}
		return panel;
	}

	/**
	 * Get mapping-panel of optimizer
	 * 
	 * @return Mapping-panel
	 */
	public PanelMapping getMappingPanel() {
		if (this.mapping == null) {
			this.mapping = new PanelMapping(this.optim);
		}

		return this.mapping;
	}

	/**
	 * Get sensitivity-panel of optimizer
	 * 
	 * @return Mapping-panel
	 */
	public PanelSensitivity getSensitivityPanel() {
		if (this.sensitivity == null) {
			this.sensitivity = new PanelSensitivity(this.optim);
		}

		return this.sensitivity;
	}

	/**
	 * Get compare-panel of optimizer
	 * 
	 * @return Compare-panel
	 */
	public PanelCompare getComparePanel() {
		return this.compare;
	}

	/**
	 * Get number of save-points to plot
	 * 
	 * @return Number save-points to plot
	 */
	public int getNumPlotSavepoints() {
		return this.getPlotPanel().getNumPlotSavepoints();
	}

	/**
	 * Set number of save-points to plot
	 * 
	 * @param num
	 *            Number save-points to plot
	 */
	public void setNumPlotSavepoints(int num) {
		if (num < 1) {
			num = 1;
		}
		this.getPlotPanel().setNumPlotSavepoints(num);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
	}

	/**
	 * Get mark mapping
	 * 
	 * @return Mark mapping
	 */
	public boolean getMarkMapping() {
		return this.getPlotPanel().getMarkMapping();
	}

	/**
	 * Set mark mapping
	 * 
	 * @param mark
	 *            Mark mapping
	 */
	public void setMarkMapping(boolean mark) {
		this.getPlotPanel().setMarkMapping(mark);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
	}

	/**
	 * Get plot hypervolume
	 * 
	 * @return Plot hypervolume
	 */
	public boolean getPlotHypervolume() {
		return this.getPlotPanel().getPlotHypervolume();
	}

	/**
	 * Set plot hypervolume
	 * 
	 * @param plot
	 *            Plot hypervolume
	 */
	public void setPlotHypervolume(boolean plot) {
		this.getPlotPanel().setPlotHypervolume(plot);
		this.optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
	}

	/**
	 * Select individual and create project
	 */
	public void selectPareto() {
		this.selectPareto(false);
	}

	/**
	 * Select individual and create project
	 * 
	 * @param open
	 *            Open created project
	 */
	public void selectPareto(final boolean open) {
		final Individual[] list;
		final int[] selected, suppressed;
		final MopFile mop;
		final boolean doc;
		JFileChooser chooser;

		list = OptimizerViews.this.getPlotPanel().getListEnabledIndividuals();
		selected = OptimizerViews.this.getPlotPanel().getSelectedIndices();
		suppressed = OptimizerViews.this.getPlotPanel().getSuppressedIndices();

		if (selected.length > 0) {
			chooser = new JFileChooser("Save MagOpt-Project as");
			chooser.setCurrentDirectory(OptimizerViews.this.optim.getProblem().getArchive().getMop().getParentFile());
			chooser.setFileFilter(new FileNameExtensionFilter("MagOpt Project File (*.mop)", "mop"));
			if (chooser.showSaveDialog(OptimizerViews.this.control) == JFileChooser.APPROVE_OPTION) {
				mop = new MopFile(chooser.getSelectedFile());
				if (!mop.exists() || (JOptionPane.showConfirmDialog(OptimizerViews.this.control,
						"File already exists. Do yout want to override the file?", "Confirm override",
						JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)) {
					doc = JOptionPane.showConfirmDialog(OptimizerViews.this.control, "Create doc files?",
							"Documentation", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION;

					new Thread() {
						@Override
						public void run() {
							Project project;

							try {
								project = OptimizerViews.this.optim.selectPareto(mop, list, selected, suppressed, doc);
								project.save();

								if ((project != null) && open) {
									if (MatlabControl.isInsideMatlab()) {
										try {
											MatlabControl.getProxy().feval("MagOptGui", new Object[] { project });
										} catch (Throwable e) {
											e.printStackTrace();
										}
									} else {
										if (Desktop.isDesktopSupported()) {
											try {
												Desktop.getDesktop().open(project.getMopFile());
											} catch (IOException e) {
												// FIXME winopen
												OptimizerViews.this.optim.getEventLog().warning("",
														"Unable to open (Exception while open)", e.getStackTrace());
											}
										} else {
											OptimizerViews.this.optim.getEventLog().warning("",
													"Unable to open (Unsupported)", new Exception().getStackTrace());
										}

									}
								}
							} catch (Exception ex) {
								Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
							}
						}
					}.start();
				}
			}
		} else {
			JOptionPane.showMessageDialog(OptimizerViews.this.control, "No global pareto is selected!",
					"Select optimum", JOptionPane.OK_OPTION);
		}
	}

	/**
	 * Select single global Pareto point for sensitivity analysis
	 */
	public void selectSingleParetoPointForSensitivityAnalysis() {
		final int[] selected;

		selected = OptimizerViews.this.getPlotPanel().getSelectedIndices();

		if (selected.length < 1) {
			JOptionPane.showMessageDialog(OptimizerViews.this.control, "No global Pareto-optimal design is selected!",
					"Select design for Sensitivity Analysis", JOptionPane.OK_OPTION);
		} else if (selected.length > 1) {
			JOptionPane.showMessageDialog(OptimizerViews.this.control,
					"Multiple Pareto-optimal designs selected! Please choose just one.",
					"Select design for Sensitivity Analysis", JOptionPane.OK_OPTION);
		} else {
			// JOptionPane.showMessageDialog(OptimizerViews.this.control,
			// "Design nr." + selected[0] + " selected for SA.",
			// "Select design for Sensitivity Analysis",
			// JOptionPane.OK_OPTION);
			Individual chosenInd = this.getOptimization().getProblem()
					.getSavePoint(this.getOptimization().getProblem().getNumSavePoints() - 1)
					.getFullPareto()[selected[0]];

			this.optim.initializeSensitivityAnalysis(chosenInd);
		}
	}

	/**
	 * Write pareto to excel
	 */
	public void writeProjectToExcel() {
		final File excel;

		JFileChooser chooser = new JFileChooser();
		chooser.setCurrentDirectory(this.optim.getProblem().getArchive().getMop().getParentFile());
		chooser.setFileFilter(new FileNameExtensionFilter("Excel File (*.xlsx / *.xls)", "xls", "xlsx"));

		if (chooser.showSaveDialog(this.control) == JFileChooser.APPROVE_OPTION) {
			excel = chooser.getSelectedFile();
			new Thread() {
				@Override
				public void run() {
					File file;
					try {
						file = OptimizerViews.this.optim.writeProjectToExcel(excel);

						if (file.exists()) {
							try {
								if (Desktop.isDesktopSupported()) {
									Desktop.getDesktop().open(file);
								}
							} catch (Exception ex) {
							}
						}
					} catch (Exception ex) {
						Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
					}

				}
			}.start();
		}
	}

	/**
	 * Shutdown Optimizer
	 */
	public void shutdown() {
		this.optim.shutdown();
	}

	/**
	 * Init project buffer
	 */
	public void initProjectBuffer() {
		new DialogInit();
		// this.controller.initProjectBuffer();
	}

	/**
	 * Clear project-buffer
	 */
	public void clearProjectBuffer() {
		this.optim.clearProjectBuffer();
	}

	/**
	 * Clear logger
	 */
	public void clearLog() {
		this.optim.clearLog();
	}

	/**
	 * Start project optimization
	 */
	public void runProject() {
		this.optim.getProjectControl().clearStopCondition();
		new DialogStopCondition(this.optim.getProjectControl());
		this.optim.runProject();
	}

	/**
	 * Finish project optimization
	 */
	public void finishProject() {
		this.optim.finishProject();
	}

	/**
	 * Stop project optimization
	 */
	public void stopProject() {
		this.optim.stopProject();
	}

	/**
	 * Set status of recompute (sensitivity)
	 * 
	 * @param recomp
	 *            Recompute mapping based
	 */
	public void setRecompute(boolean recomp) {
		this.optim.setRecompute(recomp);
	}

	/**
	 * Get status of recompute (sensitivity)
	 * 
	 * @return Recompute mapping based
	 */
	public boolean isRecompute() {
		return this.optim.isRecompute();
	}

	/**
	 * Create mapping optimization
	 */
	public void createMapping() {
		this.optim.createMapping();
	}

	/**
	 * Run mapping optimization
	 */
	public void runMapping() {
		this.optim.runMapping();
	}

	/**
	 * Refresh display
	 */
	public void refreshDisplay() {
		try {
			this.optim.refreshDisplay();
		} catch (Exception ex) {
			Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	/**
	 * Get status of optimizer
	 * 
	 * @return Status
	 */
	public String getStatus() {
		ComputeProject proj;
		String status = "Undefined";

		proj = this.getOptimization().getProjectControl();

		if (proj.isStop()) {
			if (proj.isRun()) {
				status = "Stopping";
			} else {
				status = "Stopped";
			}
		} else {
			if (proj.isAdd()) {
				status = "Running";
			} else {
				status = "Finishing";
			}
		}

		return status;
	}

	/**
	 * Is run project enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isRunProjectEnabled() {
		ComputeProject proj;

		proj = this.getOptimization().getProjectControl();

		return proj.isStop() || !proj.isRun() || !proj.isAdd();
	}

	/**
	 * Is finish project enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isFinishProjectEnabled() {
		ComputeProject proj;

		proj = this.getOptimization().getProjectControl();

		return proj.isStop() || !proj.isRun() || proj.isAdd();
	}

	/**
	 * Is stop project enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isStopProjectEnabled() {
		return this.getOptimization().getProjectControl().isRun();
	}

	/**
	 * Is clear buffer enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isClearBufferEnabled() {
		return this.getOptimization().getProjectControl().getBufferSize() > 0;
	}

	/**
	 * Is write to excel enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isWriteExcelEnabled() {
		return (this.getOptimization().getProjectControl().getProblem().getNumSavePoints() > 0)
				&& !this.getOptimization().getWriteExcel();
	}

	/**
	 * Is select pareto enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isSelectEnabled() {
		return (this.getOptimization().getProjectControl().getProblem().getNumSavePoints() > 0)
				&& !this.getOptimization().getSelectPareto();
	}

	/**
	 * Is create mapping enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isCreateMappingEnabled() {
		MappingManager map;

		map = this.getOptimization().getMappingControl();

		return (!map.areAnyMappingHoldersBusy() && !map.isOptimizationRunning() && map.hasSufficientSamples()
				&& map.isSensitivityAnalysisIdle());
	}

	/**
	 * Is run mapping enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isRunMappingEnabled() {
		MappingManager map;

		map = this.getOptimization().getMappingControl();

		return (map.areAllMappingHoldersReady() && !map.isOptimizationRunning() && map.isOptimizationIdle()
				&& map.isSensitivityAnalysisIdle());
	}

	/**
	 * Is sensitivity initialization enabled
	 * 
	 * @return Is enabled
	 */
	public boolean isSensitivityInitializationEnabled() {
		MappingManager map;

		map = this.getOptimization().getMappingControl();

		return (map.areAllMappingHoldersExceptErrorsClsReady() && map.isSensitivityAnalysisIdle());
	}

	/**
	 * Add user plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void addPlot(String name) {
		this.optim.addUserPlot(name);
	}

	/**
	 * Remove user plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void rmPlot(String name) {
		this.optim.rmUserPlot(name);
	}
}
