/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import at.lcm.bu21.optim.core.project.ComputeProject;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import jobmanager.JobDesc;

/**
 * Panel with control-elements
 * 
 * @author Werner Koppelsttter
 */
public class PanelControl extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** View-control of the optimizer */
	private OptimizerViews views;

	// display UIs
	/** Refresh panels */
	private JButton btnRefresh;
	/** Number of save-points */
	private JTextField txtSavePoints;
	/** Number of solved individual since last save-point */
	private JTextField txtSolved;
	/** Size of project buffer */
	private JTextField txtBuffer;
	/** Type of plot */
	private JComboBox cbPlotType;
	/** Number of save-points to plot */
	private JSpinner spinDisplayNum;

	// project buttons
	/** Run project optimization */
	private JButton btnRunProject;
	/** Finish project optimization */
	private JButton btnFinishProject;
	/** Stop project optimization */
	private JButton btnStopProject;
	/** Init project buffer */
	private JButton btnInitProjectBuffer;
	/** Clear project buffer */
	private JButton btnClearProjectBuffer;
	/** Write pareto to excel */
	private JButton btnWriteToExcel;
	/** Select optimum from global pareto */
	private JButton btnSelect;

	// mapping buttons
	/** Create mapping optimization */
	private JButton btnCreateMapping;
	/** Run mapping optimization */
	private JButton btnRunMapping;

	// plots
	/** Add user-plots */
	private JButton btnAddPlot;
	/** Remove user-plots */
	private JButton btnRemovePlot;

	// sensitivity
	/** Select individual for sensitivity analysis */
	private JButton btnSelectForSensitivityAnalysis;
	/** Recompute mapping based (sensitivity) */
	private JCheckBox chckbxRecompute;

	/**
	 * Create control-panbel for the controller
	 * 
	 * @param views
	 *            View-control of the optimizer
	 * @throws Exception
	 */
	public PanelControl(OptimizerViews views) throws Exception {
		this.views = views;

		this.setLayout(new GridLayout(0, 1, 0, 0));

		this.initPanelDisplay();
		this.initPanelProject();
		this.initPanelMapping();
	}

	/**
	 * Init display button panel
	 */
	private void initPanelDisplay() {
		JPanel panelDisplay;
		JLabel lblDisplay, lblSavepoints, lblSolved, lblBuffer, lblPlotType, lblDisplayNum;

		panelDisplay = new JPanel();
		this.add(panelDisplay);
		panelDisplay.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));

		lblDisplay = new JLabel("Project");
		lblDisplay.setFont(new Font("Tahoma", Font.BOLD, 11));
		lblDisplay.setPreferredSize(new Dimension(70, 20));
		panelDisplay.add(lblDisplay);

		this.btnRefresh = new JButton("Refresh");
		this.btnRefresh.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.refreshDisplay();
			}
		});
		this.btnRefresh.setPreferredSize(new Dimension(90, 30));
		this.btnRefresh.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_39.png")));
		panelDisplay.add(this.btnRefresh);

		lblSavepoints = new JLabel("SavePoints:");
		lblSavepoints.setFont(new Font("Tahoma", Font.PLAIN, 11));
		lblSavepoints.setPreferredSize(new Dimension(60, 20));
		lblSavepoints.setHorizontalAlignment(SwingConstants.RIGHT);
		panelDisplay.add(lblSavepoints);

		this.txtSavePoints = new JTextField();
		this.txtSavePoints.setPreferredSize(new Dimension(40, 20));
		this.txtSavePoints.setHorizontalAlignment(SwingConstants.RIGHT);
		this.txtSavePoints.setEditable(false);
		this.txtSavePoints.setText("0");
		panelDisplay.add(this.txtSavePoints);

		lblSolved = new JLabel("Solved Ind.:");
		lblSolved.setHorizontalAlignment(SwingConstants.RIGHT);
		lblSolved.setPreferredSize(new Dimension(60, 20));
		panelDisplay.add(lblSolved);

		this.txtSolved = new JTextField();
		this.txtSolved.setPreferredSize(new Dimension(40, 20));
		this.txtSolved.setHorizontalAlignment(SwingConstants.RIGHT);
		this.txtSolved.setEditable(false);
		this.txtSolved.setText("0");
		panelDisplay.add(this.txtSolved);

		lblBuffer = new JLabel("Buffer:");
		lblBuffer.setPreferredSize(new Dimension(40, 20));
		lblBuffer.setHorizontalAlignment(SwingConstants.RIGHT);
		panelDisplay.add(lblBuffer);

		this.txtBuffer = new JTextField();
		this.txtBuffer.setPreferredSize(new Dimension(40, 20));
		this.txtBuffer.setEditable(false);
		this.txtBuffer.setHorizontalAlignment(SwingConstants.RIGHT);
		this.txtBuffer.setText("0");
		panelDisplay.add(this.txtBuffer);

		lblPlotType = new JLabel("PlotType:");
		lblPlotType.setPreferredSize(new Dimension(50, 20));
		lblPlotType.setHorizontalAlignment(SwingConstants.RIGHT);
		panelDisplay.add(lblPlotType);

		this.cbPlotType = new JComboBox(this.views.listPlotTypes());
		this.cbPlotType.setPreferredSize(new Dimension(100, 20));
		this.cbPlotType.setEditable(false);
		this.cbPlotType.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				PanelControl.this.views.setPlotType((String) PanelControl.this.cbPlotType.getSelectedItem());
				PanelControl.this.cbPlotType.setSelectedItem(PanelControl.this.views.getPlotType());
			}
		});
		panelDisplay.add(this.cbPlotType);

		lblDisplayNum = new JLabel("Display last #SP:");
		lblDisplayNum.setPreferredSize(new Dimension(90, 20));
		lblDisplayNum.setHorizontalAlignment(SwingConstants.RIGHT);
		panelDisplay.add(lblDisplayNum);

		this.spinDisplayNum = new JSpinner();
		this.spinDisplayNum.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent event) {
				try {
					PanelControl.this.views.setNumPlotSavepoints((Integer) PanelControl.this.spinDisplayNum.getValue());
				} catch (Exception ex) {
				}
				PanelControl.this.spinDisplayNum.setValue(PanelControl.this.views.getNumPlotSavepoints());
			}
		});
		PanelControl.this.spinDisplayNum.setValue(PanelControl.this.views.getNumPlotSavepoints());
		this.spinDisplayNum.setPreferredSize(new Dimension(50, 20));
		panelDisplay.add(this.spinDisplayNum);

		JButton btnJobmanager = new JButton("JobManager");
		btnJobmanager.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_44.png")));
		btnJobmanager.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				JobDesc.setVisible(true);
			}
		});
		btnJobmanager.setPreferredSize(new Dimension(130, 30));
		panelDisplay.add(btnJobmanager);
	}

	/**
	 * Init mapping button panel
	 */
	private void initPanelMapping() {
		JPanel panelMapping;
		JLabel lblMapping;

		panelMapping = new JPanel();
		panelMapping.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
		this.add(panelMapping);

		lblMapping = new JLabel("Mappings");
		lblMapping.setFont(new Font("Tahoma", Font.BOLD, 11));
		lblMapping.setPreferredSize(new Dimension(70, 20));
		panelMapping.add(lblMapping);

		this.btnCreateMapping = new JButton("Create All");
		this.btnCreateMapping.setPreferredSize(new Dimension(135, 30));
		this.btnCreateMapping
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_40.png")));
		this.btnCreateMapping.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.createMapping();
			}
		});
		panelMapping.add(this.btnCreateMapping);

		this.btnRunMapping = new JButton("Map-Based-Opt");
		this.btnRunMapping.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.runMapping();
			}
		});
		this.btnRunMapping.setPreferredSize(new Dimension(135, 30));
		this.btnRunMapping
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_60.png")));
		panelMapping.add(this.btnRunMapping);

		this.btnSelectForSensitivityAnalysis = new JButton("");
		this.btnSelectForSensitivityAnalysis.setPreferredSize(new Dimension(30, 30));
		this.btnSelectForSensitivityAnalysis
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_38.png")));
		this.btnSelectForSensitivityAnalysis.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.selectSingleParetoPointForSensitivityAnalysis();
			}
		});
		if (OptimizerGui.enableSensitivity) {
			panelMapping.add(this.btnSelectForSensitivityAnalysis);
		} else {
			panelMapping.add(Box.createHorizontalStrut(30));
		}

		panelMapping.add(Box.createHorizontalStrut(20));

		JCheckBox chckbxMarkMapping = new JCheckBox("Mark Mapping");
		chckbxMarkMapping.setPreferredSize(new Dimension(120, 23));
		chckbxMarkMapping.setSelected(this.views.getPlotPanel().getMarkMapping());
		chckbxMarkMapping.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.getPlotPanel().setMarkMapping(((JCheckBox) event.getSource()).isSelected());
			}
		});
		panelMapping.add(chckbxMarkMapping);

		this.btnAddPlot = new JButton("Add Plot");
		this.btnAddPlot.setPreferredSize(new Dimension(130, 30));
		this.btnAddPlot.setIcon(new ImageIcon(PanelControl.class.getResource("/magopt/gui/icons/iconset1/001_01.png")));
		this.btnAddPlot.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				String name;

				name = JOptionPane.showInputDialog(PanelControl.this, "Name for new plot", "Add plot",
						JOptionPane.QUESTION_MESSAGE);

				if ((name != null) && !name.isEmpty()) {
					PanelControl.this.views.addPlot(name);
				}
			}
		});

		JCheckBox chckbxPlotHypervolume = new JCheckBox("Plot hypervolume");
		chckbxPlotHypervolume.setPreferredSize(new Dimension(120, 23));
		chckbxPlotHypervolume.setSelected(this.views.getPlotPanel().getPlotHypervolume());
		chckbxPlotHypervolume.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.getPlotPanel().setPlotHypervolume(((JCheckBox) event.getSource()).isSelected());
			}
		});
		panelMapping.add(chckbxPlotHypervolume);

		Component horizontalStrut = Box.createHorizontalStrut(20);
		panelMapping.add(horizontalStrut);
		panelMapping.add(this.btnAddPlot);

		this.btnRemovePlot = new JButton("Remove Plot");
		this.btnRemovePlot.setPreferredSize(new Dimension(130, 30));
		this.btnRemovePlot.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				String name;
				String[] plots;
				PanelPlots panel;

				panel = PanelControl.this.views.getPlotPanel();
				plots = new String[panel.getNumUserPlots()];
				for (int k = 0; k < plots.length; k++) {
					plots[k] = panel.getUserPlotName(k);
				}

				name = (String) JOptionPane.showInputDialog(PanelControl.this, "Select Plot to remove", "Remove Plot",
						JOptionPane.INFORMATION_MESSAGE, null, plots, plots[plots.length - 1]);

				if (name != null) {
					PanelControl.this.views.rmPlot(name);
				}
			}
		});
		this.btnRemovePlot
				.setIcon(new ImageIcon(PanelControl.class.getResource("/magopt/gui/icons/iconset1/001_02.png")));
		panelMapping.add(this.btnRemovePlot);

		panelMapping.add(Box.createHorizontalStrut(20));

	}

	/**
	 * Init project button panel
	 */
	private void initPanelProject() {
		JPanel panelProject;
		JLabel lblProject;
		Component horStrut2;

		panelProject = new JPanel();
		panelProject.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
		this.add(panelProject);

		lblProject = new JLabel("");
		lblProject.setFont(new Font("Tahoma", Font.BOLD, 11));
		lblProject.setPreferredSize(new Dimension(70, 20));
		panelProject.add(lblProject);

		this.btnRunProject = new JButton("Run");
		this.btnRunProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.runProject();
			}
		});
		this.btnRunProject.setPreferredSize(new Dimension(90, 30));
		this.btnRunProject
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_60.png")));
		panelProject.add(this.btnRunProject);

		this.btnFinishProject = new JButton("Finish");
		this.btnFinishProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.finishProject();
			}
		});
		this.btnFinishProject.setPreferredSize(new Dimension(90, 30));
		this.btnFinishProject
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_10.png")));
		panelProject.add(this.btnFinishProject);

		this.btnStopProject = new JButton("Stop");
		this.btnStopProject.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.stopProject();
			}
		});
		this.btnStopProject.setPreferredSize(new Dimension(90, 30));
		this.btnStopProject
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_29.png")));
		panelProject.add(this.btnStopProject);

		this.chckbxRecompute = new JCheckBox("Recompute");
		this.chckbxRecompute.setToolTipText("Recompute Sensitivity");
		this.chckbxRecompute.setPreferredSize(new Dimension(50, 23));
		this.chckbxRecompute.setSelected(this.views.isRecompute());
		this.chckbxRecompute.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.setRecompute(((JCheckBox) event.getSource()).isSelected());
			}
		});
		if (OptimizerGui.enableSensitivity) {
			panelProject.add(this.chckbxRecompute);
		} else {
			panelProject.add(Box.createHorizontalStrut(50));
		}

		this.btnInitProjectBuffer = new JButton("Init Buffer");
		this.btnInitProjectBuffer.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.initProjectBuffer();
			}
		});
		this.btnInitProjectBuffer.setPreferredSize(new Dimension(120, 30));
		this.btnInitProjectBuffer
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_51.png")));
		panelProject.add(this.btnInitProjectBuffer);

		this.btnClearProjectBuffer = new JButton("Clear Buffer");
		this.btnClearProjectBuffer.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.clearProjectBuffer();
			}
		});
		this.btnClearProjectBuffer.setPreferredSize(new Dimension(120, 30));
		this.btnClearProjectBuffer
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_49.png")));
		panelProject.add(this.btnClearProjectBuffer);

		horStrut2 = Box.createHorizontalStrut(20);
		panelProject.add(horStrut2);

		this.btnWriteToExcel = new JButton("Write to Excel");
		this.btnWriteToExcel.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.writeProjectToExcel();
			}
		});
		this.btnWriteToExcel.setPreferredSize(new Dimension(130, 30));
		this.btnWriteToExcel
				.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_45.png")));
		panelProject.add(this.btnWriteToExcel);

		this.btnSelect = new JButton("Select");
		this.btnSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				PanelControl.this.views.selectPareto(true);
			}
		});
		this.btnSelect.setIcon(new ImageIcon(OptimizerGui.class.getResource("/magopt/gui/icons/iconset1/001_46.png")));
		this.btnSelect.setPreferredSize(new Dimension(130, 30));
		panelProject.add(this.btnSelect);
	}

	/**
	 * Refresh user-interfaces
	 */
	public void refresh() {
		int num;
		ComputeProject proj;

		proj = this.views.getOptimization().getProjectControl();
		// Update # of Generations
		num = proj.getProblem().getNumSavePoints();
		if (proj.getProblem().existTempSavePoint()) {
			num--;
		}
		this.txtSavePoints.setText(Integer.toString(num));

		// update # of solved
		this.txtSolved.setText(Integer.toString(proj.getNumberSolved()));

		// Update # of inits
		this.txtBuffer.setText(Integer.toString(proj.getBufferSize()));

		// project
		this.btnRunProject.setEnabled(this.views.isRunProjectEnabled());
		this.btnFinishProject.setEnabled(this.views.isFinishProjectEnabled());
		this.btnStopProject.setEnabled(this.views.isStopProjectEnabled());

		this.btnClearProjectBuffer.setEnabled(this.views.isClearBufferEnabled());
		this.btnWriteToExcel.setEnabled(this.views.isWriteExcelEnabled());
		this.btnSelect.setEnabled(this.views.isSelectEnabled());

		// mapping
		// this.btnCreateMapping.setEnabled(this.views.isCreateMappingEnabled());
		// this.btnRunMapping.setEnabled(this.views.isRunMappingEnabled());

		// plots
		this.btnRemovePlot.setEnabled(this.views.getPlotPanel().getNumUserPlots() > 0);

		// sensitivity analysis
		// this.btnSelectForSensitivityAnalysis.setEnabled(this.views
		// .isSensitivityInitializationEnabled());
	}
}
