/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui.project;

import java.awt.Adjustable;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JScrollBar;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.plot.HOLD;
import at.lcm.bu21.general.plot.LinkPlot;
import at.lcm.bu21.general.plot.Plot;
import at.lcm.bu21.general.plot.Plot2Hist;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.Plot2Scatter;
import at.lcm.bu21.general.plot.PlotException;
import at.lcm.bu21.general.plot.Signal;
import at.lcm.bu21.general.plot.SignalStyle;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Memory;
import at.lcm.bu21.optim.core.ObjectiveType;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.SavePoint;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.project.PlotDescription;
import jmetal.qualityIndicator.Hypervolume;

/**
 * Panel to plot objectives, variables and quality-function
 * 
 * @author Werner Koppelsttter
 */
public class PanelPlots {

	/** Plot-name of quality */
	public final static String PLOT_QUALITY = "Quality";

	/** Plot-name for objectives */
	public final static String PLOT_OBJECTIVES = "Objectives";

	/** Plot-name for variables histogram */
	public final static String PLOT_VARIABLES_HIST = "Variables";

	/** Plot-name for variables */
	public final static String PLOT_VARIABLES = "Variables (*)";

	/** Plot-type name of global pareto front */
	public final static String TYPE_GLOBAL = "Global pareto front";

	/** Plot-type name of memory */
	public final static String TYPE_MEMORY = "Memory archive";

	/** Plot-type name of local pareto front */
	public final static String TYPE_LOCAL = "Solver pareto front";

	/** Plot-type name of solver archive */
	public final static String TYPE_SOLVER = "Solver archive";

//	/** Controller of optimizer */
	// private OptimizerController control;
	/** Problem to display */
	private Problem problem;

	/** List of plot-names */
	// private static String[] fixPlotNames = new String[] { PLOT_QUALITY,
	// PLOT_OBJECTIVES, PLOT_VARIABLES_HIST, PLOT_VARIABLES };
	private static String[] fixPlotNames = new String[] { PanelPlots.PLOT_QUALITY, PanelPlots.PLOT_OBJECTIVES,
			PanelPlots.PLOT_VARIABLES_HIST };

	/** List of plot-types */
	// private static String[] plotTypes = new String[] { TYPE_GLOBAL,
	// TYPE_MEMORY, TYPE_LOCAL, TYPE_SOLVER };
	private static String[] plotTypes = new String[] { PanelPlots.TYPE_GLOBAL, PanelPlots.TYPE_MEMORY };

	/** Selected plot type */
	private String plotType = PanelPlots.TYPE_GLOBAL;

	/** Mark mapping individuals */
	private boolean markMapping = false;

	/** Plot hypervolume */
	private boolean plotHypervolume = true;

	/** Number of save-points to plot */
	private int plotSavepoints = 20;

	/** List of plots */
	private Plot[] fixPlots;

	/** Object to link global plots */
	private LinkPlot linkPlots;

	/** List of plots for all userplots */
	private Hashtable<String, SubPlot> userPlots = new Hashtable<String, SubPlot>();

	/** Current displayed individuals */
	private LinkedList<Individual[]> plotValues = new LinkedList<Individual[]>();

	/** Index of enabled index of Values */
	private int enabledPlotIndex = 0;

	/**
	 * Add context-menu entries for user-defined plots
	 * 
	 * @param plot
	 *            Plot to add context-menu
	 * @param desc
	 *            Description of plot-panel
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 */
	private void addMenu(final Plot2Scatter plot, final PlotDescription desc, final int row, final int column) {
		JMenuItem item;
		JPopupMenu popup;
		JMenu menu, submenu;

		// add items to popup-menu
		popup = plot.getChartPanel().getPopupMenu();
		popup.addSeparator();

		// x-axis
		submenu = new JMenu("x-axis");
		this.addParameterMenu(submenu, desc, row, column, true);
		popup.add(submenu);

		// y-axis
		submenu = new JMenu("y-axis");
		this.addParameterMenu(submenu, desc, row, column, false);
		popup.add(submenu);

		popup.addSeparator();

		// layout to add/remove row/culumn
		menu = new JMenu("Layout");
		popup.add(menu);

		item = new JMenuItem("Add Row");
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				desc.addRow(row + 1);
			}
		});
		menu.add(item);

		item = new JMenuItem("Remove Row");
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				desc.rmRow(row);
			}
		});
		menu.add(item);

		menu.addSeparator();

		item = new JMenuItem("Add Column");
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				desc.addColumn(column + 1);
			}
		});
		menu.add(item);

		item = new JMenuItem("Remove Column");
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				desc.rmColumn(column);
			}
		});
		menu.add(item);
	}

	/**
	 * Add parameter menu (x-axis, y-axis)
	 * 
	 * @param menu
	 *            Menu of axis
	 * @param desc
	 *            Description of plot-panel
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @param xAxis
	 *            Based axis of menu
	 */
	private void addParameterMenu(final JMenu menu, final PlotDescription desc, final int row, final int column,
			final boolean xAxis) {
		final JMenu results, variables;
		JMenuItem item;

		// add results
		results = new JMenu("results");
		this.addParameterItemsMenu(results, this.problem.getArchive().getResults(), desc, row, column, xAxis);
		results.addMenuListener(new MenuListener() {

			@Override
			public void menuSelected(MenuEvent event) {
				String ident;

				if (xAxis) {
					ident = desc.getXAxis(row, column);
				} else {
					ident = desc.getYAxis(row, column);
				}

				PanelPlots.selectParameterItem(results, ident);
			}

			@Override
			public void menuDeselected(MenuEvent event) {
			}

			@Override
			public void menuCanceled(MenuEvent event) {
			}
		});
		menu.add(results);
		this.addMenuScroll(results);

		// add variables
		variables = new JMenu("variables");
		this.addParameterItemsMenu(variables, this.problem.getArchive().getVariables(), desc, row, column, xAxis);
		variables.addMenuListener(new MenuListener() {

			@Override
			public void menuSelected(MenuEvent event) {
				String ident;

				if (xAxis) {
					ident = desc.getXAxis(row, column);
				} else {
					ident = desc.getYAxis(row, column);
				}

				PanelPlots.selectParameterItem(variables, ident);
			}

			@Override
			public void menuDeselected(MenuEvent event) {
			}

			@Override
			public void menuCanceled(MenuEvent event) {
			}
		});
		menu.add(variables);
		this.addMenuScroll(variables);

		// add clear axis settings
		item = new JMenuItem("clear");
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				if (xAxis) {
					desc.setXAxis(row, column, "");
				} else {
					desc.setYAxis(row, column, "");
				}
			}
		});
		menu.add(item);
	}

	/**
	 * Add parameter item (variables/results) to context-menu
	 * 
	 * @param menu
	 *            Menu to add item
	 * @param idents
	 *            Identifier of items
	 * @param desc
	 *            Description of plot-panel
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @param xAxis
	 *            Based axis of menu
	 */
	private void addParameterItemsMenu(JMenu menu, String[] idents, final PlotDescription desc, final int row,
			final int column, final boolean xAxis) {
		JCheckBoxMenuItem item;

		for (String ident : idents) {
			// add menu-item with action
			item = new JCheckBoxMenuItem(this.problem.getArchive().getProject().getObject(ident).getShortIdentifier());
			item.setToolTipText(ident);
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent event) {
					if (xAxis) {
						desc.setXAxis(row, column, ((JMenuItem) event.getSource()).getToolTipText());
					} else {
						desc.setYAxis(row, column, ((JMenuItem) event.getSource()).getToolTipText());
					}
				}
			});
			menu.add(item);
		}
	}

	/**
	 * Add scroll-bar to JMenu
	 * 
	 * @param menu
	 *            Menu to add scroll-bar
	 */
	private void addMenuScroll(final JMenu menu) {
		final JScrollBar popupScrollBar;

		popupScrollBar = new JScrollBar(Adjustable.VERTICAL);
		// changes in scroll-bar
		popupScrollBar.addAdjustmentListener(new AdjustmentListener() {
			@Override
			public void adjustmentValueChanged(AdjustmentEvent e) {
				menu.getPopupMenu().doLayout();
				menu.getPopupMenu().repaint();

				menu.doLayout();
				menu.repaint();
			}
		});
		menu.add(popupScrollBar);

		// mouse wheel in scroll-bar
		menu.getPopupMenu().addMouseWheelListener(new MouseWheelListener() {
			@Override
			public void mouseWheelMoved(MouseWheelEvent event) {
				int amount = (event.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL)
						? event.getUnitsToScroll() * popupScrollBar.getUnitIncrement()
						: (event.getWheelRotation() < 0 ? -1 : 1) * popupScrollBar.getBlockIncrement();

				popupScrollBar.setValue(popupScrollBar.getValue() + amount);
				event.consume();
			}
		});

		// layout with scroll-bar
		menu.getPopupMenu().setLayout(new LayoutManager() {

			private final int maxRows = 25;

			@Override
			public void addLayoutComponent(String name, Component comp) {

			}

			@Override
			public void removeLayoutComponent(Component comp) {

			}

			@Override
			public Dimension preferredLayoutSize(Container parent) {
				Dimension dim = new Dimension();
				int count = 0;
				JScrollBar scrollBar = null;

				// get needed size
				for (Component comp : parent.getComponents()) {
					if (comp instanceof JScrollBar) {
						scrollBar = (JScrollBar) comp;
					} else {
						if (comp.isVisible()) {
							Dimension pref = comp.getPreferredSize();
							dim.width = Math.max(dim.width, pref.width);
							if (count++ < this.maxRows) {
								dim.height += pref.height;
							}
						}
					}
				}

				Insets insets = parent.getInsets();
				dim.width += insets.left + insets.right;
				dim.height += insets.top + insets.bottom;

				// display scroll-bar
				if (scrollBar != null) {
					if (count > this.maxRows) {
						dim.width += scrollBar.getPreferredSize().width;
						scrollBar.setVisible(true);
					} else {
						scrollBar.setVisible(false);
					}
				}

				return dim;
			}

			@Override
			public Dimension minimumLayoutSize(Container parent) {
				Dimension dim = new Dimension();
				int count = 0;
				JScrollBar scrollBar = null;

				// get needed size
				for (Component comp : parent.getComponents()) {
					if (comp instanceof JScrollBar) {
						scrollBar = (JScrollBar) comp;
					} else {
						if (comp.isVisible()) {
							Dimension pref = comp.getMinimumSize();
							dim.width = Math.max(dim.width, pref.width);
							if (count++ < this.maxRows) {
								dim.height += pref.height;
							}
						}
					}
				}

				Insets insets = parent.getInsets();
				dim.width += insets.left + insets.right;
				dim.height += insets.top + insets.bottom;

				// display scroll-bar
				if (scrollBar != null) {
					if (count > this.maxRows) {
						dim.width += scrollBar.getMinimumSize().width;
					} else {
					}
				}

				return dim;
			}

			@Override
			public void layoutContainer(Container parent) {
				Insets insets = parent.getInsets();
				JScrollBar scrollBar = null;
				Dimension dim;

				int width = parent.getWidth() - insets.left - insets.right;
				int height = parent.getHeight() - insets.top - insets.bottom;

				int x = insets.left;
				int y = insets.top;
				int position = 0;

				// get scroll-bar
				for (Component comp : parent.getComponents()) {
					if ((comp instanceof JScrollBar) && comp.isVisible()) {
						scrollBar = (JScrollBar) comp;
						dim = scrollBar.getPreferredSize();
						scrollBar.setBounds((x + width) - dim.width, y, dim.width, height);
						width -= dim.width;
						position = scrollBar.getValue();
					}
				}

				// insert items
				y -= position;
				for (Component comp : parent.getComponents()) {
					if (!(comp instanceof JScrollBar) && comp.isVisible()) {
						dim = comp.getPreferredSize();
						comp.setBounds(x, y, width, dim.height);
						y += dim.height;
					}
				}

				// setting scroll-bar
				if ((scrollBar != null) && scrollBar.isVisible()) {
					scrollBar.setUnitIncrement(height / (parent.getComponentCount() - 1));
					scrollBar.setBlockIncrement((10 * height) / (parent.getComponentCount() - 1));
					scrollBar.setMinimum(0);
					scrollBar.setMaximum((((y - insets.top) + position) - height) + scrollBar.getVisibleAmount());
				}
			}
		});

	}

	/**
	 * Select context-menu entry of axis
	 * 
	 * @param menu
	 * @param identifer
	 *            Identifier of entry to select
	 */
	private static void selectParameterItem(JMenu menu, String identifer) {
		JCheckBoxMenuItem item;

		for (Component comp : menu.getMenuComponents()) {
			item = (JCheckBoxMenuItem) comp;
			item.setState(item.getToolTipText().compareTo(identifer) == 0);
		}

	}

	/**
	 * Create panel for problem
	 * 
	 * @param optim
	 *            Optimization
	 * @throws Exception
	 */
	public PanelPlots(Optimization optim) throws Exception {

		this.problem = optim.getProblem();

		this.init();
		optim.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
	}

	/**
	 * Create panel for problem
	 * 
	 * @param problem
	 *            Problem to display
	 * @throws Exception
	 */
	public PanelPlots(Problem problem) throws Exception {

		this.problem = problem;

		this.init();
	}

	/**
	 * Init panels
	 * 
	 * @throws Exception
	 */
	private void init() throws Exception {
		this.plotHypervolume = (this.problem.getNumObjectives() < 5)
				&& (this.problem.getArchive().getMemory().getNumIndividuals() < 10000);

		this.fixPlots = new Plot[PanelPlots.fixPlotNames.length];
		for (int p = 0; p < PanelPlots.fixPlotNames.length; p++) {
			if (PanelPlots.PLOT_QUALITY.equals(PanelPlots.fixPlotNames[p])) {
				// quality plots
				this.fixPlots[p] = new Plot2Line();
				PanelPlots.initQuality((Plot2Line) this.fixPlots[p]);
			} else if (PanelPlots.PLOT_OBJECTIVES.equals(PanelPlots.fixPlotNames[p])) {
				// objectives plots
				this.fixPlots[p] = new SubPlot(1, 1);
				PanelPlots.initObjectives((SubPlot) this.fixPlots[p], this.problem);
				if (this.linkPlots == null) {
					this.linkPlots = new LinkPlot();
				}
				this.linkPlots.addPlot(this.fixPlots[p]);
			} else if (PanelPlots.PLOT_VARIABLES_HIST.equals(PanelPlots.fixPlotNames[p])) {
				// variable histogram plots
				this.fixPlots[p] = new SubPlot(1, 1);
				PanelPlots.initVariablesHist((SubPlot) this.fixPlots[p], this.problem);
				if (this.linkPlots == null) {
					this.linkPlots = new LinkPlot();
				}
				this.linkPlots.addPlot(this.fixPlots[p]);
			} else if (PanelPlots.PLOT_VARIABLES.equals(PanelPlots.fixPlotNames[p])) {
				// variables plots
				this.fixPlots[p] = new SubPlot(1, 1);
				PanelPlots.initVariables((SubPlot) this.fixPlots[p], this.problem);
				if (this.linkPlots == null) {
					this.linkPlots = new LinkPlot();
				}
				this.linkPlots.addPlot(this.fixPlots[p]);
			}
		}

		// init all tabs and all plots
		for (String name : this.problem.listUserPlots()) {
			this.performEvent(new OptimizerEvent(OptimizerEvent.PLOT, "Add: " + name));
		}
	}

	/**
	 * Init objectives-plot
	 * 
	 * @param sub
	 *            Subplot to insert objective plots
	 * @param problem
	 *            Based problem for plot
	 * @throws Exception
	 */
	public static void initObjectives(SubPlot sub, Problem problem) throws Exception {
		String xLabel, yLabel;
		Plot2Scatter plot;
		ResultDesc[] obj;
		int nObjectives, nPlot, nRows, nCols;
		int row = 0, col = 0;
		ParamField field;
		Project project;

		if (problem != null) {
			project = problem.getArchive().getProject();

			obj = problem.getObjectives();
			nObjectives = obj.length;

			if (nObjectives == 1) {
				// plot for on objective

				sub.setSize(1, 1);
				field = (ParamField) project.getObject(obj[0].getName());
				xLabel = field.getShortIdentifier() + " / " + field.getUnit();
				xLabel += " (" + obj[0].getObjective().name() + ")";

				plot = new Plot2Scatter();
				plot.setXLabel(xLabel);
				plot.setYLabel(xLabel);
				plot.setGrid(true);
				plot.setHold(HOLD.FADING);
				plot.setYAxisShowZero(false);
				sub.setPlot(0, 0, plot);

			} else {
				nPlot = (nObjectives * (nObjectives - 1)) / 2;
				nRows = (int) Math.sqrt(nPlot);
				nCols = ((nPlot + nRows) - 1) / nRows;

				sub.setSize(nRows, nCols);

				for (int p = 0; p < nObjectives; p++) {
					field = (ParamField) project.getObject(obj[p].getName());

					xLabel = field.getShortIdentifier() + " / " + field.getUnit();
					xLabel += " (" + obj[p].getObjective().name() + ")";

					for (int q = p + 1; q < nObjectives; q++) {
						field = (ParamField) project.getObject(obj[q].getName());

						yLabel = field.getShortIdentifier() + " / " + field.getUnit();
						yLabel += " (" + obj[q].getObjective().name() + ")";

						plot = new Plot2Scatter();
						plot.setXLabel(xLabel);
						plot.setYLabel(yLabel);
						plot.setGrid(true);
						plot.setHold(HOLD.FADING);
						plot.setYAxisShowZero(false);
						sub.setPlot(row, col++, plot);
						if (col >= nCols) {
							col = 0;
							row++;
						}
					}
				}
			}
		}
	}

	/**
	 * Init variables-histogram-plot
	 * 
	 * @param sub
	 *            Subplot to insert plots
	 * @param problem
	 *            Based problem for plot
	 * @throws Exception
	 */
	public static void initVariablesHist(SubPlot sub, Problem problem) throws Exception {
		String xLabel, yLabel;
		Plot2Hist plot;
		VariableDesc[] var;
		int nVariables, nRows, nCols;
		int row = 0, col = 0;
		ParamField field;
		Project project;

		if (problem != null) {
			project = problem.getArchive().getProject();

			var = problem.getVariedVariables();
			nVariables = var.length;
			nCols = (int) Math.sqrt(nVariables);
			nRows = ((nVariables + nCols) - 1) / nCols;

			sub.setSize(nRows, nCols);

			for (int p = 0; p < nVariables; p++) {
				field = (ParamField) project.getObject(var[p].getName());

				xLabel = field.getShortIdentifier() + " / " + field.getUnit();
				yLabel = "number of values";

				plot = new Plot2Hist();
				plot.setXLabel(xLabel);
				plot.setYLabel(yLabel);
				plot.setGrid(true);
				plot.setHold(HOLD.FADING);
				plot.setYAxisShowZero(false);
				sub.setPlot(row++, col, plot);
				if (row >= nRows) {
					row = 0;
					col++;
				}
			}
		}
	}

	/**
	 * Init variables-plot
	 * 
	 * @param sub
	 *            Subplot to insert variables plots
	 * @param problem
	 *            Based problem for plot
	 * @throws Exception
	 */
	public static void initVariables(SubPlot sub, Problem problem) throws Exception {
		String xLabel, yLabel;
		Plot2Scatter plot;
		VariableDesc[] var;
		int nVars, nPlot, nRows, nCols;
		int row = 0, col = 0;
		ParamField field;
		Project project;

		if (problem != null) {
			project = problem.getArchive().getProject();

			var = problem.getVariedVariables();
			nVars = var.length;

			if (nVars == 1) {
				// plot for on variable

				sub.setSize(1, 1);
				field = (ParamField) project.getObject(var[0].getName());
				xLabel = field.getShortIdentifier() + " / " + field.getUnit();

				plot = new Plot2Scatter();
				plot.setXLabel(xLabel);
				plot.setYLabel(xLabel);
				plot.setGrid(true);
				plot.setHold(HOLD.FADING);
				plot.setYAxisShowZero(false);
				sub.setPlot(0, 0, plot);

			} else {
				nPlot = (nVars * (nVars - 1)) / 2;
				nRows = (int) Math.sqrt(nPlot);
				nCols = ((nPlot + nRows) - 1) / nRows;

				sub.setSize(nRows, nCols);

				for (int p = 0; p < nVars; p++) {
					field = (ParamField) project.getObject(var[p].getName());

					xLabel = field.getShortIdentifier() + " / " + field.getUnit();

					for (int q = p + 1; q < nVars; q++) {
						field = (ParamField) project.getObject(var[q].getName());

						yLabel = field.getShortIdentifier() + " / " + field.getUnit();

						plot = new Plot2Scatter();
						plot.setXLabel(xLabel);
						plot.setYLabel(yLabel);
						plot.setGrid(true);
						plot.setHold(HOLD.FADING);
						plot.setYAxisShowZero(false);
						sub.setPlot(row, col++, plot);
						if (col >= nCols) {
							col = 0;
							row++;
						}
					}
				}
			}
		}
	}

	/**
	 * Init quality-function plot
	 * 
	 * @param plot
	 *            Plot for quality
	 * @throws Exception
	 */
	public static void initQuality(Plot2Line plot) throws Exception {

		plot.setXLabel("savepoints");
		plot.setYLabel("");
		plot.setGrid(true);
		plot.setHold(HOLD.ALL);
		plot.setYLimit(0.0, 1.0);
		plot.setLegend(true);
	}

	/**
	 * Init user-defined plot
	 * 
	 * @param desc
	 *            Description of user-defined plot
	 * @param sub
	 *            Sub plot
	 * @param problem
	 *            Based problem for plot
	 */
	public static void initUsers(PlotDescription desc, SubPlot sub, Problem problem) {
		String xIdent, yIdent, xLabel, yLabel;
		Plot2Scatter plot;
		VariableDesc[] varDesc;
		ResultDesc[] resDesc;
		ParamField field;
		Project project;

		project = problem.getArchive().getProject();

		for (int r = 0; r < desc.getNumRows(); r++) {
			for (int c = 0; c < desc.getNumColumns(); c++) {
				plot = (Plot2Scatter) sub.getPlot(r, c);

				xLabel = xIdent = desc.getXAxis(r, c);
				if (!xIdent.isEmpty()) {
					field = (ParamField) project.getObject(xIdent);
					xLabel = field.getShortIdentifier() + " / " + field.getUnit();
				}

				// get identifier of y label
				yLabel = yIdent = desc.getYAxis(r, c);
				if (!yIdent.isEmpty()) {
					field = (ParamField) project.getObject(yIdent);
					yLabel = field.getShortIdentifier() + " / " + field.getUnit();
				}

				// search for identifier in variable list
				varDesc = problem.getVariables();
				for (int k = 0; k < varDesc.length; k++) {
					if (varDesc[k].getName().compareTo(xIdent) == 0) {
						xLabel = "VARIABLE: " + xLabel;
					}
					if (varDesc[k].getName().compareTo(yIdent) == 0) {
						yLabel = "VARIABLE: " + yLabel;
					}
				}

				// search for identifier in result list
				resDesc = problem.getResults();
				for (int k = 0; k < resDesc.length; k++) {
					if (resDesc[k].getName().compareTo(xIdent) == 0) {
						xLabel = "RESULT: " + xLabel;
						if (resDesc[k].isObjective()) {
							xLabel += " (" + resDesc[k].getObjective().toString() + ")";
						}
					}
					if (resDesc[k].getName().compareTo(yIdent) == 0) {
						yLabel = "RESULT: " + yLabel;
						if (resDesc[k].isObjective()) {
							yLabel += " (" + resDesc[k].getObjective().toString() + ")";
						}
					}
				}

				// init plot
				plot.setXLabel(xLabel);
				plot.setYLabel(yLabel);
				plot.setGrid(true);
				plot.setHold(HOLD.FADING);
				plot.setYAxisShowZero(false);
			}
		}
	}

	/**
	 * Get number of save-points to plot
	 * 
	 * @return Number of save-points
	 */
	public int getNumPlotSavepoints() {
		return this.plotSavepoints;
	}

	/**
	 * Set number of save-points to plot
	 * 
	 * @param num
	 *            Number of save-points to plot
	 */
	public void setNumPlotSavepoints(int num) {
		if (num < 1) {
			num = 1;
		}
		this.plotSavepoints = num;
		this.refresh();
	}

	/**
	 * Refresh plots
	 */
	public void refresh() {
		int start, end;

		try {
			end = this.problem.getNumSavePoints() - 1;
			start = (end - this.plotSavepoints) + 1;
			if (start < 0) {
				start = 0;
			}

			// select plot type
			this.plotValues.clear();
			if (PanelPlots.TYPE_GLOBAL.equals(this.plotType)) {
				this.getGlobalValues(this.plotValues, start, end);
				this.enabledPlotIndex = this.plotValues.size() - 1;
			} else if (PanelPlots.TYPE_MEMORY.equals(this.plotType)) {
				this.getMemoryValues(this.plotValues);
				this.enabledPlotIndex = this.plotValues.size() - 2;
			} else if (PanelPlots.TYPE_LOCAL.equals(this.plotType)) {
				this.getLocalValues(this.plotValues, start, end);
				this.enabledPlotIndex = this.plotValues.size() - 1;
			} else if (PanelPlots.TYPE_SOLVER.equals(this.plotType)) {
				this.getSolverValues(this.plotValues);
				this.enabledPlotIndex = this.plotValues.size() - 2;
			}

			// draw plots
			for (int p = 0; p < PanelPlots.fixPlotNames.length; p++) {
				if (PanelPlots.PLOT_QUALITY.equals(PanelPlots.fixPlotNames[p])) {
					// quality plots
					PanelPlots.drawQuality((Plot2Line) this.fixPlots[p], this.problem, this.plotHypervolume, true);
				} else if (PanelPlots.PLOT_OBJECTIVES.equals(PanelPlots.fixPlotNames[p])) {
					// global pareto plots
					PanelPlots.drawObjectives((SubPlot) this.fixPlots[p], this.problem, this.plotValues,
							this.enabledPlotIndex, this.markMapping);
				} else if (PanelPlots.PLOT_VARIABLES_HIST.equals(PanelPlots.fixPlotNames[p])) {
					// global pareto plots
					PanelPlots.drawVariablesHist((SubPlot) this.fixPlots[p], this.problem, this.plotValues,
							this.enabledPlotIndex, this.markMapping);
				} else if (PanelPlots.PLOT_VARIABLES.equals(PanelPlots.fixPlotNames[p])) {
					// global pareto plots
					PanelPlots.drawVariables((SubPlot) this.fixPlots[p], this.problem, this.plotValues,
							this.enabledPlotIndex, this.markMapping);
				}
			}

			// draw all user plots
			for (String name : this.problem.listUserPlots()) {
				PanelPlots.drawUsers(this.problem.getUserPlot(name), this.userPlots.get(name), this.problem,
						this.plotValues, this.enabledPlotIndex, this.markMapping);
			}
		} catch (Exception ex) {
			// Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE,
			// null, ex);
		}

		if (this.linkPlots != null) {
			this.linkPlots.redraw();
		}
	}

	/**
	 * Get values of global paretos
	 * 
	 * @param values
	 *            List of values to fill
	 * @param start
	 *            Start savepoint
	 * @param end
	 *            End savepoint
	 */
	private void getGlobalValues(LinkedList<Individual[]> values, int start, int end) {
		for (int k = start; k <= end; k++) {
			values.add(this.problem.getSavePoint(k).getFullPareto());
		}
	}

	/**
	 * Get values of local paretos
	 * 
	 * @param values
	 *            List of values to fill
	 * @param start
	 *            Start savepoint
	 * @param end
	 *            End savepoint
	 */
	private void getLocalValues(LinkedList<Individual[]> values, int start, int end) {
		for (int k = start; k <= end; k++) {
			values.add(this.problem.getSavePoint(k).getLocalPareto());
		}
	}

	/**
	 * Get values of solver archive
	 * 
	 * @param values
	 *            List of values to fill
	 * @throws Exception
	 */
	private void getSolverValues(LinkedList<Individual[]> values) throws Exception {
		Individual[] inds1, inds2;
		Memory mem;

		mem = this.problem.getArchive().getMemory();
		inds1 = this.problem.getSolver().getArchive();
		inds2 = this.problem.getSolver().getPareto();

		for (int k = 0; k < inds1.length; k++) {
			inds1[k] = mem.getIndividual(inds1[k]);
		}
		values.add(inds1);

		inds2 = this.problem.getSolver().getPareto();
		for (int k = 0; k < inds2.length; k++) {
			inds2[k] = mem.getIndividual(inds2[k]);
		}
		values.add(inds2);
	}

	/**
	 * Get values of memory
	 * 
	 * @param values
	 *            List of values to fill
	 */
	private void getMemoryValues(LinkedList<Individual[]> values) {
		values.add(this.problem.getArchive().getMemory().getIndividual());
		values.add(this.problem.getArchive().getMemory().getNoErrorIndividual());
		values.add(this.problem.getArchive().getMemory().getValidIndividual(this.problem));
		values.add(this.problem.getArchive().getMemory().getParetoIndividual(this.problem));
	}

	/**
	 * Filter values which are from mapping
	 * 
	 * @param val
	 *            Values
	 * @param ind
	 *            Individuals of values
	 * @return Values from mapping
	 */
	private static double[] filterMapping(double[] val, Individual[] ind) {
		int n = 0;
		double[] map;

		map = new double[val.length];
		for (int k = 0; k < val.length; k++) {
			if (ind[k].isFromMapping()) {
				map[n++] = val[k];
			}
		}

		return Arrays.copyOf(map, n);
	}

	/**
	 * Draw objectives
	 * 
	 * @param sub
	 *            Subplot for objectives
	 * @param problem
	 *            Based problem for plot
	 * @param paretos
	 *            List of paretos to draw
	 * @param enableIndex
	 *            Enable signal of index or all if value is -1
	 * @param markMapping
	 *            Mark mappings
	 * @throws PlotException
	 * @throws TDataException
	 * @throws CloneException
	 */
	public static void drawObjectives(SubPlot sub, Problem problem, LinkedList<Individual[]> paretos, int enableIndex,
			boolean markMapping) throws PlotException, TDataException, CloneException {
		int row, col;
		Plot2Scatter plot;
		Individual[] ind;
		double[] x, y;

		int nObjectives = problem.getNumObjectives();

		if (nObjectives == 1) {
			// plot for on objective

			// clear plots
			sub.getPlot(0, 0).clearSignals();
			// add signals
			for (int n = 0; n < paretos.size(); n++) {
				ind = paretos.get(n);
				x = new double[ind.length];

				plot = (Plot2Scatter) sub.getPlot(0, 0);

				for (int k = 0; k < ind.length; k++) {
					x[k] = ind[k].getResult(problem.getObjectives()[0].getName());
				}

				plot.addSignal("series", x, x);
				if ((enableIndex != -1) && (n != enableIndex)) {
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
				}

				// add mappings individuals
				if (markMapping) {
					plot.addSignal("series", PanelPlots.filterMapping(x, ind), PanelPlots.filterMapping(x, ind));
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
					plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
				}
			}

		} else {
			int nPlot = (nObjectives * (nObjectives - 1)) / 2;
			int nLines = (int) Math.sqrt(nPlot);
			int nCols = ((nPlot + nLines) - 1) / nLines;

			// clear plots
			row = 0;
			col = 0;
			for (int p = 0; p < nObjectives; p++) {
				for (int q = p + 1; q < nObjectives; q++) {
					plot = (Plot2Scatter) sub.getPlot(row, col++);
					if (col >= nCols) {
						col = 0;
						row++;
					}
					plot.clearSignals();
				}
			}

			// add signals
			for (int n = 0; n < paretos.size(); n++) {
				ind = paretos.get(n);
				x = new double[ind.length];
				y = new double[ind.length];

				row = 0;
				col = 0;
				for (int p = 0; p < nObjectives; p++) {
					for (int q = p + 1; q < nObjectives; q++) {
						plot = (Plot2Scatter) sub.getPlot(row, col++);
						if (col >= nCols) {
							col = 0;
							row++;
						}

						for (int k = 0; k < ind.length; k++) {
							x[k] = ind[k].getResult(problem.getObjectives()[p].getName());
							y[k] = ind[k].getResult(problem.getObjectives()[q].getName());
						}

						plot.addSignal("series", x, y);
						if ((enableIndex != -1) && (n != enableIndex)) {
							plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
						}

						// add mappings individuals
						if (markMapping) {
							plot.addSignal("series", PanelPlots.filterMapping(x, ind),
									PanelPlots.filterMapping(y, ind));
							plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
							plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
						}
					}
				}
			}
		}
	}

	/**
	 * Draw variables histogram
	 * 
	 * @param sub
	 *            Subplot for variables
	 * @param problem
	 *            Based problem for plot
	 * @param paretos
	 *            List of paretos to draw
	 * @param enableIndex
	 *            Enable signal of index or all if value is -1
	 * @param markMapping
	 *            Mark mappings
	 * @throws PlotException
	 * @throws TDataException
	 * @throws CloneException
	 */
	public static void drawVariablesHist(SubPlot sub, Problem problem, LinkedList<Individual[]> paretos,
			int enableIndex, boolean markMapping) throws PlotException, TDataException, CloneException {
		int row, col;
		Plot2Hist plot;
		Individual[] ind;
		double[] y;
		VariableDesc var;

		int nVaiables = problem.getNumVariedVariables();
		int nCols = (int) Math.sqrt(nVaiables);
		int nRows = ((nVaiables + nCols) - 1) / nCols;

		// clear plots
		row = 0;
		col = 0;
		for (int p = 0; p < nVaiables; p++) {
			plot = (Plot2Hist) sub.getPlot(row++, col);
			if (row >= nRows) {
				row = 0;
				col++;
			}
			plot.clearSignals();
			var = problem.getVariedVariables()[p];
			plot.setXLimit(var.getLB(), var.getUB());
			if (var.getNumSteps() < 20.0) {
				plot.setNumIntervals((int) var.getNumSteps());
			} else {
				plot.setNumIntervals(20);
			}
		}

		// add signals
		for (int n = 0; n < paretos.size(); n++) {
			ind = paretos.get(n);
			y = new double[ind.length];

			row = 0;
			col = 0;
			for (int p = 0; p < nVaiables; p++) {
				plot = (Plot2Hist) sub.getPlot(row++, col);
				if (row >= nRows) {
					row = 0;
					col++;
				}

				for (int k = 0; k < ind.length; k++) {
					y[k] = ind[k].getVariable(problem.getVariedVariables()[p].getName());
				}

				plot.addSignal("series", y);
				if ((enableIndex != -1) && (n != enableIndex)) {
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
				}

				// add mappings individuals
				if (markMapping) {
					plot.addSignal("series", PanelPlots.filterMapping(y, ind));
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
					plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
				}
			}
		}
	}

	/**
	 * Draw variables
	 * 
	 * @param sub
	 *            Subplot for variables
	 * @param problem
	 *            Based problem for plot
	 * @param paretos
	 *            List of paretos to draw
	 * @param enableIndex
	 *            Enable signal of index or all if value is -1
	 * @param markMapping
	 *            Mark mappings
	 * @throws PlotException
	 * @throws TDataException
	 * @throws CloneException
	 */
	public static void drawVariables(SubPlot sub, Problem problem, LinkedList<Individual[]> paretos, int enableIndex,
			boolean markMapping) throws PlotException, TDataException, CloneException {
		int row, col;
		Plot2Scatter plot;
		Individual[] ind;
		double[] x, y;

		int nVars = problem.getNumVariedVariables();

		if (nVars == 1) {
			// plot for on variable

			// clear plots
			sub.getPlot(0, 0).clearSignals();
			// add signals
			for (int n = 0; n < paretos.size(); n++) {
				ind = paretos.get(n);
				x = new double[ind.length];

				plot = (Plot2Scatter) sub.getPlot(0, 0);

				for (int k = 0; k < ind.length; k++) {
					x[k] = ind[k].getVariable(problem.getVariedVariables()[0].getName());
				}

				plot.addSignal("series", x, x);
				if ((enableIndex != -1) && (n != enableIndex)) {
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
				}

				// add mappings individuals
				if (markMapping) {
					plot.addSignal("series", PanelPlots.filterMapping(x, ind), PanelPlots.filterMapping(x, ind));
					plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
					plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
				}
			}

		} else {
			int nPlot = (nVars * (nVars - 1)) / 2;
			int nLines = (int) Math.sqrt(nPlot);
			int nCols = ((nPlot + nLines) - 1) / nLines;

			// clear plots
			row = 0;
			col = 0;
			for (int p = 0; p < nVars; p++) {
				for (int q = p + 1; q < nVars; q++) {
					plot = (Plot2Scatter) sub.getPlot(row, col++);
					if (col >= nCols) {
						col = 0;
						row++;
					}
					plot.clearSignals();
				}
			}

			// add signals
			for (int n = 0; n < paretos.size(); n++) {
				ind = paretos.get(n);
				x = new double[ind.length];
				y = new double[ind.length];

				row = 0;
				col = 0;
				for (int p = 0; p < nVars; p++) {
					for (int q = p + 1; q < nVars; q++) {
						plot = (Plot2Scatter) sub.getPlot(row, col++);
						if (col >= nCols) {
							col = 0;
							row++;
						}

						for (int k = 0; k < ind.length; k++) {
							x[k] = ind[k].getVariable(problem.getVariedVariables()[p].getName());
							y[k] = ind[k].getVariable(problem.getVariedVariables()[q].getName());
						}

						plot.addSignal("series", x, y);
						if ((enableIndex != -1) && (n != enableIndex)) {
							plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
						}

						// add mappings individuals
						if (markMapping) {
							plot.addSignal("series", PanelPlots.filterMapping(x, ind),
									PanelPlots.filterMapping(y, ind));
							plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
							plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
						}
					}
				}
			}
		}
	}

	/**
	 * Draw quality-function
	 * 
	 * @param plot
	 *            Plot for quality
	 * @param problem
	 *            Based problem for plot
	 * @param hypervol
	 *            Plot hypervolume
	 * @param status
	 *            Plot status of savepoints
	 * @throws PlotException
	 * @throws TDataException
	 * @throws CloneException
	 */
	public static void drawQuality(Plot2Line plot, Problem problem, boolean hypervol, boolean status)
			throws PlotException, TDataException, CloneException {
		double[][] pareto, limitsPareto = null;
		// double[][] limitsValid;
		// double[] hypervolumeValid;
		double[] hypervolumePareto;
		// double[] hypervolumeFull;
		double[] paretoSize;
		double maxParetoSize = 0;
		double[] newPareto;
		Individual[] paretoInd;
		double[] valid, invalid, error;
		double v, i, e;
		SavePoint sp;
		int numPareto = 0;
		Color[] colors;

		if (hypervol) {
			// limitsValid =
			// problem.getArchive().getObjectivesLimitsValid(prob);
			limitsPareto = problem.getArchive().getMemory().getObjectivesLimitsPareto(problem);
			for (int k = 0; k < limitsPareto[0].length; k++) {
				if (problem.getObjectives()[k].getObjective() == ObjectiveType.MAX) {
					double tmp;
					tmp = limitsPareto[0][k];
					limitsPareto[0][k] = -1.0 * limitsPareto[1][k];
					limitsPareto[1][k] = -1.0 * tmp;

					// tmp = limitsValid[0][k];
					// limitsValid[0][k] = -1.0 * limitsValid[1][k];
					// limitsValid[1][k] = -1.0 * tmp;
				}
			}
		}

		// hypervolumeValid = new double[end];
		hypervolumePareto = new double[problem.getNumSavePoints()];
		// hypervolumeFull = new double[end];
		paretoSize = new double[problem.getNumSavePoints()];
		newPareto = new double[problem.getNumSavePoints()];
		valid = new double[problem.getNumSavePoints()];
		invalid = new double[problem.getNumSavePoints()];
		error = new double[problem.getNumSavePoints()];
		for (int k = 0; k < hypervolumePareto.length; k++) {
			sp = problem.getSavePoint(k);
			if (hypervol) {
				paretoInd = sp.getLocalPareto();
				pareto = new double[paretoInd.length][];
				for (int n = 0; n < paretoInd.length; n++) {
					pareto[n] = paretoInd[n].getObjectiveValues(problem);
					for (int m = 0; m < pareto[n].length; m++) {
						if (problem.getObjectives()[m].getObjective() == ObjectiveType.MAX) {
							pareto[n][m] *= -1.0;
						}
					}
				}
				// hypervolumeValid[k] = new Hypervolume().hypervolume(pareto,
				// limitsValid, problem.getNumObjectives());
				hypervolumePareto[k] = new Hypervolume().hypervolume(pareto, limitsPareto, problem.getNumObjectives());

				// paretoInd = prob.getFullPareto(k);
				// pareto = new double[paretoInd.length][];
				// for (int n = 0; n < paretoInd.length; n++) {
				// pareto[n] = paretoInd[n].getObjectiveValues(prob);
				// if (problem.getResults()[k].getObjective() == OBJECTIVE.MAX)
				// {
				// for (int m = 0; m < pareto[n].length; m++) {
				// pareto[n][m] *= -1.0;
				// }
				// }
				// }
				// hypervolumeFull[k] = new Hypervolume().hypervolume(pareto,
				// limitsValid, problem.getNumObjectives());
			}

			// number of new pareto-points
			if (k == 0) {
				newPareto[k] = sp.getFullPareto().length;
			} else {
				boolean found;
				int numNew = 0;
				Individual[] indsOld = problem.getSavePoint(k - 1).getFullPareto();
				for (Individual ind : sp.getFullPareto()) {
					found = false;
					for (Individual indOld : indsOld) {
						if (!found) {
							found = ind.equals(indOld);
						}
					}
					if (!found) {
						numNew++;
					}
				}
				newPareto[k] = numNew;
			}

			// size of pareto-front
			paretoSize[k] = sp.getFullPareto().length;
			maxParetoSize = maxParetoSize < paretoSize[k] ? paretoSize[k] : maxParetoSize;

			// distribution of valid, invalid and error
			v = sp.getNValid();
			i = sp.getNInvalid();
			e = sp.getNError();

			valid[k] = (v + i + e) / (v + i + e);
			invalid[k] = (i + e) / (v + i + e);
			error[k] = e / (v + i + e);

			// normalize
			newPareto[k] /= (v + i + e);

		}

		// normalize values
		for (int k = 0; k < paretoSize.length; k++) {
			paretoSize[k] /= maxParetoSize;
		}

		plot.clearSignals();

		// define colors
		colors = new Color[] { PanelProjectOverview.COLOR_OK, PanelProjectOverview.COLOR_INVALID,
				PanelProjectOverview.COLOR_ERROR, Color.gray, Color.ORANGE, Color.blue.brighter() };
		if (!status) {
			colors = Arrays.copyOfRange(colors, 3, colors.length);
		}
		// if (!hypervol) {
		// colors = Arrays.copyOfRange(colors, 0, colors.length - 1);
		// }
		plot.setColorOrder(colors);

		if (status) {
			plot.getDefaultStyle().setSignalStyle(SignalStyle.Area);
			plot.addSignal("valid", valid);
			plot.addSignal("invalid", invalid);
			plot.addSignal("error", error);
		}

		plot.getDefaultStyle().setSignalStyle(SignalStyle.Line);
		plot.getDefaultStyle().setLineWidth(2.0f);
		plot.addSignal("New in pareto-front normalized", newPareto);
		if (problem.getNumSavePoints() > 0) {
			numPareto = (problem.getSavePoint(problem.getNumSavePoints() - 1).getFullPareto().length);
		}
		plot.addSignal("Size of pareto-front normalized (Current " + numPareto + ")", paretoSize);
		if (hypervol) {
			// plot.addSignal("valid", hypervolumeValid);
			plot.addSignal("hypervolume", hypervolumePareto);
			// plot.addSignal("full", hypervolumeFull);
			// plot.setLegend(true);
		}

	}

	/**
	 * Draw user-defined Plots
	 * 
	 * @param desc
	 *            Description of user-defined plot
	 * @param sub
	 *            Sub plot
	 * @param problem
	 *            Based problem for plot
	 * @param paretos
	 *            List of paretos to draw
	 * @param enableIndex
	 *            Enable signal of index or all if value is -1
	 * @param markMapping
	 *            Mark mappings
	 */
	public static void drawUsers(PlotDescription desc, SubPlot sub, Problem problem, LinkedList<Individual[]> paretos,
			int enableIndex, boolean markMapping) {
		Plot2Scatter plot;
		double[] x, y;
		Individual[] ind;
		String xIdent, yIdent;
		int xIndex = -1, yIndex = -1;
		boolean xVar = false, yVar = false;
		VariableDesc[] varDesc;
		ResultDesc[] resDesc;

		for (int r = 0; r < desc.getNumRows(); r++) {
			for (int c = 0; c < desc.getNumColumns(); c++) {
				plot = (Plot2Scatter) sub.getPlot(r, c);

				// clear plot and get settings
				plot.clearSignals();
				xIdent = desc.getXAxis(r, c);
				yIdent = desc.getYAxis(r, c);
				if (!xIdent.isEmpty() && !yIdent.isEmpty()) {
					try {
						// get location of axis
						varDesc = problem.getVariables();
						for (int k = 0; k < varDesc.length; k++) {
							if (varDesc[k].getName().compareTo(xIdent) == 0) {
								xVar = true;
								xIndex = k;
							}
							if (varDesc[k].getName().compareTo(yIdent) == 0) {
								yVar = true;
								yIndex = k;
							}
						}
						resDesc = problem.getResults();
						for (int k = 0; k < resDesc.length; k++) {
							if (resDesc[k].getName().compareTo(xIdent) == 0) {
								xVar = false;
								xIndex = k;
							}
							if (resDesc[k].getName().compareTo(yIdent) == 0) {
								yVar = false;
								yIndex = k;
							}
						}

						// get values of paretos
						for (int n = 0; n < paretos.size(); n++) {
							ind = paretos.get(n);
							x = new double[ind.length];
							y = new double[ind.length];

							for (int k = 0; k < ind.length; k++) {
								if (xVar) {
									x[k] = ind[k].getVariable(xIndex);
								} else {
									x[k] = ind[k].getResult(xIndex);
								}

								if (yVar) {
									y[k] = ind[k].getVariable(yIndex);
								} else {
									y[k] = ind[k].getResult(yIndex);
								}
							}

							plot.addSignal("series", x, y);
							if ((enableIndex != -1) && (n != enableIndex)) {
								plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
							}

							// add mappings individuals
							if (markMapping) {
								plot.addSignal("series", PanelPlots.filterMapping(x, ind),
										PanelPlots.filterMapping(y, ind));
								plot.getSignal(plot.getNumSignals() - 1).setEnable(false);
								plot.getSignal(plot.getNumSignals() - 1).getStyle().setColor(Color.GREEN);
							}
						}
					} catch (Exception e) {
					}
				}
			}
		}
	}

	/**
	 * Get selected indices
	 * 
	 * @return Indices
	 */
	public int[] getSelectedIndices() {
		Plot2Scatter plot;
		Signal sig;

		plot = (Plot2Scatter) ((SubPlot) this.getFixPlot(PanelPlots.PLOT_OBJECTIVES)).getPlot(0, 0);
		if (this.markMapping) {
			sig = plot.getSignal(this.enabledPlotIndex * 2);
		} else {
			sig = plot.getSignal(this.enabledPlotIndex);
		}

		return sig.getSelectedIndices();
	}

	/**
	 * Get suppressed indices
	 * 
	 * @return indices
	 */
	public int[] getSuppressedIndices() {
		Plot2Scatter plot;
		Signal sig;

		plot = (Plot2Scatter) ((SubPlot) this.getFixPlot(PanelPlots.PLOT_OBJECTIVES)).getPlot(0, 0);
		if (this.markMapping) {
			sig = plot.getSignal(plot.getNumSignals() - 2);
		} else {
			sig = plot.getSignal(plot.getNumSignals() - 1);
		}

		return sig.getSuppressedIndices();

	}

	/**
	 * List types of plots
	 * 
	 * @return List of strings
	 */
	public static String[] listPlotTypes() {
		return PanelPlots.plotTypes;
	}

	/**
	 * Get type of plot
	 * 
	 * @return Plot type
	 */
	public String getPlotType() {
		return this.plotType;
	}

	/**
	 * Set type of plot
	 * 
	 * @param plotType
	 *            Plot type
	 */
	public void setPlotType(String plotType) {
		if ((this.plotType == null) || !this.plotType.equals(plotType)) {
			this.plotType = plotType;
			this.refresh();
		}
	}

	/**
	 * Get number of fixed plots
	 * 
	 * @return Number of plots
	 * 
	 */
	public int getNumFixPlots() {
		return PanelPlots.fixPlotNames.length;
	}

	/**
	 * Get name of fixed plot
	 * 
	 * @param index
	 *            Index of plot
	 * @return Name of plot
	 */
	public String getFixPlotName(int index) {
		return PanelPlots.fixPlotNames[index];
	}

	/**
	 * Get fixed plot
	 * 
	 * @param index
	 *            Index of plot
	 * @return Plot
	 */
	public Plot getFixPlot(int index) {
		return this.getFixPlot(PanelPlots.fixPlotNames[index]);
	}

	/**
	 * Get index of fixed plot
	 * 
	 * @param name
	 *            Name of plot
	 * @return Index of plot
	 */
	private int getFixPlotIndex(String name) {
		int index = -1;

		for (int k = 0; k < PanelPlots.fixPlotNames.length; k++) {
			if (PanelPlots.fixPlotNames[k].compareTo(name) == 0) {
				index = k;
			}
		}

		return index;
	}

	/**
	 * Get fixed plot
	 * 
	 * @param name
	 *            Name of plot
	 * @return Plot
	 */
	public Plot getFixPlot(String name) {
		Plot plot = null;
		int index;

		index = this.getFixPlotIndex(name);
		if (index != -1) {
			plot = this.fixPlots[index];
		}

		return plot;
	}

	/**
	 * Get number of user-defined plots
	 * 
	 * @return Number of plots
	 */
	public int getNumUserPlots() {
		return this.userPlots.size();
	}

	/**
	 * Get description of user-defined plots
	 * 
	 * @param index
	 *            Index of plot
	 * @return Plot-description
	 */
	public PlotDescription getUserPlotDescription(int index) {
		String[] plots;
		PlotDescription desc = null;

		plots = this.problem.listUserPlots();
		if ((0 <= index) && (index < plots.length)) {
			desc = this.problem.getUserPlot(plots[index]);
		}

		return desc;
	}

	/**
	 * Get description of user-defined plots
	 * 
	 * @param name
	 *            Name of plot
	 * @return Plot-description
	 */
	public PlotDescription getUserPlotDescription(String name) {
		return this.problem.getUserPlot(name);
	}

	/**
	 * Get name of user-defined plot
	 * 
	 * @param index
	 *            Index of plot
	 * @return Name of plot
	 */
	public String getUserPlotName(int index) {
		PlotDescription desc;
		String name = null;

		desc = this.getUserPlotDescription(index);
		if (desc != null) {
			name = desc.getName();
		}

		return name;
	}

	/**
	 * Get user-defined plot
	 * 
	 * @param index
	 *            Index of plot
	 * @return Plot
	 */
	public Plot getUserPlot(int index) {
		String[] plots;
		Plot plot = null;

		plots = this.problem.listUserPlots();
		if ((0 <= index) && (index < plots.length)) {
			plot = this.userPlots.get(plots[index]);
		}

		return plot;
	}

	/**
	 * Get user-defined plot
	 * 
	 * @param name
	 *            Name of plot
	 * @return Plot
	 */
	public Plot getUserPlot(String name) {
		return this.userPlots.get(name);
	}

	/**
	 * Get mark mapping
	 * 
	 * @return Mark mapping
	 */
	public boolean getMarkMapping() {
		return this.markMapping;
	}

	/**
	 * Set mark mapping
	 * 
	 * @param mark
	 *            Mark mapping
	 */
	public void setMarkMapping(boolean mark) {
		this.markMapping = mark;
		this.refresh();
	}

	/**
	 * Get plot hypervolume
	 * 
	 * @return Mark mapping
	 */
	public boolean getPlotHypervolume() {
		return this.plotHypervolume;
	}

	/**
	 * Set plot hypervolume
	 * 
	 * @param plot
	 *            Plot hypervolume
	 */
	public void setPlotHypervolume(boolean plot) {
		this.plotHypervolume = plot;
		this.refresh();
	}

	/**
	 * Perform optimizer event
	 * 
	 * @param event
	 *            Optimizer event
	 */
	public void performEvent(OptimizerEvent event) {
		String name = null;
		SubPlot sub = null;
		PlotDescription desc;
		Plot2Scatter plot;
		Plot[][] plots;

		if (event.getMessage().startsWith("Add: ")) {
			name = event.getMessage().substring(5);
			sub = this.userPlots.get(name);
			if (sub == null) {
				plot = new Plot2Scatter();
				this.addMenu(plot, this.problem.getUserPlot(name), 0, 0);
				sub = new SubPlot(1, 1);
				sub.setPlot(0, 0, plot);
				this.linkPlots.addPlot(sub);
				this.userPlots.put(name, sub);
			}
		} else if (event.getMessage().startsWith("Remove: ")) {
			name = event.getMessage().substring(8);
			sub = this.userPlots.get(name);
			if (sub != null) {
				this.linkPlots.rmPlot(sub);
				this.userPlots.remove(name);
				sub = null;
			}
		} else if (event.getMessage().startsWith("Init: ")) {
			name = event.getMessage().substring(6);
			sub = this.userPlots.get(name);
		}

		if (sub != null) {
			desc = this.problem.getUserPlot(name);

			plots = new Plot[desc.getNumRows()][desc.getNumColumns()];
			for (int r = 0; r < plots.length; r++) {
				for (int c = 0; c < plots[r].length; c++) {
					plot = new Plot2Scatter();
					this.addMenu(plot, desc, r, c);
					plots[r][c] = plot;
				}
			}

			this.linkPlots.rmPlot(sub);
			sub.setPlots(plots);
			this.linkPlots.addPlot(sub);
			PanelPlots.initUsers(desc, sub, this.problem);

		}
	}

	/**
	 * Get list of enabled individuals
	 * 
	 * @return Enabled individuals
	 */
	public Individual[] getListEnabledIndividuals() {
		return this.plotValues.get(this.enabledPlotIndex);
	}
}
