# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10181 $
$LastChangedDate:: 2021-05-05 13:32:50 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import re
import sys
import traceback
import importlib

try:
    # noinspection PyUnresolvedReferences
    from java.lang import StackTraceElement
except Exception as e:
    print(e)


def eval_fnc_method(structure, *args):
    """
    Generic function to evaluate Python functions in SyMSpace

    This function is called very often, hence performance is critical!

    Parameters
    ----------
    structure : ParamPythonFunction
        Java Object of class ParamPythonFunction
    args : tuple
        - args[0] => 'hasShow', 'reset', 'refreshDependency', 'run', ....
        - args[1] => dict(RunSim=False, RemoteCompute=False, Optimize=False) if mode is 'run'

    Returns
    -------
    tuple
    """
    val = None
    errMsg = None
    errStack = None
    cp = None
    files = []
    try:
        (cp, files) = _path_load(structure)

        '''import class'''
        importlib.reload(importlib.import_module(structure.getModuleName()))
        exec('from ' + structure.getModuleName() + ' import ' + structure.getClassName())

        # module = importlib.import_module(structure.getModuleName())  # no exec/eval alternative
        # module = importlib.reload(module)                            # no exec/eval alternative

        '''load fnc'''
        fnc = eval(structure.getClassName() + '(structure)')
        # cls = getattr(module, structure.getClassName())  # no exec/eval alternative
        # fnc = cls(structure)                             # no exec/eval alternative

        '''eval method'''
        if args[0] == 'refreshDependency':
            if len(fnc.getImports()) == 0 and len(fnc.getExports()) == 0:
                fnc.setImportVars(fnc.importVars())
                fnc.setExportVars(fnc.exportVars())

        method = eval(structure.getClassName() + '.' + args[0])
        # method = fnc.__getattribute__(args[0])  # no exec/eval alternative

        if len(args) == 1:
            val = method(fnc)
            # val = method()  # no exec/eval alternative
        else:
            val = method(fnc, *(args[1:]))
            # val = method(*(args[1:]))  # no exec/eval alternative
    except:
        errStack = []
        exc_type, exc_value, exc_traceback = sys.exc_info()
        errMsg = str(exc_value)
        for trace in traceback.extract_tb(exc_traceback):
            errStack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        _path_unload(cp, files)

    return errMsg, errStack, val


def _path_load(structure):
    """
    Set dynamic system-paths in SyMSpace

    This function is called very often, hence performance is critical!

    Parameters
    ----------
    structure : ParamPythonFunction
        Java Object of class ParamPythonFunction

    Returns
    -------
    tuple
        cp: str
            Classpath
        files: list
            imported file paths (.egg, .zip)
    """
    cp = structure.getClassPath()

    files = _load_latest_dist(cp)

    if cp in sys.path:
        cp = None
    else:
        sys.path.append(cp)

    return cp, files


def _path_unload(cp: str, files: list):
    """
    Restore system-path in SyMSpace

    Parameters
    ----------
    cp: str
        Classpath
    files: list
        imported file paths (.egg, .zip)
    """
    if cp is not None:
        if cp in sys.path:
            sys.path.remove(cp)

    for file in files:
        if file in sys.path:
            sys.path.remove(file)


def _load_latest_dist(cp: str) -> list:
    """
    Search for zip-distribution files in the classpath folder and add them to the python path.
    If multiple files of the same package are found, only the latest version will be added to the python path.

    Parameters
    ----------
    cp : str
     Classpath

    Returns
    -------
    files : list
        List of file paths, which have been added to the python path
    """
    files = []

    if os.path.exists(cp):
        '''make sure the path exists'''
        pkgs = dict()
        regex_pattern = r"^(?P<pkg_name>\w+)-(?P<pkg_version>[\d]+\.[\d]+\.[\d]+)-[\w\.]+\.\w+$"
        for file in os.listdir(cp):
            '''look for files in the classpath 
            (do not use the glob functions for this e.g. 
            files = glob.glob(os.path.join(cp, '*.zip'), recursive=False)) 
            '''
            for ext in ['.zip', '.egg']:
                try:
                    if file.endswith(ext):
                        '''extract package name and versions'''
                        match_obj = re.search(regex_pattern, file)
                        if match_obj is None:
                            '''no pattern match -> skip this file'''
                            continue

                        pkg_name = match_obj.group(1)
                        pkg_version = match_obj.group(2)

                        if pkg_name not in pkgs:
                            pkgs[pkg_name] = (pkg_version, file)
                        elif parse_version(pkgs[pkg_name][0]) < parse_version(pkg_version):
                            pkgs[pkg_name] = (pkg_version, file)
                except Exception as ex:
                    print('Error in "_load_latest_dist"', ex)

        for key, value in pkgs.items():
            file = os.path.join(cp, value[1])
            if file not in sys.path:
                '''add to sys path'''
                files.append(file)
    sys.path.extend(files)
    return files


def parse_version(v):
    """Simple version parser

    Alternative to:
    - from packaging.version import parse as parse_version  # recommended
    - from distutils.version import LooseVersion as parse_version # older

    """
    return tuple(map(int, (v.split("."))))


def getVariables(loc: dict):
    import inspect

    variables = []
    for key in loc:
        if not key.startswith('_'):
            if not inspect.isfunction(loc[key]) and not inspect.ismethod(loc[key]) and not inspect.ismodule(
                    loc[key]) and not inspect.isbuiltin(loc[key]):
                variables.append(key)

    return variables


def redirect_output(output_printstream, output_encoding, error_printstream, error_encoding):
    sys.stdout = _RedirectOutputstream(output_printstream, output_encoding)
    sys.stderr = _RedirectErrorstream(error_printstream, error_encoding)


class _RedirectOutputstream:

    def __init__(self, print_stream, encoding):
        self.printmethod = getattr(print_stream, 'print')
        self.flushmethod = getattr(print_stream, 'flush')
        self.encoding = encoding

    def write(self, msg):
        self.printmethod(msg)

    def flush(self):
        self.flushmethod()


class _RedirectErrorstream:

    def __init__(self, print_stream, encoding):
        self.printmethod = getattr(print_stream, 'print')
        self.flushmethod = getattr(print_stream, 'flush')
        self.encoding = encoding

    def write(self, msg):
        self.printmethod(msg)

    def flush(self):
        self.flushmethod()
