#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10175 $
$LastChangedDate:: 2021-04-26 16:45:48 +0200#$

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import pickle
from symspace.parampythonfnc import ParamPythonFnc


def debugexternal(Cls: ParamPythonFnc):
    """
    Decorator for debugging of SyMSpace functions (ParamPythonFnc)
    """

    class DebugWrapper(Cls):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def run(self, *args, **kwargs) -> int:
            """ Run method for debugging """
            if self.debug_mode:
                '''run in debug mode: override standard methods with debug specific methods'''
                self.importValues = self.importValuesDebug
                self.exportValues = self.exportValuesDebug
                self.info = self.infoDebug
                self.warning = self.warningDebug
                self.error = self.errorDebug
                self.getParamDirectory = self.getParamDirectoryDebug
                return super().run(*args, **kwargs)
            else:
                '''create debug files'''
                class_path = self._structure.getClassPath()
                class_name = self._structure.getClassName()
                module_name = self._structure.getModuleName()
                fnc_debug = 'debug_' + module_name.replace('fnc_', '') + '.py'
                fnc_debug_path = os.path.join(class_path, fnc_debug)

                if not os.path.isfile(fnc_debug_path):
                    run_args = args[0] if len(args) == 1 else {}
                    remote_compute = run_args['RemoteCompute'] if 'RemoteCompute' in run_args else False
                    run_sim = run_args['RunSim'] if 'RunSim' in run_args else False
                    optimize = run_args['Optimize'] if 'Optimize' in run_args else False

                    with open(fnc_debug_path, 'w') as fh:
                        '''write debug starter file'''
                        fh.write('from {module_name} import {class_name}\n\n'.format(module_name=module_name,
                                                                                     class_name=class_name))
                        fh.write('dbg = {class_name}(structure=None)\n'.format(class_name=class_name))
                        fh.write('dbg.debug_mode = True\n')
                        fh.write('run_args = dict(RemoteCompute={remote_compute}, '
                                 'RunSim={run_sim}, '
                                 'Optimize={optimize})\n'.format(remote_compute=remote_compute,
                                                                 run_sim=run_sim,
                                                                 optimize=optimize))
                        fh.write('dbg.run(run_args)\n')

                pklfile = os.path.join(self._structure.getClassPath(), 'inputParameter.p')
                with open(pklfile, 'wb') as fh:
                    pickle.dump(self.importValues(), fh)

                self.info('Open {} for debugging with Python debugger'.format(fnc_debug_path))

                return super().run(*args, **kwargs)

        def importValuesDebug(self) -> dict:
            """provide importValues for DebugWrapper"""
            pklfile = 'inputParameter.p'
            if os.path.isfile(pklfile):
                with open(pklfile, 'rb') as fh:
                    data = pickle.load(fh)
            else:
                print('Pickle file not found')
                data = {}
            return data

        def exportValuesDebug(self, d: dict):
            """provide exportValues for DebugWrapper"""
            self.return_values = d

        def infoDebug(self, message: str):
            """provide info for DebugWrapper"""
            print(message)

        def warningDebug(self, message: str):
            """provide warning for DebugWrapper"""
            print(message)

        def errorDebug(self, message: str):
            """provide error for DebugWrapper"""
            # Todo: error usually just prints a message but does not raise an Exception!
            raise Exception(message)

        def getParamDirectoryDebug(self) -> str:
            """provide getParamDirectory for DebugWrapper"""
            return os.getcwd()

    return DebugWrapper


def remotecompute(requires: str, variation: list = None):
    """
    Decorator to distribute run methode to remote instance

    Parameters
    ----------
    requires : str
        Example:
        - @remotecompute('(PYTHON >= \"3.5.3\")') requires Python >= 3.5.3 at remote instance
        or with Python in java
        - @remotecompute('(PYTHON >= \"3.5.3\") && (JEP >= \"3.6.3\")') requires Python >= 3.5.3 and
          JEP >= 3.6.3 at remote instance
    variation : list
        A list of the variation import parameter keys
        The selected import parameters need to be of type np.ndarrays!

        Example:
        - @remotecompute('(PYTHON >= \"3.5.3\")', variation=['a', 'b']) requires Python >= 3.5.3 at remote instance and
          list of variation-variables

    """

    def wrap(Cls: ParamPythonFnc):

        class RemoteWrapper(Cls):

            def __init__(self, *args, **kwargs):
                self.data = {}
                super().__init__(*args, **kwargs)

            def isDistributable(self) -> bool:
                """Return distributable flag"""
                return True

            # def getUniverse(self):
            #    return 'DOCKER'

            def getJobRequirement(self):
                """Get Condor compatible string for remote machine requirements"""
                return requires

            def getVariation(self) -> list:
                """Get list of variation"""
                return variation

            def getFile(self) -> str:
                """Return current Python script"""
                class_path = self._structure.getClassPath()
                module_name = self._structure.getModuleName()
                file = os.path.join(class_path, module_name + '.py')
                return file

            def getAdditionalModules(self) -> list:
                """List of additional files to add"""
                lstModules = super().getAdditionalModules()
                if lstModules is None:
                    lstModules = []
                return lstModules

            def importValuesRemote(self) -> dict:
                """provide importValues for RemoteWrapper"""
                return self.data

            def exportValuesRemote(self, d: dict):
                """provide exportValues for RemoteWrapper"""
                for key in d:
                    self.data[key] = d[key]

            def getParamDirectoryRemote(self) -> str:
                """provide getParamDirectory for RemoteWrapper"""
                return os.getcwd()

            def infoRemote(self, message: str):
                """provide info for RemoteWrapper"""
                print(message)

            def warningRemote(self, message: str):
                """provide warning for RemoteWrapper"""
                print(message)

            def errorRemote(self, message: str):
                """provide error for RemoteWrapper"""
                # Todo: error usually just prints a message but does not raise an Exception!
                raise Exception(message)

            def runJob(self, d, run_args) -> int:
                if hasattr(Cls, 'runJob'):
                    '''run this only if the method runJob is explicitly implemented in the wrapped class'''
                    # Todo: should self.importValues/info/getParamDirectory also be available here?
                    status = super().runJob(d, run_args)
                else:
                    '''standard case'''
                    self.importValues = self.importValuesRemote
                    self.exportValues = self.exportValuesRemote
                    self.info = self.infoRemote
                    self.warning = self.warningRemote
                    self.error = self.errorRemote
                    self.getParamDirectory = self.getParamDirectoryRemote
                    self.data = d
                    status = super().run(run_args)
                return status

            def run(self, run_args) -> int:
                status = ParamPythonFnc.run(self, run_args)

                return status

        return RemoteWrapper

    return wrap


def remotedockercompute(image, variation: list = None):
    """
    Decorator to distribute run methode to remote docker instance

    Parameters
    ----------
    image : str
        Example:
        - @remotecompute('python:3.7-openjdk13') docker image with python at remote instance
    variation : list
        A list of the variation import parameter keys
        The selected import parameters need to be of type np.ndarrays!


    """

    def wrapDocker(Cls: ParamPythonFnc):

        class RemoteDockerWrapper(Cls):

            def __init__(self, *args, **kwargs):
                self.data = {}
                super().__init__(*args, **kwargs)

            def isDistributable(self) -> bool:
                """Return distributable flag"""
                return True

            def getUniverse(self) -> str:
                return 'DOCKER'

            def getDockerImage(self) -> str:
                return image

            def getJobRequirement(self) -> str:
                """Get Condor compatible string for remote machine requirements"""
                return '(TARGET.OPSYS == "LINUX")'

            def getVariation(self) -> list:
                """Get list of variation"""
                return variation

            def getFile(self) -> str:
                """Return current Python script"""
                class_path = self._structure.getClassPath()
                module_name = self._structure.getModuleName()
                file = os.path.join(class_path, module_name + '.py')
                return file

            def getAdditionalModules(self) -> list:
                """List of additional files to add"""
                lstModules = super().getAdditionalModules()
                if lstModules is None:
                    lstModules = []
                return lstModules

            def importValuesRemote(self) -> dict:
                """provide importValues for RemoteWrapper"""
                return self.data

            def exportValuesRemote(self, d: dict):
                """provide exportValues for RemoteWrapper"""
                for key in d:
                    self.data[key] = d[key]

            def getParamDirectoryRemote(self) -> str:
                """provide getParamDirectory for RemoteWrapper"""
                return os.getcwd()

            def infoRemote(self, message: str):
                """provide info for RemoteWrapper"""
                print(message)

            def warningRemote(self, message: str):
                """provide warning for RemoteWrapper"""
                print(message)

            def errorRemote(self, message: str):
                """provide error for RemoteWrapper"""
                # Todo: error usually just prints a message but does not raise an Exception!
                raise Exception(message)

            def runJob(self, d: dict, run_args: dict) -> int:
                if hasattr(Cls, 'runJob'):
                    status = super().runJob(d, run_args)
                else:
                    '''standard case'''
                    # ToDo: if remote compute then ...
                    self.importValues = self.importValuesRemote
                    self.exportValues = self.exportValuesRemote
                    self.info = self.infoRemote
                    self.warning = self.warningRemote
                    self.error = self.errorRemote
                    self.getParamDirectory = self.getParamDirectoryRemote
                    self.data = d
                    status = super().run(run_args)
                return status

            def run(self, run_args) -> int:
                return ParamPythonFnc.run(self, run_args)

        return RemoteDockerWrapper

    return wrapDocker


if __name__ == '__main__':
    pass
