#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10211 $
$LastChangedDate:: 2021-05-20 14:18:02 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import socket
import traceback
import pickle
import argparse
import logging
import glob
import importlib
import shutil


def addLogHandler(logger: logging.Logger, log_file: str, log_level: int, use_format: bool = True):
    """
    Add a file-handler for logging

    Parameters
    ----------
    logger: logging.Logger
        Logger instance
    log_file: str
        Log file to write log messages into
    log_level: int
        Logging level (see logging module)
    use_format: bool
        Use a special format for log messages
    """
    f_handler = logging.FileHandler(log_file)
    f_handler.setLevel(log_level)

    if use_format:
        f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S')
        f_handler.setFormatter(f_format)

    '''Add handlers to the logger'''
    logger.addHandler(f_handler)


def init(logger):
    """
    Load essential packages from a distribution files

    Parameters
    ----------
    logger: logging.Logger
        Logger instance
    """

    '''unpack the zip-archives'''
    logger.debug(50 * '-')
    logger.debug('Unpacking zip archives in job folder')
    zip_files = glob.glob('*.zip')
    zip_files.extend(glob.glob('*.egg'))  # get py-egg files (legacy)

    '''allow egg-files (legacy) to be unpacked like zip files'''
    shutil.register_unpack_format('egg', ['.egg'], shutil._unpack_zipfile)

    for file in zip_files:
        '''extract zip files to the working directory'''
        logger.debug(file)
        shutil.unpack_archive(file)

    logger.debug('Archives unpacked successfully')

    '''redirect the import of the parampythonfnc to the dummy parampythonfnc'''
    std_module = 'symspace.parampythonfnc'
    dummy_module = 'symspace.jobinterface.starter.dummy_parampythonfnc'
    logger.debug('Redirect "{}" to "{}"'.format(std_module, dummy_module))
    importlib.import_module(std_module)
    importlib.import_module(dummy_module)
    sys.modules[std_module] = sys.modules[dummy_module]
    logger.debug('Redirect finished')


def main(logger):
    parser = argparse.ArgumentParser(description='Run a python job')
    parser.add_argument('moduleName', help='')
    parser.add_argument('className', help='')
    parser.add_argument('infile', help='')
    parser.add_argument('outfile', help='')
    parser.add_argument("max_file_size_mb", default=100.0, type=float, help="")

    args = parser.parse_args()

    moduleName = args.moduleName
    className = args.className
    infile = args.infile
    outfile = args.outfile
    max_file_size_mb = args.max_file_size_mb

    job_dir = os.path.abspath(os.curdir)

    logger.debug('symstarter.py called on machine {}'.format(socket.gethostname()))
    logger.debug('Arguments:')
    for index, arg in enumerate(sys.argv):
        logger.debug('arg[{}] = {}'.format(index, arg))

    '''read input data'''
    logger.debug(50 * '-')
    logger.debug('Reading input data from file: {}'.format(infile))
    with open(infile, 'rb') as f:
        d = pickle.load(f)
    logger.debug('Input data read successfully')

    '''retrieve the run_args dictionary'''
    run_args = d['__run_args']

    '''import class'''
    logger.debug(50 * '-')
    module0 = os.path.splitext(moduleName)[0]
    logger.debug('Importing module: {}'.format(module0))
    module = __import__(module0)
    logger.debug('Importing class: {}'.format(className))
    classHandle = getattr(module, className)
    logger.debug('Creating Instance of class {}'.format(className))
    inst = classHandle(None)
    logger.debug('Call function runJob with arguments: d={}, run_args={}'.format('<inputDataDict>', str(run_args)))
    import_set = set(d.keys())
    status = inst.runJob(d, run_args)
    logger.debug('status runJob: {:d}'.format(status))

    if status == 0:
        '''export data'''
        logger.debug('Export data: {}'.format('<exportDataDict>'))
        logger.debug('Writing export data to file {}'.format(outfile))
        with open(outfile, 'wb') as f:
            pickle.dump(d, f)
        logger.debug('Outfile written successfully')

        '''zip result folders, since remote jobs only send files but no folders'''
        export_set = set(d.keys()).difference(import_set)
        for name in export_set:
            val = d[name]
            if isinstance(val, str):
                if os.path.isdir(val):
                    path = os.path.abspath(val)
                    folder_name = os.path.basename(path)
                    if path == os.path.join(job_dir, folder_name):
                        '''the folder is in the toplevel JobDirectory and not somewhere else'''
                        if os.path.getsize(path) < max_file_size_mb * 1024 ** 2:
                            '''check if the directory does not exceed the maximum defined sized'''

                            '''name the zip-file according to the dictionary key'''
                            target_path = os.path.join(job_dir, name)
                            zip_file = '{}.zip'.format(target_path)
                            if not os.path.exists(zip_file):
                                '''create zip file from folder'''
                                parent_folder = os.path.dirname(path)
                                shutil.make_archive(target_path, format='zip', root_dir=parent_folder, base_dir=folder_name)

                                if not os.path.isfile(zip_file):
                                    raise FileNotFoundError('File does not exist: {}'.format(zip_file))
                                if not os.path.getsize(zip_file) > 0:
                                    raise FileExistsError('File is empty: {}'.format(zip_file))

                                logger.debug('File created: {}'.format(zip_file))
                        else:
                            '''folder size is to big'''
                            # Todo: raise Exception?
                            logger.debug('Folder size is bigger than {} MB'.format(max_file_size_mb))

                    elif path.startswith(job_dir):
                        # Todo: allow this?
                        logger.debug('Transfer of Job subdirectories is not implemented: {}'.format(path))
                    else:
                        # Todo: allow this?
                        logger.debug('Transfer of directories outside of the JobDirectory is not implemented: {}'.format(path))

        logger.debug(50 * '-')
        logger.debug('Starter finished successfully!')
    else:
        raise Exception('Error: {}.runJob() received nonzero return code {:d}!'.format(className, status))


if __name__ == '__main__':
    '''Python Starter for Symspace ParamPythonFunction remote job execution'''

    sym_logger = logging.getLogger('symstarter')
    sym_logger.setLevel(logging.DEBUG)

    try:
        addLogHandler(sym_logger, 'symstarter_debug.log', logging.DEBUG)
        init(sym_logger)
        main(sym_logger)
    except Exception:
        '''Handle error'''
        sym_logger.exception("Exception occurred")

        error_logger = logging.getLogger('symstarter_error')
        error_logger.setLevel(logging.ERROR)
        addLogHandler(error_logger, 'symstarter_error.log', logging.ERROR, use_format=False)

        exc_type, exc_value, exc_traceback = sys.exc_info()
        error_logger.error(str(exc_type))
        error_logger.error(str(exc_value))
        tb = traceback.extract_tb(exc_traceback)
        for i, item in enumerate(tb):
            error_logger.error('{},{},{},{}'.format(item.name, os.path.basename(item.filename), item.lineno, item.line))
