#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision$
$LastChangedDate::                           $

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import sys
import os


class Paths:
    """Helper class to get access paths"""

    @staticmethod
    def getSyMSpaceRoot() -> str:
        """
        Get the SyMSpace root folder

        Returns
        -------
        symspace_root : str
        """
        symspace_root = os.getenv('MagOpt_Home')

        if symspace_root is None:
            raise EnvironmentError('Environment Variable "MagOpt_Home" is not defined!')
        elif not os.path.isdir(symspace_root):
            raise NotADirectoryError('"MagOpt_Home" is not a directory!')

        return symspace_root

    @staticmethod
    def getComponentSpaceRoot() -> str:
        """
        Get the ComponentSpace root folder

        Returns
        -------
        cs_root : str
        """
        try:
            from symspace.wrapper.symutils import SymConfig
            cs_root = SymConfig.getConfig('ComponentSpacePath')
        except ValueError:
            cs_root = None

        if cs_root is not None:
            cs_root = cs_root.replace('\\\\', '\\')
        else:
            cs_root = os.getenv('ComponentSpace')

        if cs_root is None:
            raise EnvironmentError('Environment Variable "ComponentSpace" is not defined!')
        elif not os.path.isdir(cs_root):
            raise NotADirectoryError('"ComponentSpace" is not a directory!')

        return cs_root

    @staticmethod
    def getPythonRoot() -> str:
        """
        Get the Python root folder

        Returns
        -------
        python_root : str
        """
        python_root = os.path.dirname(os.path.dirname(os.__file__))
        return python_root

    @classmethod
    def getCSyMSpaceExe(cls) -> str:
        """
        Get the CSyMSpace executable

        Returns
        -------
        csymspace_exe : str
        """
        symspace_root = cls.getSyMSpaceRoot()
        csymspace_exe = os.path.join(symspace_root, 'CSyMSpace.exe')

        if not os.path.isfile(csymspace_exe):
            raise FileNotFoundError('"{}" does not exist!'.format(csymspace_exe))

        return csymspace_exe

    @classmethod
    def getPythonExe(cls) -> str:
        """
        Get the Python executable

        Returns
        -------
        python_exe : str
        """
        python_root = cls.getPythonRoot()
        python_exe = os.path.join(python_root, 'python.exe')

        if not os.path.isfile(python_exe):
            raise FileNotFoundError('"{}" does not exist!'.format(python_exe))

        return python_exe

    @classmethod
    def getTestReportsDir(cls) -> str:
        """
        Get the default directory for SyMSpace test reports

        Returns
        -------
        str
        """
        symspace_root = cls.getSyMSpaceRoot()
        return os.path.join(symspace_root, 'TEST_REPORTS')

    @classmethod
    def getCSTestReportsDir(cls) -> str:
        """
        Get the default directory for ComponentSpace test reports

        Returns
        -------
        str
        """
        cs_root = cls.getComponentSpaceRoot()
        return os.path.join(cs_root, 'TEST_REPORTS')


class CMDParser:

    @staticmethod
    def jepSysArgs():
        """
        Get sys args in the correct form.

        Bug in jep returns tuple instead of list,
        this causes problems when using functions of the optparse package.

        Returns
        -------
        args : list
        """
        import symspace.osutils.processutils as pu

        if pu.getInterpreter == 'python':
            '''change nothing if run in python interpreter'''
            args = sys.argv
        elif isinstance(sys.argv, str):
            '''if run jep interpreter and only one argument exists (it is the python file), remove it'''
            args = []
        else:
            '''if run jep interpreter and multiple argument remove the first argument (the python file)'''
            args = list(sys.argv[1:])
        return args

    @staticmethod
    def jepArgPath(path: str) -> str:
        """
        Correct the path passed as commandline argument to a jep console

        SyMSpace Bug #612
        https://redmine.lcm.at/redmine/issues/612

        e.g:

        CSyMSpace.exe myscript.py 'C:\opt'

        path gets 'C\:\opt' due to a jep bug

        Parameters
        ----------
        path : str
            input path

        Returns
        -------
        new_path : str
            corrected path
        """
        path_list = list(path)
        if path_list[1] == '\\':
            path_list[1] = ''
        new_path = os.path.abspath(''.join(path_list))
        return new_path
