# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10184 $
$LastChangedDate:: 2021-05-05 14:10:10 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
from symspace.pyutils.eggpkg import EggPackage
from symspace.pyutils.zippkg import ZipPackage


class DistPkgUtil:
    """Helper class for Python distribution packages"""
    file_formats = ('zip', 'egg')
    sub_classes = dict(zip=ZipPackage, egg=EggPackage)

    @classmethod
    def isDistPkg(cls, file: str) -> bool:
        """
        Check if the path points to a supported python distribution package

        Parameters
        ----------
        file: str
            File path to check

        Returns
        -------
        bool
        """

        if os.path.isfile(file):
            all_suffixes = cls.file_formats
            for suffix in all_suffixes:
                if file.endswith(suffix):
                    return True

        return False

    @classmethod
    def getSubClass(cls, key: str):
        """
        Get the correct class according to the key

        Parameters
        ----------
        key : str
            Key which maps to a class

        Examples
        --------
        >>> zip_cls = DistPkgUtil.getSubClass('zip')
        >>> egg_cls = DistPkgUtil.getSubClass('egg')

        Returns
        -------
        sub_cls : ZipPackage or EggPackage
        """
        key = key.lower()
        sub_cls = None

        if key in cls.sub_classes:
            sub_cls = cls.sub_classes[key]
        return sub_cls

    @classmethod
    def createFromPkgDir(cls, directory: str, dist_format: str, target_dir: str = None, pkg_version: str = None,
                         compile_bytecode: bool = True, dry_run: bool = False) -> str:
        """
        Create a distribution from a package directory

        Parameters
        ----------
        directory: str
            Package directory
        dist_format: str
            Distribution format ('zip' or 'egg')
        target_dir: str
            Target directory for the created distribution file
        pkg_version: str
            Package version
        compile_bytecode: bool
            Compile py to pyc
        dry_run: bool
            Don't create a file (but return the file path, where it would be if it was created)

        Examples
        --------
        >>> dist_file = DistPkgUtil.createFromPkgDir('x:\\my_pkg', 'zip')

        Returns
        -------
        dist_pkg_file: str
            Distribution file
        """
        if os.path.isdir(directory):
            sub_cls = cls.getSubClass(dist_format)
            if sub_cls is not None:
                dist_pkg_file = sub_cls.createFromPkgDir(directory, target_dir, pkg_version, compile_bytecode, dry_run)
            else:
                raise Exception('Distribution "{}" format is not supported!'.format(dist_format))
        else:
            raise Exception('Folder "{}" not found'.format(directory))
        return dist_pkg_file

    @classmethod
    def findLatestPackages(cls, include_files: list, import_order=('zip', 'egg')) -> dict:
        """
        Find the latest packages in the list of files

        Parameters
        ----------
        include_files : list
            List of distribution files to use
        import_order : tuple
            Import order of distributions classes.
            Default: ('zip', 'egg'), first add zip-distributions then egg-distributions

        Examples
        --------
        >>> latest_pkgs = DistPkgUtil.findLatestPackages(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-1.0.0-py3.6.egg'])

        Returns
        -------
        pkgs: dict
            Collection of latest-packages
        """
        latest_pkgs = dict()
        for dist_format in import_order:
            sub_cls = cls.getSubClass(dist_format)
            if sub_cls is not None:
                tmp_latest_pkgs = sub_cls.findLatestPackages(include_files)
                for key, value in tmp_latest_pkgs.items():
                    if key not in latest_pkgs:
                        '''add entries, but do not overwrite existing'''
                        latest_pkgs[key] = value

        return latest_pkgs

    @classmethod
    def extendImportPath(cls, include_files: list, import_order=('zip', 'egg')) -> list:
        """
        Try to extend the python import path (=sys.path) with these files

        Parameters
        ----------
        include_files : list
            List of distribution files to use
        import_order : tuple
            Import order of distributions classes.
            Default: ('zip', 'egg'), first add zip-distributions then egg-distributions

        Examples
        --------
        >>> imported_files = DistPkgUtil.extendImportPath(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-1.0.0-py3.6.egg'])

        Returns
        -------
        files: list
            List of files, which have been added to the the python import path (=sys.path)
        """
        files = []
        for dist_format in import_order:
            sub_cls = cls.getSubClass(dist_format)
            if sub_cls is not None:
                f = sub_cls.extendImportPath(include_files)
                files.extend(f)
        return files

    @classmethod
    def getFilesFromDir(cls, directory: str, import_order=('zip', 'egg')) -> list:
        """

        Parameters
        ----------
        directory: str
            Directory to search
        import_order : tuple
            Import order of distributions classes.
            Default: ('zip', 'egg'), first add zip-distributions then egg-distributions

        Examples
        --------
        >>> imported_files = DistPkgUtil.getFilesFromDir('x:\\dist')

        Returns
        -------
        files: list
            List of files, which should be used to extend the python import path (=sys.path)
        """
        files = []
        if os.path.isdir(directory):
            for dist_format in import_order:
                sub_cls = cls.getSubClass(dist_format)
                if sub_cls is not None:
                    include_files = sub_cls.getFilesFromDir(directory)
                    files.extend(include_files)
        return files

    @staticmethod
    def clearImportPath(paths: list):
        """
        Remove paths (files/directories) from import path (= sys.path)

        Parameters
        ----------
        paths: list
            List of paths to remove

        Examples
        --------
        >>> DistPkgUtil.clearImportPath(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-1.0.0-py3.6.egg'])
        """
        if paths is not None:
            for path in paths:
                if path in sys.path:
                    sys.path.remove(path)
