# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10105 $
$LastChangedDate:: 2021-02-08 10:10:23 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import cProfile
import functools


def profile_wrapper(func, output_file: str = None,
                    save_snapshot: bool = True, print_out: bool = False, sort_print_by: str = 'cumtime'):
    """
    Profiling wrapper function

    Wraps the function that should be the entry point for profiling.

    Examples
    --------
    def my_function(a, b, c=3):
        print(a, b, c)

    p = profile_it(my_function, save_snapshot=False, print_out=True)
    p(1, 2, c=6)

    Parameters
    ----------
    func
        handle of the target function
    output_file: str or None. Default is None
            Path of the output file. If only name of the file is given, it's
            saved in the current directory.
            If it's None, the name of the decorated function is used.
    save_snapshot :bool
        Save the pstat result file
        A 3rd party tool for visualization is required:
        - JetBrains PyCharm: "Tools -> Open CProfile snapshot"
        - SnakeViz (a html-viewer for Python profiling data):
            pip install snakeviz
            snakeviz my_func.profile
    print_out: bool
        Print the results
    sort_print_by : str
        Sort the printed results by
        "calls"     , "call count"
        "ncalls"    , "call count"
        "cumtime"   , "cumulative time"
        "cumulative", "cumulative time"
        "file"      , "file name"
        "filename"  , "file name"
        "line"      , "line number"
        "module"    , "file name"
        "name"      , "function name"
        "nfl"       , "name/file/line"
        "pcalls"    , "primitive call count"
        "stdname"   , "standard name"
        "time"      , "internal time"
        "tottime"   , "internal time"

    Returns
    -------
    wrapper
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _output_file = output_file or func.__name__ + '.profile'
        print('Starting cProfile')
        prof = cProfile.Profile()
        retval = prof.runcall(func, *args, **kwargs)
        if print_out:
            prof.print_stats(sort=sort_print_by)
        if save_snapshot:
            prof.dump_stats(_output_file)
            print('Profiling snapshot saved to: {}'.format(os.path.abspath(_output_file)))
        return retval
    return wrapper


def profile_it(output_file: str = None, save_snapshot: bool = True, print_out: bool = False,
               sort_print_by: str = 'cumtime'):
    """
    Profiling decorator

    Decorate the function that should be the entry point for profiling.

    Examples
    --------
    @profile_it(save_snapshot=False, print_out=True)
    def my_function(a, b, c=3):
        print(a, b, c)

    my_function(1, 2)

    Parameters
    ----------
    output_file: str or None. Default is None
            Path of the output file. If only name of the file is given, it's
            saved in the current directory.
            If it's None, the name of the decorated function is used.
    save_snapshot :bool
        Save the pstat result file
        A 3rd party tool for visualization is required:
        - JetBrains PyCharm: "Tools -> Open CProfile snapshot"
        - SnakeViz (a html-viewer for Python profiling data):
            pip install snakeviz
            snakeviz my_func.profile
    print_out: bool
        Print the results
    sort_print_by : str
        Sort the printed results by
        "calls"     , "call count"
        "ncalls"    , "call count"
        "cumtime"   , "cumulative time"
        "cumulative", "cumulative time"
        "file"      , "file name"
        "filename"  , "file name"
        "line"      , "line number"
        "module"    , "file name"
        "name"      , "function name"
        "nfl"       , "name/file/line"
        "pcalls"    , "primitive call count"
        "stdname"   , "standard name"
        "time"      , "internal time"
        "tottime"   , "internal time"

    Returns
    -------
    inner
    """

    def inner(func):
        return profile_wrapper(func, output_file, save_snapshot, print_out, sort_print_by)
    return inner


if __name__ == '__main__':
    @profile_it(save_snapshot=False, print_out=True)
    def _my_function(a, b, c=3):
        print(a, b, c)
        my_list = [i for i in range(100)]
        return my_list


    _my_function(1, 2)
    # _p = profile_wrapper(_my_function, save_snapshot=False, print_out=True)
    # _p(1, 2, 6)
