# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10106 $
$LastChangedDate:: 2021-02-08 10:48:36 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
from symspace.osutils.pathutils import Paths

try:
    # Start of java imports
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.utils import MagOptConfig as JMagOptConfig
except Exception as e:
    JMagOptConfig = None
    print('Unable to load Java Libraries!\n{}'.format(e))


class SymConfig:
    """Wrapper class for MagOptConfig
    This class which might offer some alternative native-python implementations to the Java Class MagOptConfig"""

    # cl = None  # class-loader
    # classs = None  # List of java-classpath
    # pyPath = None  # List of library
    # libs = None  # List of library
    # mPath = None  # List of Matlab-path

    @staticmethod
    def __raiseImportError():
        raise ImportError('Unable to import at.lcm.bu21.magopt.utils.MagOptConfig, '
                          'no native python implementation available')

    @staticmethod
    def __getListOfStrings(jobj) -> list:
        py_obj = []
        for element in jobj:
            py_obj.append(str(element))
        return py_obj

    @classmethod
    def getHome(cls) -> str:
        if JMagOptConfig is None:
            cls.__raiseImportError()
        else:
            # noinspection PyUnresolvedReferences
            return JMagOptConfig.getHome()

    @staticmethod
    def getConfig(name: str) -> str:
        """
        Get |sym| config-entry

        Parameters
        ----------
        name: str
            Name of config-entry
        Returns
        -------
        value :str
        """
        if JMagOptConfig is None:
            '''run python implementation'''
            # Todo: probably this algorithm needs to be improves
            config = os.path.join(Paths.getSyMSpaceRoot(), 'config')
            config_local = os.path.join(Paths.getSyMSpaceRoot(), 'config.local')

            with open(config, 'r') as fh:
                lines = fh.readlines()
            with open(config_local, 'r') as fh:
                lines.extend(fh.readlines())

            value = None
            for line in lines:
                tmp_line = line.strip(' ')
                head = '{}='.format(name)
                if tmp_line.startswith(head):
                    '''the last occurrence in the config will be used'''
                    value = tmp_line.split(head)[1].strip('\n')

        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            value = JMagOptConfig.getConfig(name)

        if value is None:
            '''Error if value is None'''
            raise ValueError('Key not found in SyMSpace Config: {}'.format(name))

        return value

    @classmethod
    def getJavaPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return cls.__getListOfStrings(JMagOptConfig.getJavaPath())

    @classmethod
    def getLibPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return cls.__getListOfStrings(JMagOptConfig.getLibPath())

    @classmethod
    def getMPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return cls.__getListOfStrings(JMagOptConfig.getMPath())

    @classmethod
    def getPyPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return cls.__getListOfStrings(JMagOptConfig.getPyPath())

    @classmethod
    def getJepPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return str(JMagOptConfig.getJepPath())

    @classmethod
    def getLibraryPath(cls):
        if JMagOptConfig is None:
            '''run python implementation'''
            cls.__raiseImportError()
        else:
            '''run java implementation'''
            # noinspection PyUnresolvedReferences
            return cls.__getListOfStrings(JMagOptConfig.getLibraryPath())


if __name__ == '__main__':
    print(SymConfig.getHome())
    print(SymConfig.getConfig('PythonPath'))
    print(SymConfig.getJavaPath())
    print(SymConfig.getLibPath())
    print(SymConfig.getMPath())
    print(SymConfig.getPyPath())
    print(SymConfig.getJepPath())
    print(SymConfig.getLibraryPath())
