/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10153 $
 * $Date: 2021-03-31 18:41:56 +0200 (Wed, 31 Mar 2021) $
 */
package jobmanager.python;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import jobmanager.JobInterface;

/**
 * Invoke local Command Line Interface
 * 
 * @author M. Kitzberger
 */
public class CLILocalJepJob implements JobInterface {

	BufferedWriter bw;
	FileWriter fw;

	// constructor for testing
	public CLILocalJepJob() {

	}

	@Override
	public void runJob(String... args) throws Exception {

		this.bw = null;
		this.fw = null;

		try {

			this.setupDebugLog();

			this.bw.write("Arguments:\n");
			for (int i = 0; i < args.length; i++) {
				this.bw.write(String.format("arg%d = %s\n", i, args[i]));
			}
			this.invokeCLI(args);

		} catch (Exception ex) {

			this.bw.write(String.format("\nException occured see file starter_error.log for traceback"));
			if (this.bw != null) {
				this.bw.close();
			}
			if (this.fw != null) {
				this.fw.close();
			}
			String timeStamp = new SimpleDateFormat("HH:mm:ss yyyy.MM.dd").format(new Date());
			StringWriter errors = new StringWriter();
			ex.printStackTrace(new PrintWriter(errors));
			this.fw = new FileWriter("starter_error.log");
			this.bw = new BufferedWriter(this.fw);
			this.bw.write(String.format("Exception occured during runJob called at %s:\n\n", timeStamp));
			this.bw.write(errors.toString());

			// raise to caller
			throw ex;

		} finally {
			if (this.bw != null) {
				this.bw.close();
			}
			if (this.fw != null) {
				this.fw.close();
			}
		}
	}

	protected void setupDebugLog() throws IOException {
		String timeStamp = new SimpleDateFormat("HH:mm:ss yyyy.MM.dd").format(new Date());
		this.fw = new FileWriter("starter_debug.log");
		this.bw = new BufferedWriter(this.fw);
		this.bw.write(String.format("runJob of CLILocalJepJob called at %s:\n\n", timeStamp));
	}

	protected int invokeCLI(String... args) throws Exception {
		List<String> list = new ArrayList<String>(Arrays.asList(args));
		if (System.getProperty("os.name").equalsIgnoreCase("Linux")) {
			list.add(0, "jep");
		} else {
			list.add(0, "jep.bat");
		}

		this.bw.write("Run CLI with command:\n");
		for (String s : list) {
			this.bw.write(s);
			this.bw.write(" ");
		}
		this.bw.write("\n");

		this.bw.write("java.class.path:\n");
		this.bw.write(System.getProperty("java.class.path"));
		this.bw.write("\n");

		this.bw.write("path:\n");
		this.bw.write(System.getenv("Path"));
		this.bw.write("\n");

		ProcessBuilder pb = new ProcessBuilder();
		pb.command(list);
		pb.environment().put("CLASSPATH", System.getProperty("java.class.path"));
//		pb.inheritIO();
		final Process subprocess = pb.start();

		// log output
		new Thread() {
			@Override
			public void run() {
				FileOutputStream out = null;
				InputStream in = subprocess.getInputStream();
				int val;
				try {
					out = new FileOutputStream(new File("SyMOutput.txt"));
					while ((val = in.read()) != -1) {
						out.write(val);
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					if (out != null) {
						try {
							out.close();
						} catch (Exception e) {
						}
					}
				}

			}
		}.start();
		// log error
		new Thread() {
			@Override
			public void run() {
				FileOutputStream out = null;
				InputStream in = subprocess.getErrorStream();
				int val;
				try {
					out = new FileOutputStream(new File("SyMError.txt"));
					while ((val = in.read()) != -1) {
						out.write(val);
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					if (out != null) {
						try {
							out.close();
						} catch (Exception e) {
						}
					}
				}

			}
		}.start();

		int exitCode = subprocess.waitFor();
		this.bw.write(String.format("status: %d", exitCode));
		// Check if Python subprocess created and error file
		File f = new File(".\\symstarter_error.log");
		if (f.exists()) {
			throw new Exception("Error occured during execution (symstarter_error.log)");
		}
		return exitCode;
	}

}
